///////////////////////////////////////////////////////////////////////////
//
//	VGAPaint 386 - anim.c
//
//	General routines for handling animations.
//

#define	VGAPAINT_ANIM_C

#include <a/bits.h>

#include "vgapaint.h"
#include "anim.h"
#include "picture.h"
#include "request.h"
#include "listbox.h"
#include "asmfunc.h"

void	Anim_SetPBS_Updater(struct Slider *sld);

static struct Slider ASPBS_sliders[]={
	{
		NULL,
		{ NULL },
		NULL,		/* bitmap will be filled in */
		{ 24,22,208,8 },
		NULL,		/* use autoknob */
		0,0,		/* will be filled in */
		4,4,		/* knob sizes */
		0,0,		/* listbox.c */
		0x20000,0,		/* clicks */
		128,128,		/* bump */
		0,0,0,0,	/* positions */
		SLD_ARROW1*4 + SLD_ARROW2*6 + SLD_OUTLINE3D + SLD_ATTACHX,	// flags
		Anim_SetPBS_Updater,	// updater
		{  10,20,12,12 },	// arrow1
		{ 234,20,12,12 },	// arrow2
		-1,0,1,0,		// arrow taps
		NULL			// no controlled listbox
	}
};

static struct Requester ASPBS_req={
	0,0,
	256,36,
	REQ_BORDER + REQ_POPCEN + REQ_FORCECOL + REQ_SLDREPHIT,
	"Adjust Playback Speed",		/* no title */
	St2,
	NULL,
	NULL,
	NULL,
	&ASPBS_sliders,
	NULL,
	NULL,			// negatives
	NULL
};

void	Anim_SetPBS_Updater(struct Slider *sld) {
	if (sld->cposx<=0x10000 && sld->cposx)
		sprintf(packmem,"%08lx: %8.2f frames/sec",sld->cposx,65536.0/sld->cposx);
	else
		sprintf(packmem,"%08lx: %8.2f secs/frame",sld->cposx,sld->cposx/65536.0);
	mouse_hide();
	SysText(sld->destbm,(ASPBS_req.width-sysfontwidth*29)>>1,12,packmem,St0,St2);
	mouse_show();
}


void Anim_SetPlaybackRate(struct Picture *pict) {
	mouse_coord x,y,b;
	int	key;

	ASPBS_req.width = 42*sysfontwidth;
	ASPBS_sliders[0].boundbox.width = ASPBS_req.width-48;

	if (!OpenRequester(&ASPBS_req,0,0)) return;

	SliderAbs(&ASPBS_sliders[0],pict->playback_speed,0);

	for(;;) {
		mouse_wait_unkey();
		do
			kbhit();
		while(!mouse_buttons() && -1==(key=kbd_check()));

		if (key==' ' || key==0x1b) break;

		mouse_read2(&x,&y,&b);
		if (b&2) break;
		CheckRequester(&ASPBS_req,x,y,b);
	};
	
	mouse_wait_unkey();

	pict->playback_speed = ASPBS_sliders[0].cposx;

	CloseRequester(&ASPBS_req);
	return;
}

void Anim_Playback(struct Picture *pict,ulong playflags) {
	ulong	ticks;
	ulong	ntick;
	ulong	tick64k = 0xffff0000;
	ushort	htick1,htick2;
	boolean	back = FALSE;

	// convert playback speed to ticks

	ticks = scale32u(pict->playback_speed,1193181,65536);

	// playback at top speed

	mouse_hide();
	vd->set(&vidbm,0);

	init_timer2_counter();

	ntick = tick64k+(ulong)(htick2=read_timer2()) - ticks;

	while(!kbhit()) {

		do {
			htick1=htick2;	htick2=read_timer2();
			if (htick2>htick1) tick64k-=0x10000;
		} while((slong)(ntick-(tick64k+(ulong)htick2))<0 && !kbhit());

		if (back)
			PicturePrevFrame(pict);
		else
			PictureNextFrame(pict);

		htick1=htick2;	htick2=read_timer2();
		if (htick2>htick1) tick64k-=0x10000;
			
		vd->bitblt(pict->bm,0,0,&vidbm,0,0,vidbm.width,vidbm.height-8);
		if (pict->current->frame_pal)
			LoadFramePalette(pict,pict->current);
		vid_printf6(&vidbm,0,vidbm.y2-8,St2,St0,"%5d/%-5d  last: %08lx(%lx)  size %7ld",pict->current_num,pict->total_frames,
			ticks+ntick-(tick64k+(ulong)htick2),ticks,pict->current->frame_siz);

		htick1=htick2;	htick2=read_timer2();
		if (htick2>htick1) tick64k-=0x10000;
			
		vd->refresh();

		if (pict->current_num == pict->total_frames) {
			if (playflags & ANIM_PLAYBACK_PINGPONG)
				back=TRUE;
			else if (playflags & ANIM_PLAYBACK_ONCE)
				break;
		} else if (pict->current_num == 1) {
			if (playflags & ANIM_PLAYBACK_PINGPONG)
				back=FALSE;
		}

		ntick -= ticks;
	}
	mouse_show();
	if (kbhit()) getch();

	restore_screen();
}

/////////////////////////////////////////////////////////////////////////////

//	ulong size = AnimSize(struct Picture *pict);
//
//	Returns the total size of an animation.

ulong	AnimSize(struct Picture *pict) {
	struct	Frame *frame;
	ulong	size=0;
	int	i;

	frame = pict->current;
	for(i=0;i<pict->total_frames;i++) {
		if ((i+1)==pict->current_num && pict->dirty)
			size += pict->maxframe;
		else
			size += frame->frame_siz;
		frame = frame->next;
	}

	return size;
}


////////////////////// Anim>Process>Align Palettes //////////////////////////

//	boolean success = AnimAlignPalettes(struct Picture *pict);
//
//	Align adjacent frame palettes for highest compression.
//
//	Returns TRUE on success, FALSE on failure.

static uchar *mhs_pal;
static int mhs_r,mhs_g,mhs_b;

static int match_histosort(const uchar *a,const uchar *b) {
	int	err1,err2;

	err1 = ((int)mhs_pal[0+3L**a]-mhs_r)*((int)mhs_pal[0+3L**a]-mhs_r)
	     + ((int)mhs_pal[1+3L**a]-mhs_g)*((int)mhs_pal[1+3L**a]-mhs_g)
	     + ((int)mhs_pal[2+3L**a]-mhs_b)*((int)mhs_pal[2+3L**a]-mhs_b);
	err2 = ((int)mhs_pal[0+3L**b]-mhs_r)*((int)mhs_pal[0+3L**b]-mhs_r)
	     + ((int)mhs_pal[1+3L**b]-mhs_g)*((int)mhs_pal[1+3L**b]-mhs_g)
	     + ((int)mhs_pal[2+3L**b]-mhs_b)*((int)mhs_pal[2+3L**b]-mhs_b);
	return (int)(err1>err2)-(int)(err1<err2);
}

boolean	AnimAlignPalettes(struct Picture *pict) {
	int	i,j,k;
	uchar	*trn,*matchmtx;
	ulong	*histoptr;
	struct FramePalette *fp;
	struct StatusBar *sb;

	// allocate translation buffer memory 
	if (!(trn = malloc(1024+768+768+4096+256*256))) return FALSE;
	matchmtx = trn + 1024+768*2 + 4096;
	histoptr = (ulong *)(trn + 1024+768*2);
	mhs_pal = trn+1024;

	// seek to the first frame
	PictureGotoFrame(pict,pict->first,1);

	// go through each frame in turn

	sb = CreateStatusBar(pict->total_frames,"Aligning Frame Palettes...");

	for(i=0;i<pict->total_frames;i++) {
		if (kbd_check()==0x1b) break;

		// feather the palettes, so that each frame has a distinct palette

		fp = pict_cur->current->frame_pal;

		if (fp ? fp->usecount>1 : TRUE) {	// multiplexed palette or no palette?
			if (!fp)
				memcpy(trn+1024,pict->palette,768);
			else {	// multiplexed palette
				memcpy(trn+1024,fp->palette,768);
				DetachFramePalette(pict,pict->current);
			}

			if (!AttachPaletteToFrame(pict,pict->current,NULL)) {
				DestroyStatusBar(sb);
				free(trn);
				return FALSE;
			} 
		} else memcpy(trn+1024,fp->palette,768);

		// trn+1024: current frame
		// trn+1792: last frame

		if (i) {
			uchar	*cmtx;
			uchar *dst,*src;

			// take a histo

			Histogram(histoptr,pict->bm);

			// match current palette to last palette

			for(j=0;j<256;j++) trn[j+256]=0;

			for(j=0;j<256;j++) {
				for(k=0;k<256;k++) {
					matchmtx[256*j+k] = k;
				}
				mhs_r = mhs_pal[768+3*j+0];
				mhs_g = mhs_pal[768+3*j+1];
				mhs_b = mhs_pal[768+3*j+2];
				qsort(matchmtx+256*j,256,1,match_histosort);
			}

			cmtx = matchmtx;
			for(j=0;j<256;j++) {
				if (histoptr[j] || !j) {
					for(k=0;k<256;k++) if (!trn[(ulong)cmtx[k]+256]) break;
					trn[j] = cmtx[k];
					trn[(ulong)cmtx[k]+256]=1;
				}
				cmtx += 256;
			}
			cmtx = matchmtx;
			for(j=1;j<256;j++) {
				if (!histoptr[j]) {
					for(k=0;k<256;k++) if (!trn[(ulong)cmtx[k]+256]) break;
					trn[j] = cmtx[k];
					trn[(ulong)cmtx[k]+256]=1;
				}
				cmtx += 256;
			}

			for(j=0;j<256;j++)
				trn[256+trn[j]] = j;

			vid13_translate(pict->bm,pict->bm->x1,pict->bm->y1,pict->bm->width,pict->bm->height,trn+256);

			dst = pict_cur->current->frame_pal->palette;
			src = trn + 1024;
			for(j=0;j<256;j++) {
				dst[0] = src[trn[j]*3+0];
				dst[1] = src[trn[j]*3+1];
				dst[2] = src[trn[j]*3+2];
				dst+=3;
			}

			memcpy(trn+1024,pict_cur->current->frame_pal->palette,768);

			PictureMarkDirty(pict);
		}

		memcpy(trn+1792,trn+1024,768);
		PictureNextFrame(pict);
		ModifyStatusBar(sb,i+1);
	}

	DestroyStatusBar(sb);
	free(trn);
	return TRUE;
}




