#define	VGAPAINT_BRUSH_C

#include "vgapaint.h"
#include "sprite.h"
#include "brush.h"
#include "picture.h"

void OffBrush(void) {
	sprite_hide2(&brush_sprite);
};

void OnBrush(void) {
	sprite_show2(&brush_sprite);
};

/* PasteBrush() is defined as a macro in defines.h */

void PasteBrush2(BMP bm,coord x,coord y,coord color) {
	if (color>=0 && brush) {
		PasteBrush(bm,x-brush_hx,y-brush_hy);
		TAP_RECT(x-brush_hx,y-brush_hy);
		TAP_RECT(x+brush_sprite.width-brush_hx,y+brush_sprite.height-brush_hy);
	} else {
		br_plotback(bm,x,y,color);
		if (color>=0) TAP_RECT(x,y);
	};
}

boolean AllocBrush(coord x,coord y) {
	FreeBrush();
	if (!(PictureInit(&brush_picture,&brush_bm,x,y)))
		return FALSE;
	if (!(brushsave=malloc(vid13_rassize(x,y)))) {
		FreeBrush();
		return FALSE;
	};
	brush_sprite.width=x;
	brush_sprite.height=y;
	brush=TRUE;
//	vid_initBitMap(&brushbm,brush_bm.vm,0,0,brush_sprite.width,brush_sprite.height,brush_bm.bpr,picbm->pal);
	return TRUE;
}

void FreeBrush(void) {
	OffBrush();
	if (brush_picture_alloc) {
		PictureDispose(&brush_picture);
		brush_picture_alloc = FALSE;
	}
	if (brushsave){ free(brushsave); brushsave=brush_sprite.savebm.vm=NULL;};
}

void GetBrush(coord x1,coord y1,coord x2,coord y2) {
	coord xs,ys;

	if (x2<x1) swap(coord,x1,x2);
	if (y2<y1) swap(coord,y1,y2);

	xs=x2+1-x1;	ys=y2+1-y1;
	vid_rectclip(winbm,&x1,&y1,&xs,&ys);
	x2=x1+xs-1;	y2=y1+ys-1;

	if (!AllocBrush(x2+1-x1,y2+1-y1)) {
		Message("Out of memory!");
		return;
	};
	mouse_hide();
//	vid13_bitblt(winbm,x1,y1,&brushbm,0,0,brush_sprite.width,brush_sprite.height);
	vid13_bitblt(winbm,x1,y1,&brush_bm,0,0,brush_sprite.width,brush_sprite.height);
	mouse_show();
	PictureMarkDirty(&brush_picture);
	brush_handle_readjust();
	brush_ox=(x1+vposx) % brush_sprite.width;
	brush_oy=(y1+vposy) % brush_sprite.height;

	brush_sprite.trans_colr=bcolor;
	if (opt.autotransp) {
		uchar c;

		c=vd->point(winbm,x1,y1);
		if (c==vd->point(winbm,x1,y2))
		  if (c==vd->point(winbm,x2,y1))
		    if (c==vd->point(winbm,x2,y2))
		    	brush_sprite.trans_colr=c;
	};

	sprite_init(&brush_sprite,brush_bm.vm,brushsave,winbm,brush_bm.bpr); 	/* have to follow protocol */
	memcpy(brush_picture.palette,piccol,768);			/* copy colors */
}

void GetStdBrush(int id) {
	FreeBrush();
	brush=FALSE;
	brush_sprite.width=1;
	brush_sprite.height=1;

	if (!AllocBrush(brush_sprite.width,brush_sprite.height)) return;
	brush_bm.vm[0] = 1;
	PictureMarkDirty(&brush_picture);
	brush_sprite.flags=SPR_NOPATTERN;
	brush_hx=brush_sprite.width>>1;
	brush_hy=brush_sprite.height>>1;
	brush_ox=brush_oy=0;
	brush=(id==0) ? FALSE : TRUE;
	sprite_init(&brush_sprite,brush_bm.vm,brushsave,winbm,brush_bm.bpr);
	brush_sprite.trans_colr = 0;
}

void BrushMode(int cnt) {
	if (brush_bm.vm) {
		OffBrush();
		brush_sprite.flags &= ~SPR_DRAWMODE;
		switch(cnt) {
		case 0: brush_sprite.flags += SPR_MATTE; break;
		case 1: brush_sprite.flags += SPR_NOPATTERN; break;
		case 2: brush_sprite.flags += SPR_REPLACE; break;
		case 3: brush_sprite.flags += SPR_NOIMAGE; break;
		};
		sprite_init(&brush_sprite,brush_bm.vm,brushsave,winbm,brush_bm.bpr);
		OnBrush();
	};
}


void brush_flipx(void) {
	int i;

	OffBrush();
	if (brush) {
		for(i=0;i<brush_picture.total_frames;i++) {
//			vid13_flipx(&brushbm,0,0,brush_sprite.width,brush_sprite.height);
			vid13_flipx(&brush_bm,0,0,brush_sprite.width,brush_sprite.height);
			PictureMarkDirty(&brush_picture);
			PictureNextFrame(&brush_picture);
		}
		sprite_reinit(&brush_sprite);
	};
}

void brush_flipy(void) {
	int i;

	OffBrush();
	if (brush) {
		for(i=0;i<brush_picture.total_frames;i++) {
//			vid13_flipy(&brushbm,0,0,brush_sprite.width,brush_sprite.height);
			vid13_flipy(&brush_bm,0,0,brush_sprite.width,brush_sprite.height);
			PictureMarkDirty(&brush_picture);
			PictureNextFrame(&brush_picture);
		}
		sprite_reinit(&brush_sprite);
	};
}

/* brush_prepare_translate: ready the brush structures for a change operation
** returns TRUE if OK to continue; change, then brush_finish_translate() */

void brush_translate(long width,long height,void (*func)(BMP dst,BMP src,void *),void *data) {
	int	i;

	if (brush) {
		OffBrush();
		brush_oldbm=brush_bm;
		brush_old_picture = brush_picture;
		brush = FALSE;
		brush_picture_alloc = FALSE;

		/* snag the bitmap away so it's not freed */
		
		if (!AllocBrush(width,height)) {
			/* oops... */
			brush_bm.vm = brush_oldbm.vm;
			brush_picture = brush_old_picture;
			brush = TRUE;
			brush_picture_alloc = TRUE;
			FreeBrush();
			GetStdBrush(0);
			return;
		};

		brush_old_picture.bm = &brush_oldbm;

		PictureGotoFrame(&brush_old_picture,brush_old_picture.first,1);
		for(i=0;i<brush_old_picture.total_frames;i++) {
			if (i) InsertAfterFrame(&brush_picture);
			func(&brush_bm,&brush_oldbm,data);
			PictureMarkDirty(&brush_picture);
			if (i)	DeleteCurrentFrame(&brush_old_picture);
			else	PictureNextFrame(&brush_old_picture);
		}
		PictureNextFrame(&brush_picture);
		PictureDispose(&brush_old_picture);
		sprite_init(&brush_sprite,brush_bm.vm,brushsave,winbm,brush_bm.bpr);
		brush_handle_readjust();
	};
}

void brush_rotright_func(BMP dst,BMP src,void *data) {
	vid13_rotateright(dst,0,0,src,0,0,src->width,src->height);
}

void brush_rotright(void) {
	brush_translate(brush_sprite.height,brush_sprite.width,brush_rotright_func,NULL);
}

void brush_handle_readjust(void) {
	OffBrush();
	switch(brush_hpos) {
	case 0: case 3: case 6:	brush_hx=0; break;
	case 1: case 4: case 7:	brush_hx=brush_sprite.width>>1; break;
	case 2: case 5: case 8:	brush_hx=brush_sprite.width-1; break;
	};
	switch(brush_hpos) {
	case 0: case 1: case 2:	brush_hy=0; break;
	case 3: case 4: case 5:	brush_hy=brush_sprite.height>>1; break;
	case 6: case 7: case 8:	brush_hy=brush_sprite.height-1; break;
	};
}

void brush_handle_cycle(void) {
	if (++brush_hpos > 8) brush_hpos=0;
	brush_handle_readjust();
}

void brush_handle_center(void) {
	brush_hpos=4;
	brush_handle_readjust();
}

void brush_scale_func(BMP dst,BMP src,void *data) {
	vid13_scale(src,0,0,src->width,src->height,dst,0,0,dst->width,dst->height);
}

void brush_scale(void) {
	mouse_coord x,y,b,ox,oy;
	coord ax,ay;

	dabort=FALSE;
	OffBrush();

	/* drag around the brush's rectangle with LR handle */

	mouse_read2(&x,&y,&b);
	b=0;
	do {
		mouse_hide();
		vd->box(winbm,x,y,x+1-brush_sprite.width,y+1-brush_sprite.height,STI+0x80000000);
		mouse_show();
		ox=x; oy=y;
		mouse_readmove(&x,&y,&b);
		mouse_hide();
		vd->box(winbm,ox,oy,ox+1-brush_sprite.width,oy+1-brush_sprite.height,STI+0x80000000);
		mouse_show();
		if (dabort) goto abort;
	} while(!b);
	ax=x+1-brush_sprite.width;
	ay=y+1-brush_sprite.height;

	/* resize the rectangle with the LR handle again */

	do {
		mouse_hide();
		vd->box(winbm,ax,ay,x,y,STI+0x80000000);
		mouse_show();
		ox=x; oy=y;
		mouse_readmove(&x,&y,&b);
		if (x<ax) x=ax;
		if (y<ay) y=ay;
		mouse_hide();
		vd->box(winbm,ax,ay,ox,oy,STI+0x80000000);
		mouse_show();
		if (dabort) goto abort;
	} while(b);

	/* actually scale the brush */

	brush_translate(x+1-ax,y+1-ay,brush_scale_func,NULL);

	return;
abort:
	mouse_wait_unkey();
	return;
}

void brush_halve(void) {
	brush_translate(brush_bm.width>>1,brush_bm.height>>1,brush_scale_func,NULL);
}

void brush_double(void) {
	brush_translate(brush_bm.width<<1,brush_bm.height<<1,brush_scale_func,NULL);
}

void brush_double_x(void) {
	brush_translate(brush_bm.width<<1,brush_bm.height,brush_scale_func,NULL);
}

void brush_double_y(void) {
	brush_translate(brush_bm.width,brush_bm.height<<1,brush_scale_func,NULL);
}

/////////////////

struct BrushFreeRotateData {
	long	cx,cy;
	double	angle;
};

void brush_free_rotate_func(BMP dst,BMP src,struct BrushFreeRotateData *bfrd) {
	vd->set(dst,brush_sprite.trans_colr);
	vid13_rotate(src,0,0,src->width,src->height,src->width>>1,src->height>>1,dst,bfrd->cx,bfrd->cy,-bfrd->angle);
}

void brush_free_rotate(void) {
	mouse_coord x,y,b,ox,oy;
	coord ax,ay;
	coord xs1,xs2,ys1,ys2;
	long cx,cy,sx,sy;
	double angle;
	struct BrushFreeRotateData bfrd;

	dabort=FALSE;
	OffBrush();

	/* drag around the brush's rectangle with center handle */

	xs1=brush_sprite.width>>1;
	ys1=brush_sprite.height>>1;
	mouse_read2(&x,&y,&b);
	b=0;
	do {
		mouse_hide();
		vd->box(winbm,x-xs1,y-ys1,x+brush_sprite.width-xs1-1,y+brush_sprite.height-ys1-1,STI+0x80000000);
		mouse_show();
		ox=x; oy=y;
		mouse_readmove(&x,&y,&b);
		mouse_hide();
		vd->box(winbm,ox-xs1,oy-ys1,ox+brush_sprite.width-xs1-1,oy+brush_sprite.height-ys1-1,STI+0x80000000);
		mouse_show();
		if (dabort) goto abort;
	} while(!b);
	ax=x;	ay=y;
	xs2=brush_sprite.width-xs1;
	ys2=brush_sprite.height-ys1;
	do {
		mouse_hide();
//		vid_printf6(&vidbm,0,0,ST0,ST2,"Angle: %6.3f\n",((angle=Angle(ax,ay,x,y))*180.0)/M_PI);
		SetTitle("Angle: %6.3f\n",((angle=Angle(ax,ay,x,y))*180.0)/M_PI);
		vd->line(winbm,ax,ay,x,y,STI+0x80000000);
		vid_polyline(winbm,STI+0x80000000,3+0x80000000,
			(coord)(ax+xs1*cos(angle)+ys1*sin(angle)),(coord)(ay-xs1*sin(angle)+ys1*cos(angle)),
			(coord)(ax-xs2*cos(angle)+ys1*sin(angle)),(coord)(ay+xs2*sin(angle)+ys1*cos(angle)),
			(coord)(ax-xs2*cos(angle)-ys2*sin(angle)),(coord)(ay+xs2*sin(angle)-ys2*cos(angle)),
			(coord)(ax+xs1*cos(angle)-ys2*sin(angle)),(coord)(ay-xs1*sin(angle)-ys2*cos(angle))
		);
		mouse_show();
		ox=x;	oy=y;
		mouse_readmove(&x,&y,&b);
		mouse_hide();
		vd->line(winbm,ax,ay,ox,oy,STI+0x80000000);
		vid_polyline(winbm,STI+0x80000000,3+0x80000000,
			(coord)(ax+xs1*cos(angle)+ys1*sin(angle)),(coord)(ay-xs1*sin(angle)+ys1*cos(angle)),
			(coord)(ax-xs2*cos(angle)+ys1*sin(angle)),(coord)(ay+xs2*sin(angle)+ys1*cos(angle)),
			(coord)(ax-xs2*cos(angle)-ys2*sin(angle)),(coord)(ay+xs2*sin(angle)-ys2*cos(angle)),
			(coord)(ax+xs1*cos(angle)-ys2*sin(angle)),(coord)(ay-xs1*sin(angle)-ys2*cos(angle))
		);
		mouse_show();
	} while(b);

	/* rotate the brush */

	vid_rotate_size(brush_sprite.width,brush_sprite.height,xs1,ys1,&sx,&sy,&cx,&cy,angle);

	bfrd.cx = cx;
	bfrd.cy = cy;
	bfrd.angle = angle;
	brush_translate(sx,sy,brush_free_rotate_func,&bfrd);

abort:
	mouse_wait_unkey();
}

double Angle(long ax,long ay,long x,long y) {
	double a1;

	x-=ax;	y=ay-y;
	if (x==0) if (y>0) return M_PI/2; else if (y<0) return (M_PI*3)/2; else return 0;

	if (abs(y)>=abs(x)) {
		a1=atan((double)y/(double)abs(x));
	} else
		a1=(y<0 ? ((M_PI*3)/2) : (M_PI/2))-atan((double)abs(x)/(double)y);

	if (x<0) a1=M_PI-a1;
	if (a1<0) a1+=2*M_PI;

	return a1;
}

void 	brush_remap(void) {
	OffBrush();
	vid_make_remap(brush_picture.palette,pict_cur->palette,packmem,256,256);
	vid13_translate(&brush_bm,0,0,brush_sprite.width,brush_sprite.height,packmem);
	sprite_init(&brush_sprite,brush_bm.vm,brushsave,winbm,brush_bm.bpr);
}

void	BrushNextFrame(void) {
	OffBrush();
	PictureNextFrame(&brush_picture);
	SetTitle(NULL);
}

void	BrushPrevFrame(void) {
	OffBrush();
	PicturePrevFrame(&brush_picture);
	SetTitle(NULL);
}

void	BrushInsertFrameBefore(void) {
	OffBrush();
	InsertBeforeFrame(&brush_picture);
	SetTitle(NULL);
}

void	BrushInsertFrameAfter(void) {
	OffBrush();
	InsertAfterFrame(&brush_picture);
	SetTitle(NULL);
}

void	BrushDeleteCurrentFrame(void) {
	if (brush_picture.total_frames>1) {
		OffBrush();
		DeleteCurrentFrame(&brush_picture);
		SetTitle(NULL);
	}
}

