//////////////////////////////////////////////////////////////////////////
//
//	VGAPaint 386 - help.c
//
//	Help driver.  Why do so many people need help?
//

#include <ctype.h>
#include <a/bits.h>

#include "help.h"
#include "video.h"
#include "request.h"
#include "listbox.h"
#include "io.h"
#include "toolicon.h"

//	struct HelpFileHandle *handle = OpenHelpFile(char *filename);
//
//	Opens a help file and returns a handle for that file.
//
//	Returns NULL if the open fails for any reason.

struct HelpFileHandle *OpenHelpFile(char *name) {
	struct HelpFileHandle *hfh;
	struct {
		ulong id,version;
	} hhd;

	if (hfh = malloc(sizeof(struct HelpFileHandle))) {
		if (hfh->f = VPfopen(name,"rb")) {
			if (1==fread(&hhd,sizeof hhd,1,hfh->f)) {
				if (hhd.id=='PLHV' && hhd.version==0) {
					if (1==fread(hfh->hash_table,256*sizeof(ulong),1,hfh->f)) {
						return hfh;
					}
				}
			}
		}
	}
	
	return NULL;
}

//	void CloseHelpFile(struct HelpFileHandle *hfh);
//
//	Closes a previously opened help file and disposes the handle
//	associated with it;

void CloseHelpFile(struct HelpFileHandle *hfh) {
	fclose(hfh->f);
	free(hfh);
}

//	char *DistillHelpTopicName(char *dest,char *src);
//
//	Converts a help topic name to a more generalized form.
//
//	Returns <dest>.
//
//	The following transformations occur:
//	- spaces are converted to dashes
//	- uppercase chars become lowercase

char * DistillHelpTopicName(char *dest,char *src) {
	uchar c,lc = 'z';
	char *od=dest;

	while(c=*src++) {
		if (c==' ') {
			if (lc!=' ')
				*dest++='-';
		} else
			*dest++ = tolower(c);
		lc = c;
	}
	*dest++=0;

	return od;
}

//	uchar HashHelpTopicName(char *s);
//
//	Reduce a help topic down to a hash character.

uchar	HashHelpTopicName(char *s) {
	char c;
	uchar hash=0;

	while(c=*s++)
		hash = brotl(hash,5) ^ c;

	return hash;
}

//	ulong FindHelpTopic(struct HelpFileHandle *hfh,char *topic,ulong *len_ptr);
//
//	Returns the start and length of a topic in the help file.
//
//	Returns NULL if the topic could not be found.

ulong FindHelpTopic(struct HelpFileHandle *hfh,char *topic,ulong *len_ptr) {
	ulong	pos;
	uchar	*t,ch;

	// hash the name down and look it up in the hash table

	pos = hfh->hash_table[HashHelpTopicName(topic)];
	if (!pos) return NULL;		// no hash chain, can't exist

	// seek to the position in the file and look for the entry

	if (fseek(hfh->f,pos,SEEK_SET)) return NULL;

	for(;;) {
		if (1!=fread(&pos,sizeof pos,1,hfh->f)) return NULL;
		if (!pos) return NULL;
		if (1!=fread(len_ptr,sizeof(ulong),1,hfh->f)) return NULL;

		// this is really sick; do the compare as we read the name

		t=topic;
		do {
			if (EOF==(ch=getc(hfh->f))) return NULL;
			if (ch!=*t++) break;
		} while(ch);

		if (ch==t[-1]) return pos;	// match!

		// if we stopped because our string was too short, read the rest

		if (ch)
			do {
				if (EOF==(ch=getc(hfh->f))) return NULL;
			} while(ch);
	}

}

//	struct HelpTopic *LoadHelpTopic(struct HelpFileHandle *hfh,ulong pos,ulong len,ulong wrap);
//
//	Loads a help topic into memory.
//
//	Returns NULL if the topic could not be loaded for any reason.

struct HelpTopic *LoadHelpTopic(struct HelpFileHandle *hfh,ulong pos,ulong len,ulong wrap) {
	struct HelpTopic *ht;
	struct	HelpLineChain *hlc_end=NULL;

	// seek to the topic start

	if (fseek(hfh->f,pos,SEEK_SET)) return NULL;

	// allocate memory to hold the whole thing

	if (!(ht = malloc(sizeof(struct HelpTopic)))) return NULL;
	memset(ht,0,sizeof(struct HelpTopic));
	if (!(ht->topic_mem = malloc(len+1))) goto abort;

	// read in the topic

	if (1!=fread(ht->topic_mem,len,1,hfh->f)) goto abort;
	ht->topic_mem[len]=1;

	// trace through the topic and break it into 50char lines

	{
		uchar *s=ht->topic_mem;		// current position
		uchar *break_start=NULL;    	// last position where we can start break
		uchar *break_end=NULL;		// last position where we can end break
		uchar *line_start=ht->topic_mem;// start of line
		uchar c;		    	// current char
		int chars=0;			// number of characters in current line
		int bchars=0;			// number of characters after last break
		boolean	in_space_break=FALSE;
		boolean in_link=FALSE;
		boolean	first_search=FALSE;
		struct HelpLineChain *hlc;

		do {
			c=*s;

			if (isspace(c)) {
				bchars=0;
				if (!in_space_break) break_start = break_end = s;
				in_space_break=TRUE;
				if (first_search) {
					++s;
					continue;
				}
			} else {
				if (in_space_break) {
					break_end = s;
					in_space_break=FALSE;
				}
				if (first_search) {
					line_start=s;
					first_search=FALSE;
				}
			}
			if (c==174)
				in_link=TRUE;
			else if (c==175)
				in_link=FALSE;
			else if (c=='|' && in_link) {
				while((c=*s++)!=1) if (c==175) break;
				if (c==1) --s;
				--s;
				in_link=FALSE;
			} else if (c!=29 && c!=19) {
				++bchars;
				if (++chars>wrap || c==1 || !c) {
					if (!hlc_end ? TRUE : hlc_end->line_cnt>=sizeof(hlc_end->lines)/sizeof(hlc_end->lines[0])) {
						if (!(hlc = malloc(sizeof(struct HelpLineChain)))) goto abort;

						if (!ht->hlc) ht->hlc=hlc;
						hlc->line_cnt=0;
						hlc->next=NULL;
						hlc_end->next = hlc;
						hlc_end = hlc;
					}
					hlc_end->lines[hlc_end->line_cnt].s = line_start;
					if (!break_start || c==1 || !c) {
						hlc_end->lines[hlc_end->line_cnt].l = s-line_start;
						line_start = s;
						if (!c) ++line_start;
						first_search = FALSE;
						chars=0;
					} else {
						hlc_end->lines[hlc_end->line_cnt].l = break_start-line_start;
						line_start = break_end;
						if (!(first_search=in_space_break))
							chars = bchars;
						else	chars=0;
					}
					hlc_end->line_cnt++;
					break_start=NULL;
					++ht->total_lines;
				}
			}
			++s;
		} while(c!=1);
	}
	return ht;

abort:
	UnloadHelpTopic(ht);
	return NULL;
}

//	void UnloadHelpTopic(struct HelpTopic *ht);
//
//	Unloads a help topic from memory.

void UnloadHelpTopic(struct HelpTopic *ht) {
	struct HelpLineChain *hlc,*hlc2;

	hlc=ht->hlc;
	while(hlc2=hlc) {
		hlc=hlc2->next;
		free(hlc2);
	}

	free(ht->topic_mem);
	free(ht);
}

//	int result = DisplayHelpTopic(struct HelpFileHandle *hfh,char *topic);
//
//	Displays the help text associated with a topic in the given help
//	file.
//
//	Returns:
//		0 if help was not successfully displayed,
//		1 if the help topic didn't exist and "no help" was displayed
//		2 if the user selected "Cancel,"
//		3 if the user selected "Ok."

static void *DisplayHelpGetIndex(struct Listbox *lbox,int idx) {
	struct HelpLineChain *hlc = lbox->listadd;

	while(hlc) {
		if (idx<hlc->line_cnt)
			return hlc->lines+idx;
		idx -= hlc->line_cnt;
		hlc=hlc->next;
	}

	return NULL;
}

static void DisplayHelpDIndex(struct Listbox *lbox,void *entry,int ypos,int highlight) {
	struct HelpLine *hl = entry;
	coord y = lbox->y+lbox->textheight*ypos;
	coord x = lbox->x;
	int len;
	uchar c,*s,*t,*lim;
	boolean	htstate=FALSE;
	boolean	center=FALSE;
	boolean reverse=FALSE;
	color	fore,back,tcolor;

	lbox->clearIndex(lbox,ypos);
	if (hl) {
		// scan through the string looking for   pairs

		s = t	= hl->s;
		lim	= s+hl->l;

		while(s<lim) {
			c=*s++;
			if (c==174 || c==175 || s==lim || (c=='|' && htstate) || c==19 || c==29) {
				len = s-t;
				if (c==174 || c==175 || c==19 || c==29 || (c=='|' && htstate)) --len;

				fore = htstate ? St1 : St0;
				back = htstate ? St3 : St2;

				if (reverse) {
					tcolor=fore; fore=back; back=tcolor;
				}

				if (len)
					SysTextLim(lbox->destbm,x,y,t,fore,back,len);
				x += sysfontwidth*len;

				if (c=='|' && htstate) {
					while(s<lim) {
						if (*s++==175) break;
					}
				}

				t=s;
				if (c==174) htstate=TRUE; else htstate=FALSE;
				if (c==19) reverse=!reverse;
				else if (c==29) center=TRUE;
			}
		}
		if (reverse)
			vd->bar(lbox->destbm,x-1,y,lbox->x+lbox->width+1-x,lbox->textheight,St0);
		if (center) {
			color xd;

			// this is a lame way to center, but it's easy...

			xd = (lbox->width - (x - lbox->x))>>1;
			vd->hblit(lbox->destbm,lbox->x,y,lbox->width,lbox->textheight,xd);
			vd->bar(lbox->destbm,lbox->x,y,xd,lbox->textheight,back);
		}
	}
}

#define HELP_COLS (49)
#define	HISTORY_DEPTH	(32)

int DisplayHelpTopic(struct HelpFileHandle *hfh,char *topic) {
	char name[128],name2[128];
	ulong	history_pos[HISTORY_DEPTH];
	ulong	history_len[HISTORY_DEPTH];
	int	history_depth=0;
	ulong	pos;
	ulong	len;
	struct	ListRequester *lr;
	static char *dumb_table[]={ 0 };
	struct HelpTopic *ht;
	mouse_coord	mx,my,mb;
	ulong		msg;
	boolean		cont;
	boolean		quit=FALSE;
	boolean		help_tried=FALSE;

	if (!hfh) return 0;

	// distill the topic name down to an acceptable form
	// - change any number of spaces to dashes
	// - change uppercase to lowercase
	// - hash the name down at the same time

	DistillHelpTopicName(name,topic);

	// open the requester

	if (!(lr = MakeListRequester(HELP_COLS,19,"VP386 Help"))) {
		Message("Not enough memory to display help.");
		return 0;
	}

	// put in a stupid dummy table
	lr->tbox->table = dumb_table;
	TextboxCreate(lr->tbox,TRUE);

	lr->lbox->getIndex	= DisplayHelpGetIndex;
	lr->lbox->displayIndex	= DisplayHelpDIndex;
	lr->lbox->listadd	= NULL;
	lr->lbox->alen		= 0;
	lr->zones[1].text	= "Back";

	// activate the requster

	if (!OpenListRequesterAutosize(lr))
		goto open_failure;

	while(!quit) {
		// attempt to load topic

		if (!stricmp(name,"go-back")) {
			if (history_depth) {
				--history_depth;
				pos = history_pos[history_depth];
				len = history_len[history_depth];
			}
		} else {
			char *t,*slash;

			help_tried = FALSE;

			while(!(pos = FindHelpTopic(hfh,name,&len))) {
				if (help_tried) goto open_failure_1;

				// trace back through the name, try to cd ..

				t=name;
				slash = NULL;
				while(*t) {
					if (*t=='/') slash=t;
					++t;
				}
				if (slash) *slash=0; else {
					strcpy(name,"no-help");
					help_tried=TRUE;
				}
			}
		}

		if (!(ht = LoadHelpTopic(hfh,pos,len,HELP_COLS))) goto open_failure_1;

		lr->lbox->listadd	= (void *)ht->hlc;
		lr->lbox->alen		= ht->total_lines;
		lr->lbox->sel		= -1;
		lr->lbox->seladd	= NULL;

		mouse_hide();
		lr->lbox->clearWindow(lr->lbox);
		lr->sld->cposy=0;
		SliderRecompute(lr->sld);
		DrawListbox(lr->lbox);
		mouse_show();

		if (strcmp(name,"no-help")) {
			if (history_depth==HISTORY_DEPTH) {
				history_depth--;
				memmove(history_pos,history_pos+1,sizeof(ulong)*(HISTORY_DEPTH-1));
			}
			history_pos[history_depth] = pos;
			history_len[history_depth] = len;
			++history_depth;
		}

		cont=TRUE;

		do {
			mouse_read2(&mx,&my,&mb);
			if (mb & 1) {
				msg=CheckRequester(lr->req,mx,my,mb);
				mouse_wait_unkey();
				switch(msg) {
				case REQM_LISTBOX+0:
					{
					struct HelpLine *hl;
					uchar *str,*end,*t,c;
					int ch;
					char *start_link=NULL;
					boolean in_link=FALSE;
					boolean center=FALSE;

					// isolate line

					if (!(hl = DisplayHelpGetIndex(lr->lbox,
						lr->lbox->cpos+(my-lr->lbox->y-lr->req->reqbm.ybias)/lr->lbox->textheight))) break;

					// trace through line looking for the link

					str = hl->s;
					end = str + hl->l;
					ch = (mx - lr->lbox->x - lr->req->reqbm.xbias)/sysfontwidth;
					while(str<end) {
						c = *str++;
						if (c==174) {
							in_link = TRUE;
							start_link = str;
						} else if (c==175) {
							in_link = FALSE;
							start_link = NULL;
						} else if (in_link && c=='|') {
							while(str<end)
								if (*str++==175) break;
							in_link = FALSE;
							start_link=NULL;
						} else if (c==29) {
							center=TRUE;
						} else if (c!=19) {
							if (!ch) break;
							--ch;
						}
					}

					// did we hit a hyperlink?

					if (!ch && in_link) {
						t = name2;
						while(str<end) {
							c=*str++;
							if (c=='|') t=name2;
							else if (c==175) break;
							else *t++=c;
						}
						*t=0;

						DistillHelpTopicName(name,name2);
					}
					cont=FALSE;

					}break;
				case RZoneMsg(0,0):	/* Ok */
					quit=TRUE;
					cont=FALSE;
					break;
				case RZoneMsg(0,1):	/* Back */
					cont=FALSE;
					strcpy(name,"go-back");
					if (history_depth) --history_depth;
					break;
				};
			};
		} while(cont);

		UnloadHelpTopic(ht);
	}

	CloseRequester(lr->req);

	// cleanup

	DisposeListRequester(lr);

	return 3;

open_failure_1:	CloseRequester(lr->req);
		DisposeListRequester(lr);
open_failure:	Message("Unable to open help.");
		return 0;
}
