#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <dos.h>

#include <a/types.h>
#include <a/bits.h>

struct HashLink {
	struct HashLink *next;
	ulong	offset;
	ulong	length;
	char	name[1];
} *hash_chain[256]={NULL};

ulong	vphelp_header[258];
char line_buffer[256];

struct	HashLink *current_topic = FALSE;
int	line_number=0;
char	*source_fn;

uchar	help_hash(uchar *str) {
	uchar c;
	uchar hash=0;

	while(c=*str++)
		hash = brotl(hash,5) ^ c;

	return hash;
}

void	Error(char *s) {
	printf("%s(%ld): Error! %s\n",source_fn,line_number,s);
	exit(EXIT_FAILURE);
}

void	InsertImage(FILE *fo,char *n);

int main(int argc,char **argv) {
	FILE *fi,*fo;
	char *t;
	int i;
	boolean showhashtable=FALSE;

	printf("\nVP386 Help File Distiller - Version 0.9 by Avery Lee\n\n");

	if (argc<3) {
		puts(	"distill [options] <source file> <dest file>\n"
			"\t"	"-h\t"	"show hash table lengths"
			);
		return EXIT_FAILURE;
	}

	while(argv[1][0]=='-') {
		switch(argv[1][1]) {
		case 'h':
			showhashtable=TRUE;
			break;
		default:
			printf("Bad option: -%c\n",argv[1][1]);
		}
		++argv;
	}

	printf("Distilling %s to %s.\n",argv[1],argv[2]);
	source_fn = argv[1];

	if (!(fi=fopen(argv[1],"r"))) {
		printf("Error opening %s for read.\n",argv[1]);
		return EXIT_FAILURE;
	}

	if (!(fo=fopen(argv[2],"wb"))) {
		printf("Error opening %s for write.\n",argv[2]);
		return EXIT_FAILURE;
	}

	vphelp_header[0]='PLHV';	// VHLP tag
	vphelp_header[1]=0;		// version 0
	memset(vphelp_header+2,0,256*sizeof(long));	// hash table

	if (1!=fwrite(vphelp_header,sizeof vphelp_header,1,fo)) goto write_error;

	// start processing lines from the helpfile

	while(fgets(line_buffer,sizeof line_buffer,fi)) {
		++line_number;

		// kill the newline at the end of the string; if there
		// isn't one, kill the user's dreams

		t = line_buffer+strlen(line_buffer)-1;

		if (*t!='\n')
			Error("Line too long");
		else *t=0;

		// check for a '#' sign at the start

		if (*line_buffer=='#') {

			// advance through the line looking for either a space or the null

			t=line_buffer;
			while(*t && !isspace(*t)) ++t;
			if (*t) *t++=0;

			if (!stricmp(line_buffer+1,"topic")) {
				struct HashLink *hl;
				uchar h = help_hash(t);

				// terminate a running topic

				if (current_topic) {
 					current_topic->length = ftell(fo)-current_topic->offset;
 					fputc(0x01,fo);
				}

				// allocate a node

				if (!(hl = malloc(sizeof(struct HashLink)+strlen(t)))) goto memory_error;

				strcpy(hl->name,t);
				hl->offset = ftell(fo);

				// link node into hash chain

				hl->next = hash_chain[h];
				hash_chain[h] = hl;

				current_topic=hl;
			} else if (!stricmp(line_buffer+1,"end")) {
				if (!current_topic) Error("#end without #topic");
				current_topic->length = ftell(fo)-current_topic->offset;
				current_topic=NULL;
			} else if (!stricmp(line_buffer+1,"reverse")) {
				if (!current_topic) Error("#reverse not inside topic");
				fputc(19,fo);
			} else if (!stricmp(line_buffer+1,"center")) {
				if (!current_topic) Error("#center not inside topic");
				fputc(29,fo);
			} else if (!stricmp(line_buffer+1,"irev")) {
				if (!current_topic) Error("#irev not inside topic");
				fputc(19,fo);
				fputs(t,fo);
				fputc(19,fo);
			} else if (!stricmp(line_buffer+1,"title")) {
				if (!current_topic) Error("#title not inside topic");
				fputc(29,fo);
				fputc(19,fo);
				fputs(t,fo);
				fputc(0,fo);
#if 0
			} else if (!stricmp(line_buffer+1,"image")) {
				InsertImage(fo,t);
#endif
			} else if (line_buffer[1]) {
				Error("Unknown # directive");
			}
		} else {
			if (!*line_buffer) {
				if (current_topic) fputc(0,fo);
			} else if (!current_topic)
				Error("Help text not within topic");
			else {
				// look at the last character in the buffer; if it is not a "~" then
				// add a null at the end

				t=line_buffer;
				while(*t) ++t;
				if (t>line_buffer) --t;

				if (*t=='~') {
					if (1!=fwrite(line_buffer,t-line_buffer,1,fo)) goto write_error;
				} else {
					if (1!=fwrite(line_buffer,t+2-line_buffer,1,fo)) goto write_error;
				}
			}
		}
	}

	if (current_topic)
		Error("Help topic not terminated");

	// write out hash tables

	for(i=0;i<256;i++) {
		struct HashLink *hl;
		int htlen = 0;
		if (hl = hash_chain[i]) {

			vphelp_header[i+2] = ftell(fo);

			while(hl) {
				if (1!=fwrite(&hl->offset,sizeof(ulong)*2+1+strlen(hl->name),1,fo)) goto write_error;
				hl=hl->next;
				++htlen;
			}
			fputc(0,fo);
			fputc(0,fo);
			fputc(0,fo);
			fputc(0,fo);
			if (showhashtable && htlen) printf("Hash chain %3d: length %ld\n",i,htlen);
		}
	}

	// seek back to the start and rewrite out to header

	if (fseek(fo,0,SEEK_SET)) goto write_error;
	if (1!=fwrite(vphelp_header,sizeof vphelp_header,1,fo)) goto write_error;

	return EXIT_SUCCESS;
write_error:
	Error("Unable to write to file");
memory_error:
	Error("Out of memory");
}

#if 0

struct Color {
	uchar red,grn,blu;
};

struct PCXHeader {
	uchar			magic;		// 00
	uchar			version;	// 01
	char			encoding;	// 02
	char			bitsPerPixel;	// 03
	short			x1,y1,x2,y2;	// 04
	short			xdpi,ydpi;	// 0c
	struct	Color		palette[16];	// 10
	uchar			videomode; /* reserved */
	char			planes;
	short			bytesPerRow;
	short			palinfo;
	short			xsize;
	short			ysize;
	char			filler[54];
};

char readerr[]="Unable to read PCX file";
char formaterr[]="PCX file bad";
char memoryerr[]="Out of memory";

boolean unpack_PCX_row(uchar *dmem,FILE *file,long xs) {
	long left=xs;
	long c,c2;
	while(left) {
		if (-1==(c=fgetc(file))) return TRUE;
		if (c>=0xC0) {
			if (-1==(c2=fgetc(file))) return TRUE;
			if ((left-=(c-0xC0))<0) return TRUE;
			for(c-=0xC0;c;c--) *dmem++=(uchar)c2;
		} else {
			*dmem++=c;
			--left;
		};
	};
	return FALSE;
}

long pack_PCX_row(FILE *fo,uchar *src,long left) {
	uchar data,new,count=0;
	uchar *start=dst;

	data=*src;

	while(left--) {
		new=*src++;
		if (count==63 || (new!=data && count)) {
			if (count>=2 || data>=0xc0)
				fputc(count+0xc0,fo);
			fputc(data,fo);
			data=new;
			count=1;
		} else data=new,++count;
	};
	if (count) {
		if (count>=2 || data>=0xc0) fputc(count+0xc0,fo);
		*dst++=data;
	};
	return dst-start;
}

void InsertImage(FILE *fo,char *t) {
	FILE *fi;
	struct PCXHeader pcxh;
	uchar	*image_mem,*iptr;
	int	y;

	// open the file

	if (!(fi = fopen(t,"rb"))) Error("Unable to open PCX file");

	// read in the header

	if (1!=fread(&pcxh,sizeof pcxh,1,fi)) Error(readerr);

	// process header

	if (pcxh.magic != 0x0a) Error(formaterr);
	if (pcxh.planes!=1 || pcxh.bitsPerPixel!=8) Error("PCX file must be 8-bit");

	// allocate memory for image

	if (!(image_mem = malloc(pcxh.xsize*pcxh.ysize))) Error(memoryerr);

	// decompress the picture

	iptr = image_mem;
	for(y=pcxh.ysize;y;y--) {
		unpack_PCX_row(iptr,fi,pcxh.xsize);
		iptr += pcxh.xsize;
	}

	// compress the picture

}

#endif

