#define	VGAPAINT_IO_C

#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>

#include "vgapaint.h"
#include "io.h"


///////////////////////////////////////////////////////////
//
//	void SetSystemFlags(void);
//
//	Set the system flags in the <sys> variable
//	so that we know whether we should <ahem> sidestep
//	idiosyncracies.  (Or hack through them.)


void	SetSystemFlags(void) {
	struct RealInt rregs;
	void *mptr;
	ushort desc;
	int i;

	// detect VESA

	if (mptr=DosMemoryAllocate(17,&desc)) {
		memset(mptr,0,264);
		memset(&rregs,0,sizeof rregs);
		rregs.w.ax=0x4f00;
		rregs.es=(ulong)mptr>>4;
		rregs.w.di=(ulong)mptr&15;
		DPMIRealInt(0x10,&rregs);

		if (rregs.w.ax == 0x004f) {
			sys.VESA = TRUE;
			if (((ushort *)mptr)[2]>=0x0200)
				sys.VESA2_0 = TRUE;
		}

		DosMemoryFree(desc);
	}

	// detect multitaskers

	desc = DetectMWindows();

	if (desc) sys.MTask = TRUE;
	if (desc>=0x0100)	// Detect all Windows
		sys.Windows=TRUE;

	if (desc>=0x0400)	// Detect Windows 95 and NT
		sys.Win95=TRUE;

	// look for Linux DOSEMU

	if (!memcmp("02/25/93",(void *)0xFFFF5,8)) {
		if (mptr = DosMemoryAllocate(1,&desc)) {
			static uchar DOSEMUdetect[]={
				0x31,0xc0,		// xor al,al
				0xcd,0xe6,		// int e6
				0x2e,			// cs:
				0xa3,0x09,0x00,		// mov [0009],ax
				0xcf			// iret
			};

			memcpy(mptr,DOSEMUdetect,sizeof DOSEMUdetect);

			// scan from 80-FF (not E6) looking for an unused interrupt

			for(i=0x80;i<0x100;i++)
				if (!*(ulong *)(i*4) && i!=0xE6) {
					*(ulong *)(i*4) = ((ulong)mptr)<<12;
					DPMIRealInt(i,&rregs);
					if (*(ushort *)((char *)mptr + 9) == 0xaa55)
						sys.LinuxDOSEMU=TRUE;
					*(ulong *)(i*4) = NULL;
					break;
				}

			DosMemoryFree(desc);
		}
	}

	// Fix for Caldera OpenDOS: check for LFN support regardless of
	// whether Windows 95 is installed

	if (mptr = DosMemoryAllocate(3,&desc)) {
		strcpy((uchar *)mptr+32,"c:\\");

		rregs.w.ax = 0x71a0;	// get volume info
		rregs.w.cx = 32;	/// fstype buffer

		rregs.ds   = ((ulong)mptr>>4)+2; // DS:DX = drive path, eg 'c:\'
		rregs.w.dx = ((ulong)mptr & 15);

		rregs.es = ((ulong)mptr >> 4);  // ES:DI = filesystem type buffer
		rregs.w.di = ((ulong)mptr & 15);

		DPMIRealInt(0x21,&rregs);
		opt.useLFN = ( rregs.w.ax != 0x7100);
		DosMemoryFree(desc);
	}

//	opt.useLFN = sys.Win95;

	sys.realDPMI = Detect_True_DPMI();
}


ushort DetectMWindows(void) {
	union REGS r;
	r.w.ax = 0x1600;
	int386( 0x2F, &r, &r );
	if( r.h.al == 0x00 || r.h.al == 0x80 ) {

		// Not running MS-Windows -- are we running OS/2?

        	r.h.ah = 0x30;
	        intdos( &r, &r );
	        if( r.h.al >= 20 ) return 20;    // running OS/2
		return 0;
	}

	if( r.h.al == 0x01 || r.h.al == 0xff )
		// MS-Windows/386 version 2.xx
		return 0x200;

	// MS-Windows version AL.AH
	return ( r.h.al << 8 ) | r.h.ah;
}

int kbd_check(void) {
#if 0
	if (!kbhit())
		return -1;
	else {
		vid_setRGB6(0,255,255,255);
		regs.h.ah=1;
		int386(0x16,&regs,&regs);
		if (!regs.h.al & !regs.h.ah) return -1;
		if (regs.h.ah != 0xE0) {
			regs.h.ah=0;
			int386(0x16,&regs,&regs);
		}
		vid_setRGB6(0,0,0,0);

//			if (regs.h.ah==0x4A) {
//				PrtSc_Interrupt();
//				restore_scrn=TRUE;
//			};

		if (regs.h.al)
			return (short)regs.h.al;
		else if (regs.h.ah==0xE0) {
			regs.h.ah=1;
			int386(0x16,&regs,&regs);
			return (short)regs.h.ah+0x200;
		} else	return (short)regs.h.ah+0x100;
	};
#else
	int asm_kbd_check(void);
	return asm_kbd_check();
#endif
}

uchar kbd_shift(void) {
	regs.h.ah=0x02;
	int386(0x16,&regs,&regs);
	return (uchar)regs.h.al;
}

ulong read_tick(void) {
//	regs.h.ah=0x00;
//	int386(0x1a,&regs,&regs);
//	return ((ulong)regs.w.cx << 16) + (ushort)regs.w.dx;

	return *(ulong *)0x46c;		// I like speed
}

void *DosMemoryAllocate(ushort paragraphs,ushort *desc) {
	union REGS regs;

	regs.x.eax=0x0100;	/* DPMI allocate DOS memory */
	regs.x.ebx=paragraphs;	/* allocate paras */
	int386(0x31,&regs,&regs);
	*desc=regs.w.dx;
	if (regs.x.cflag) return 0;
	else return (void *)((regs.x.eax & 0xFFFF)<<4);
}

void DosMemoryFree(ushort descriptor) {
	union REGS regs;

	regs.x.eax=0x0101;	/* DPMI free DOS memory */
	regs.x.edx=descriptor;
	int386(0x31,&regs,&regs);
}

void	DPMIRealInt(uchar i,struct RealInt *ri) {
	ri->sp=ri->ss=0;
	regs.w.ax=0x0300;
	regs.h.bl=i;
	regs.h.bh=0;
	regs.w.cx=0;
	regs.x.edi=(ulong)ri;
	segread(&sregs);
	int386x(0x31,&regs,&regs,&sregs);
}


ushort	DPMIAllocateSelector(void) {
	union REGS regs;

	regs.x.eax=0;
	regs.x.ecx=1;
	int386(0x31,&regs,&regs);
	if (regs.x.cflag & INTR_CF) return NULL;
	return regs.w.ax;
}

void	DPMIFreeSelector(ushort sel) {
	union REGS regs;

	regs.x.eax = 1;
	regs.w.bx = sel;
	int386(0x31,&regs,&regs);
}

void	DPMISetSelector(ushort sel,ulong base,ulong lim,ushort access) {
	union REGS regs;

	regs.x.eax=9;
	regs.w.bx=sel;
	regs.w.cx=access;
	int386(0x31,&regs,&regs);

	regs.x.eax=7;
	regs.w.bx=sel;
	regs.w.cx=base>>16;
	regs.w.dx=base;
	int386(0x31,&regs,&regs);

	regs.x.eax=8;
	regs.w.bx=sel;
	regs.w.cx=lim>>16;
	regs.w.dx=lim;
	int386(0x31,&regs,&regs);
}

//////// The following DPMI routines will only work under PMODE/W -- NOT DOS/4GW (toy extender...)

//	Detect_True_DPMI()
//
//	Returns TRUE if running under PMODE/W or a DPMI server, FALSE if running under DOS/4GW

boolean	Detect_True_DPMI(void) {
	union REGS regs;

	regs.w.ax=0x0306;
	regs.w.bx=1;		// set these values for ID
	regs.w.cx=4;
	int386(0x31,&regs,&regs);

	// Rational screwed DOS/4GW and 4GW/Pro by not including callbacks

//	if (!(regs.w.bx==1 && regs.w.cx==4)) {printf("PMODE/W\n"); getch();}

	return !(regs.w.bx==1 && regs.w.cx==4);
}


//	CBack=DPMI_AllocCBack(PMProc,buf32)
//
//	Allocate a DPMI callback address

ulong DPMI_AllocCBack(void (*PMProc)(void),void *buf32) {
	union REGS regs;
	struct SREGS sregs;

 	regs.w.ax=0x0303;
//	segread(&sregs);
	regs.x.esi=(ulong)FP_OFF( (void (far *)(void)) PMProc);
	regs.x.edi=(ulong)FP_OFF( (void far *) buf32);
	sregs.ds=(ushort)FP_SEG( (void (far *)(void)) PMProc);
	sregs.es=(ushort)FP_SEG( (void far *) buf32);
	int386x(0x31,&regs,&regs,&sregs);

	if (regs.x.cflag & INTR_CF) return 0; else return ((ulong)(ushort)regs.w.cx<<16)+(ushort)regs.w.dx;
}

//	DPMI_FreeCBack(ulong CBack)
//
//	Free a DPMI callback address

void	DPMI_FreeCBack(ulong CBack) {
	union REGS regs;

	regs.w.ax=0x0304;
	regs.w.cx=CBack>>16;
	regs.w.dx=(ushort)CBack;
	int386(0x31,&regs,&regs);
}


//	DPMI_Yield(void)
//
//	Release VP386's current time slice so multitasking systems get a wack at the cpu.

void	DPMI_Yield(void) {
	if (sys.realDPMI && !opt.CPUhog) {
		union REGS regs;

		regs.w.ax=0x1680;
		int386(0x2f,&regs,&regs);
	};
}

ulong DPMI_MemFree(void) {
	union REGS regs;
	struct SREGS sregs;
	struct DPMIMemInfo dpmimi;

	regs.x.eax = 0x00000500;
	segread(&sregs);
	sregs.es = FP_SEG( &dpmimi );
	regs.x.edi = FP_OFF( &dpmimi );

	int386x(0x31, &regs, &regs, &sregs );

	return (dpmimi.NumFreePagesAvail<<12);
}

void *DPMIPhysMapAttempt(ulong addr,ulong len) {
	union REGS regs;
	regs.x.eax = 0x0800;
	regs.x.ebx = addr>>16;
	regs.x.ecx = addr;
	regs.x.esi = len>>16;
	regs.x.edi = len;
	int386(0x31,&regs,&regs);

	if (regs.x.cflag & INTR_CF) return NULL;
	else return (void *)(((ulong)regs.x.ebx<<16) | ((ulong)regs.x.ecx & 0xFFFF));
}



//////////////////////////// Hacked File I/O ///////////////////////////////

//static char VP_Start_Path[256];

static struct VPAssign *VP_Assign_List = NULL;
static int VP_Assign_Count = 0;

//	void	VP_Set_Start_Path(char *argv0);
//
//	Sets the path of the "VP:" prefix used in VP file I/O.

void	VP_Set_Start_Path(char *argv0) {
	char buf[256];
	char *t=buf;
	char *s=buf;

	while(*s++=*argv0++);
	--s;

	// for some reason Watcom/DOS4GW gives me mutilated garbage
	// like "C:\\Athena\Vgapaint" when I run VP386 over a Win95
	// peer-to-peer network
	//
	// Look for the "C:\\" garbage and cancel the "C:"

	if (isalpha(buf[0]) && buf[1]==':' && buf[2]=='\\' && buf[3]=='\\')
		t+=2;

	// step backwards through the string, looking for a '\' or a ':'

	while(s>t) {
		if (*s=='\\' || *s==':') break;
		--s;
	}

	if (s>t) ++s;
	*s=0;

	VPAssign("VP",t);
}

//	void VPAssign(char *name,char *path);
//
//	Associates a path with a name.
//
//	If you assign "Work:" to "C:\dumb", then "Work:project" refers to
//	"C:\dumb\project".
//
//	Unfortunately this also means that "Work:..\temp" refers to
//	"C:\dumb\..\temp".  Good thing DOS/Windows 95 allows this.
//
//	The strings supplied are duplicated and may be modified/destroyed
//	afterward.
//
//	Any existing assign with the same name is destroyed.
//
//	V1.3: Assigns in the path are now resolved.
//
//	Returns TRUE on success, FALSE on failure.

boolean	VPAssign(char *name,char *path) {
	struct VPAssign *vpa;
	char *s,*t,*m;
	char path2[512];

	VPAssignDestroy(name);

	if (!(vpa = malloc(sizeof(struct VPAssign)))) return FALSE;

	// obtain length of assign name string; slice off ending colon

	s = name;
	while(*s && *s!=':') ++s;

	// resolve assigns in the path

	path = VPAssignResolve(path2,path);

	// obtain length of path string; slice off ending slashes

	t = path2;
	while(*t) ++t;
	while(t>path2) if (t[-1]=='\\') --t; else break;

	// allocate one block of memory for both strings

	if (!(m = malloc((s-name)+(t-path2)+2))) {
		free(vpa);
		return FALSE;
	}

	vpa->name = m;
	vpa->path = m + (s-name) + 1;

	memcpy(vpa->name,name ,(s-name ));	vpa->name[s-name ]=0;
	memcpy(vpa->path,path2,(t-path2));	vpa->path[t-path2]=0;

	// link new assign into list

	vpa->prev = NULL;
	vpa->next = VP_Assign_List;
	VP_Assign_List = vpa;
	vpa->next->prev = VP_Assign_List;

	++VP_Assign_Count;

	return TRUE;
}

//	void	VPAssignDestroy(char *name);
//
//	Destroy an assign.

void	VPAssignDestroy(char *name) {
	struct VPAssign *vpa;

	if (!(vpa = VPAssignFind(name))) return;

	if (!vpa->prev) VP_Assign_List = vpa->next;
	if (vpa->next)	vpa->next->prev = vpa->prev;
	if (vpa->prev)	vpa->prev->next = vpa->next;

	free(vpa->name);
	free(vpa);

	--VP_Assign_Count;
}

//	struct VPAssign *VPAssignFind(char *name);
//
//	Find the address of a VPAssign structure with a certain name.

struct VPAssign *VPAssignFind(char *name) {
	struct VPAssign *vpa = VP_Assign_List;
	char n[128],*s,*t;

	s = name; t=n;
	while (*s && *s!=':') *t++=*s++;
	*t++=0;

	while(vpa) {
		if (!stricmp(vpa->name,n)) return vpa;
		vpa=vpa->next;
	}
	return NULL;
}


//	char *VPAssignResolve(char *outbuffer,char *name);
//
//	Check for an assign in a name ("VP:\vp386.ini") and resolve the
//	assign to a path if there is one.
//
//	Returns outbuffer for ease of use.
//
//	V1.1 fix: Don't cancel a pair of starting backslashes.  That way
//		  we don't disturb network references (\\Athena\vgapaint)

char *	VPAssignResolve(char *outbuffer,char *name) {
	struct VPAssign *vpa;
	char *t;

	if (vpa = VPAssignFind(name)) {
		strcpy(outbuffer,vpa->path);
		t=outbuffer; while(*t) ++t;
		*t++ = '\\';

		// skip past the colon

		while(*name && *name!=':') ++name;
		if (*name==':') ++name;

		// skip a starting backslash, but leave a double backslash
		// alone

		if (name[0]=='\\' && name[1]!='\\') ++name;

		// concatenate path

		strcpy(t,name);

	} else strcpy(outbuffer,name);

	return outbuffer;
}

//	struct VPAssign *VPAssignEnum(struct VPAssign *prev);
//
//	Enumerates the assign list.
//
//	A value of NULL for prev returns the next assign; otherwise VPAssignEnum
//	returns either a pointer to the next assign or NULL for the end.

struct VPAssign *VPAssignEnum(struct VPAssign *prev) {
	if (prev) return prev->next;
	else return VP_Assign_List;
}

int	VPAssignCount(void) {
	return VP_Assign_Count;
}

//	FILE *handle = VPfopen(char *name,char *usage);
//
//	Opens a file the same way as fopen(), but checking for "VP:" in front
//	of file name.  If "VP:" is found, the starting path to VP386 is added
//	in front of the name.
//
//	Uses VPopen() and fdopen() to take advantage of Windows 95's long filenames.
//	To do this, the undocumented Watcom function __open_flags() function is used.

extern int __open_flags(char *usage);

FILE *	VPfopen(char *name,char *usage) {
	uchar nambuf[384];
	FILE *t;
	int handle;
	int open_flags;
	int sopen_flags = 0;
	int create_flags = 0;

	name = VPAssignResolve(nambuf,name);

//	if (!strnicmp(name,"VP:",3)) {
//		strcpy(tmpbuf,VP_Start_Path);
//		if (name[3]=='\\')	++name;
//		strcat(tmpbuf,name+3);
//		name=tmpbuf;
//	}

	if (opt.useLFN) {
		open_flags = __open_flags(usage);

		if (tolower(*usage)=='r') {
			if (open_flags & _WRITE ) sopen_flags  =O_RDWR;
		} else {	// 'w' or 'a'
			sopen_flags = O_CREAT + O_WRONLY;
			if (open_flags & _READ  ) ++sopen_flags;		// promote from O_WRONLY to O_RDWR
			if (tolower(*usage)=='a') sopen_flags |=O_APPEND;	// 'a' --> append mode
			else			  sopen_flags |=O_TRUNC;	// 'w' --> truncate mode
			create_flags = S_IRUSR + S_IWUSR;
		};
		if (open_flags & _BINARY) sopen_flags |=O_BINARY;
		else			  sopen_flags |=O_TEXT;

		if (-1==(handle=VPsopen(name,sopen_flags,0,create_flags))) return NULL;

		if (t=fdopen(handle,usage)) {
			if (sopen_flags & O_APPEND) fseek(t,0,SEEK_END);
			return t;
		}
		close(handle);
		return NULL;
	} else
		return fopen(name,usage);	// no lfn support
}


//	ulong handle = VPfindfirst(char *spec,uchar attrib,struct VPFileFind *vpff);
//
//	Initiate a FileFind operation, using LFN filenames if operating under Win95.
//
//	You must call VPfindclose() after finishing the search, whether you have
//	reached the end or not.
//
//	If you are passed a DOS handle (handle 1) for any reason, you must call
//	VPfindnext() with the same VPFileFind structure; however, you need not call
//	VPfindclose().  It won't hurt, though.  DOS FileFind data will be returned
//	in the Win95 LFN structure but you may refer to the DOS structure.  Note that
//	the DOS FileFind structure is embedded in the <lfn> structure.

ulong VPfindfirst(char *spec,uchar attrib,struct VPFileFind *vpff) {
	struct RealInt regs;
	int rc;

	if (opt.useLFN) {
		strcpy(realtmp+0x1c0,spec);

		regs.w.ax = 0x714e;
		regs.w.si = 0x0000;
		regs.w.cx = attrib;
		regs.w.dx = ((ulong)realtmp+0x1c0)&15;
		regs.ds   = ((ulong)realtmp+0x1c0)>>4;
		regs.w.di = ((ulong)realtmp      )&15;
		regs.es   = ((ulong)realtmp      )>>4;
		DPMIRealInt(0x21,&regs);

		memcpy(vpff,realtmp,sizeof(struct VPFileFind));

		if (regs.eflags & INTR_CF) {
			if (regs.w.ax != 0x7100) return NULL;
		} else return 0x10000 + (ulong)regs.w.ax;

		// error 0x7100: fall through (function not supported)
	};

//	rc = _dos_findfirst(spec,attrib,(void *)vpff->std._reserved);
	rc = _dos_findfirst(spec,attrib,(void *)((char *)realtmp+128));

	// translate DOS -> Win95 filefind structure

	memset(vpff->std.__padding,0,128);
	vpff->lfn.attrib	= ((struct VPFileFind *)realtmp)->std.attrib;
	vpff->lfn.fsize_low	= ((struct VPFileFind *)realtmp)->std.attrib;
	memcpy(vpff->lfn.lfname,&((struct VPFileFind *)realtmp)->std.name,13);	// set lfname to DOS name
	memcpy(vpff->lfn.sfname,&((struct VPFileFind *)realtmp)->std.name,13);

	return !rc;  // 1 is our DOS filefind handle
}

//	int error = VPfindnext(ulong handle,struct VPFileFind *vpff);
//
//	Continue a FileFind operation, using LFN filenames if operating under Win95.
//
//	You need to call VPfindfirst() to start the file search and to obtain a handle.
//	VPfindclose() must be called to terminate the search whether the search is
//	followed to completion or not.
//
//	When using a DOS filefind handle (which you may be passed even in Win95) you MUST
//	return the same VPFileFind structure which was passed to the previous
//	VPfindnext()/VPfindfirst() on the same handle.

int	VPfindnext(ulong handle,struct VPFileFind *vpff) {
	struct RealInt regs;
	int rc;

	if (handle >= 0x10000) {	// win95 lfn
		regs.w.ax = 0x714f;
		regs.w.si = 0;
		regs.w.bx = (ushort)handle;
		regs.w.di = ((ulong)realtmp) & 15;
		regs.es   = ((ulong)realtmp)>>4;
		DPMIRealInt(0x21,&regs);

		memcpy(vpff,realtmp,sizeof(struct VPFileFind));

		if (regs.eflags & INTR_CF) {
			if (regs.w.ax == 0x7100) return VPfindnext(1,vpff);	// recall with DOS handle
			else return regs.w.ax;
		} else return 0;
	} else if (handle == 1) {	// dos
//		rc = _dos_findnext((void *)vpff->std._reserved);
		rc = _dos_findnext((void *)((char *)realtmp+128));

		// translate DOS -> Win95 filefind structure

		memset(vpff->std.__padding,0,128);
		vpff->lfn.attrib	= ((struct VPFileFind *)realtmp)->std.attrib;
		vpff->lfn.fsize_low	= ((struct VPFileFind *)realtmp)->std.attrib;
		memcpy(vpff->lfn.lfname,&((struct VPFileFind *)realtmp)->std.name,13);	// set lfname to DOS name
		memcpy(vpff->lfn.sfname,&((struct VPFileFind *)realtmp)->std.name,13);

		return rc;
	} else return 12;		// dos no files remaining
}

//	void VPfindclose(ulong handle);
//
//	Terminate a FileFind operation.
//
//	You must call this function to terminate a FileFind operation, esp. when passed
//	a Win95 function.  It is not required to call this function for a DOS function
//	but it couldn't hurt.

void VPfindclose(ulong handle) {
	struct RealInt regs;

	if (handle != 1) {
		regs.w.ax = 0x71a1;
		regs.w.bx = handle - 0x10000;
		DPMIRealInt(0x21,&regs);
	}
}


//	int handle = VPsopen(const char *path,int access,int share,...);
//
//	Open a file for access at the OS level, using LFNames if available.
//
//	Designed for function call compatibility with the standard C function sopen().

int	VPsopen(const char *path,int access,int share,...) {
	int create_flags;
	int handle;
	ulong	io_flags;
	struct RealInt regs;
	union REGS r;

	if (access & O_CREAT) {
		va_list val;
		va_start(val,share);
		create_flags = va_arg(val,int);
		va_end(val);
	};

	// pass options through to regular open()

	if (opt.useLFN) {
		regs.x.eax = 0x716c;
		regs.x.ebx = (access & 0x03) + (share & 0xf0);
		regs.x.ecx = 0;

		switch(access & (O_CREAT+O_TRUNC)) {
		case NULL:		regs.x.edx = 0x01;	break;
		case O_CREAT:		regs.x.edx = 0x10;	break;
		case O_TRUNC:		regs.x.edx = 0x02;	break;
		case O_CREAT+O_TRUNC:	regs.x.edx = 0x12;	break;
		}

		if (access&O_APPEND && !(regs.x.edx&3)) regs.x.edx|=1;

		strcpy(realtmp,path);
		regs.x.esi = ((ulong)realtmp)&15;
		regs.ds = ((ulong)realtmp)>>4;
		DPMIRealInt(0x21,&regs);

		if (regs.eflags & INTR_CF) {
			if (regs.w.ax != 0x7100) return -1;

			// function not supported (IFSMgr not running) - fall through to old function
		} else {

			handle = (ushort)regs.w.ax;

			io_flags = 0;

			if (access & O_RDWR) io_flags |= _READ | _WRITE;
			else {
				if (access & O_WRONLY) io_flags |= _WRITE;
				else io_flags |= _READ;
			}
			if (access & O_APPEND) io_flags |= _APPEND;
			if (access & O_BINARY) io_flags |= _BINARY;

			// change file mode

			if (isatty(handle)) {
				io_flags |= _ISTTY;
				if (access & O_BINARY) {
					r.x.eax = 0x4400;
					r.x.ebx = handle;
					int386(0x21,&r,&r);

					if (r.x.cflag & INTR_CF) { close(handle); return -1; }

					r.h.dl |= 0x20L;
					r.x.edx &= 0x000000ff;

					r.x.eax = 0x4401;
					r.x.ebx = handle;
					int386(0x21,&r,&r);

					if (r.x.cflag & INTR_CF) { close(handle); return -1; }
				}
			}

			__SetIOMode(handle,io_flags);

			return handle;
		}
	}
	if (access & O_CREAT)
		return sopen(path,access,share,create_flags);
	else
		return sopen(path,access,share);
}

//	int filesize = VPfilesize(FILE *f);
//
//	Returns the length of an opened file.
//
//	Returns -1 upon failure.

int	VPfilesize(FILE *f) {
	int s;
	if (fseek(f,0,SEEK_END)) return -1L;
	if ((s=ftell(f))<0) return -1L;
	if (fseek(f,0,SEEK_SET)) return -1L;

	return s;
}


/////////////////////// Function calling system /////////////////////////

//	I'll be frank: DOS/4GW sucks.  That's why I have to work
//	through int 12h in order to make pass-ups work.  Since this
//	is the equipment information vector, it shouldn't be called
//	too much and definitely never from interrupt.
//
//	Note: This means multiple copies of VP386 will take over int12h
//	      multiple times.  This shouldn't impact the system too much,
//	      but other DOS/4GW programs may do the same thing.  Other
//	      programs tend to use other vectors (one popular game uses
//	      the Ctrl-Break vector).

static char VPpassup_routine[]={
	0x00,				// passup flag
	0x2e,0x80,0x0e,0x00,0x00,0x01,	// or cs:byte ptr [0],1
	0xcd,0x12,			// int 12h
	0xcb,				// retf
};

static boolean	VPCallbackSystemInstalled=FALSE;

//	To do a raw passup, place the function address in EBP and Far call
//	the routine.


//	boolean failed=InitCallbackSystem(void);
//
//	Initialize the callback system.
//
//	Returns FALSE if the Callback system was unable to initialize.

boolean	InitCallbackSystem(void) {
	// copy callback routine to real memory

	memcpy(realmem,VPpassup_routine,sizeof VPpassup_routine);

	// set the int12h vector

	prev_int_12=_dos_getvect(0x12);
	_dos_setvect(0x12,CBS_IntRoutine);

	VPCallbackSystemInstalled=TRUE;

	return TRUE;
}


//	void DeinitCallbackSystem(void)
//
//	Shutdown the callback system.

void	DeinitCallbackSystem(void) {
	if (!VPCallbackSystemInstalled) return;
	VPCallbackSystemInstalled=FALSE;
	_dos_setvect(0x12,prev_int_12);
}

//	ushort cb_seg = AllocCallback(VPpassup_fptr function,boolean interrupt);
//
//	Setup a use.
//
//	Returns NULL if unable to allocate a callback for any reason;
//	returns a real-mode segment value from 0x0001-0xFFFF
//	if allocate was successful.
//
//	To use the passup, far call the routine.  If you are going to use
//	the callback for hooking a non-passed up interrupt vector, you need
//	to set the <interrupt> flag, and then place <segment>:0000 into the
//	vector.

uchar	VPcallback_routine[32]={
	0x66,0x55,			// push ebp
	0x66,0xbd,0x00,0x00,0x00,0x00,	// mov ebp,00000000h
	0x9a,0x01,0x00,0x00,0x00,	// call far 0000:0001h
	0x66,0x5d,			// pop ebp
	0xcf,				// iret

	0x66,0xbd,0x00,0x00,0x00,0x00,	// mov ebp,00000000h
	0xea,0x01,0x00,0x00,0x00,	// jmp far 0000:0001h
	0x00,0x00,0x00,0x00,0x00,	// pad bytes >16
};

ushort	AllocCallback(VPpassup_fptr vpf,boolean interrpt) {
	struct RealInt RMI;
	uchar *p;

	RMI.b.ah=0x48;
	RMI.w.bx=1;
	DPMIRealInt(0x21,&RMI);
	if (RMI.eflags & INTR_CF)
		return NULL;

	p=(uchar *)(((long)RMI.w.ax)<<4);

	memcpy(p,VPcallback_routine+(interrpt?0:16),16);

	if (interrpt) {
		*(VPpassup_fptr *)(p+4) = vpf;
		*(ushort *)(p+11) = ((long)realmem)>>4;
	} else {
		*(VPpassup_fptr *)(p+2) = vpf;
		*(ushort *)(p+9) = ((long)realmem)>>4;
	}

	return RMI.w.ax;
}


//	void	FreeCallback(ushort id);
//
//	Frees a VP callback previously allocated with the AllocCallback()
//	function.
//
//	It is safe to pass NULL to this function.
//
//	Do not attempt to free a callback previously allocated, since it may
//	have been reallocated.


void	FreeCallback(ushort callback) {
	struct RealInt RMI;

	if (!callback) return;

	RMI.b.ah=0x49;
	RMI.es=callback;
	DPMIRealInt(0x21,&RMI);
}


//	struct CallList *call_list = FindCallList(ulong id);
//
//	Find a call list in the global call list.

struct CallList *FindCallList(ulong id) {
	struct Node *n,*next;

	n=GlobCallList.prev;

	while(next=n->prev) {
		if (((struct CallList *)n)->id == id) return (struct CallList *)n;
		n=next;
	}

	return NULL;
}

//	void	AddCallList(struct CallList *cl,ulong id);
//
//	Add a Call List to the global Call List.

void	AddCallList(struct CallList *cl,ulong id) {
	cl->id = id;
	disable();
	AddHead(&GlobCallList,&cl->node);
	enable();
}


//	void	RemCallList(struct CallList *cl);
//
//	Remove a Call List from the global Call List.

void	RemCallList(struct CallList *cl) {
	disable();
	Remove(&cl->node);
	enable();
}


//	void	AddCall(struct CallList *cl,struct Call *c);
//
//	Add a Call to a Call List.


void	AddCall(struct CallList *cl,struct Call *c) {
	disable();
	AddHead(&cl->list,&c->node);
	enable();
}


//	void	RemCall(struct Call *c);
//
//	Remove a Call from a Call List.

void	RemCall(struct Call *c) {
	disable();
	Remove(&c->node);
	enable();
}

//	ulong	ExecuteCallList(struct CallList *cl, ...);
//
//	Executes all functions on a call list.
//
//	Returns NULL, or a non-zero value if one of the routines
//	returned so.  Returning a non-zero value terminates
//	execution of the list.
//
//	Functions are called with the <data> argument in their
//	Call structure and a pointer to the varargs list passed
//	to this function.

long	ExecuteCallList(struct CallList *cl, ...) {
	struct Node *n,*next;
	va_list	val;
	long	retval;

	va_start(val,cl);
	n=cl->list.prev;
	while(next=n->prev) {
		retval=((struct Call *)n)->func(((struct Call *)n)->data,val);
		if (retval) return retval;
		n=next;
	}
	va_end(val);

	return 0;
}


///////////////////////////////// interrupt handling

// Vectors 08h-20h and 22-2eh are redirected up by DOS/4GW, so we have
// special handling for those vectors, for faster interrupt processing.
// We handle vector 05h and 2fh-ffh through our routines.

struct CallList		IntCallList[17];		// high ints not supported yet
static ulong		IntCallers[17]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static ulong		IntHandlerInstalled=0x00;

void (__interrupt __far *IntOldHandlers[17])(void);	// old interrupt vectors
void (*IntVectors[17])(void) = {
	&asm_int08_handler,
	&asm_int09_handler,
	&asm_int0a_handler,
	&asm_int0b_handler,
	&asm_int0c_handler,
	&asm_int0d_handler,
	&asm_int0e_handler,
	&asm_int0f_handler,
};

//	void	InitInterruptHandler(void);
//
//	Initializes the interrupt system for its first use.

boolean	InitInterruptHandler(void) {
	int i;

	for(i=0;i<17;i++) NewList(&IntCallList[i].list);

	memset(IntStackPtrs,0,sizeof IntStackPtrs);

	if (!(IntStackPtrs[0]=(ulong *)malloc(8192))) {
		DeinitInterruptHandler();
		return TRUE;
	}
	IntStacks[0]=IntStackPtrs[0]+2048;

	if (!(IntStackPtrs[16]=(ulong *)malloc(8192))) {
		DeinitInterruptHandler();
		return TRUE;
	}
	IntStacks[16]=IntStackPtrs[16]+2048;

//	for(i=1;i<16;i++)
//		if (!(IntStacks[i]=(ulong *)malloc(2048))) {
//			DeinitInterruptHandler();
//			return TRUE;
// 		} else IntStackPtrs[i]=IntStacks[i]+512;

	return FALSE;
}

//	void	DeinitInterruptHandler(void);
//
//	Deinitializes the interrupt system.

void	DeinitInterruptHandler(void) {
	int i;

	DeinstallAllInterruptHandlers();
	for(i=0;i<17;i++) if (IntStackPtrs[i]) { free(IntStackPtrs[i]); IntStackPtrs[i]=0; };
}

//	void	AddInterruptHandler(uchar intnum, struct IHandler *ih);
//
//	Adds an interrupt handler to a particular interrupt chain.

void	AddInterruptHandler(uchar intnum,struct IHandler *ih) {
	AddCall(&IntCallList[intnum],&ih->call);
	if (!IntCallers[intnum]++)
		InstallInterruptHandler(intnum);
}

//	void	RemInterruptHandler(uchar intnum, struct IHandler *ih);
//
//	Removes an interrupt handler from a particular interrupt chain.

void	RemInterruptHandler(uchar intnum, struct IHandler *ih) {
	if (!--IntCallers[intnum])
		DeinstallInterruptHandler(intnum);
	RemCall(&ih->call);
}

//	void	LockInterruptHandler(uchar intnum);
//
//	Lock the given interrupt handler # so that the handler isn't deinstalled.

void	LockInterruptHandler(uchar i) {
	if (!IntCallers[i]) InstallInterruptHandler(i);
	++IntCallers[i];
}

//	void	UnlockInterruptHandler(uchar intnum);
//
//	Unlock the given interrupt handler # so that the handler can be deinstalled.

void	UnlockInterruptHandler(uchar i) {
	--IntCallers[i];
	if (!IntCallers[i]) DeinstallInterruptHandler(i);
}

//	void	InstallInterruptHandler(uchar intnum);
//
//	Install the internal interrupt handler # specified.

void	__far __interrupt dummy(void) {
	_chain_intr(IntOldHandlers[0]);
}

void	InstallInterruptHandler(uchar intnum) {
	if (IntHandlerInstalled & (1<<intnum)) return;

	disable();

	IntOldHandlers[intnum] = _dos_getvect(intnum + 0x08);

	_dos_setvect(intnum + 0x08, (void (__interrupt __far *)(void))IntVectors[intnum]);

	IntHandlerInstalled |= (1L<<intnum);

	if (intnum == INT_TIMER)
		if (Int08TimerPeriod != 0x10000)
			_InternalSetTimerPeriod(Int08TimerPeriod);

	enable();
}

//	void	DeinstallInterruptHandler(char intnum);
//
//	Deinstall the internal interrupt handler # specified.

void	DeinstallInterruptHandler(uchar intnum) {
	if (!(IntHandlerInstalled & (1<<intnum))) return;

	disable();

	_dos_setvect(intnum+0x08,IntOldHandlers[intnum]);

	IntHandlerInstalled &= ~(1L<<intnum);

	if (intnum == INT_TIMER)
		if (Int08TimerPeriod != 0x10000)
			_InternalSetTimerPeriod(0);

	enable();
}

//	void	DeinstallAllInterruptHandlers(void);
//
//	Deinstall ALL internal interrupt handlers

void	DeinstallAllInterruptHandlers(void) {
	uchar i;

	for(i=0;i<16;i++) DeinstallInterruptHandler(i);
}

//	void	SetTimerPeriod(ushort period);
//
//	Sets the period of 8253 Timer 0.
//
//	All values of <period> are valid.  A period of 1 is 1193181 Hz.  A period of 10
//	is 119318.1 Hz.  A period of 100 is 11931.81 Hz, etc.  The exception is that a
//	period of 0 is 1193181/65536 Hz, or 18.2 Hz - this is the normal period for the
//	timer.  This 0 period MUST be set before you exit, or the clock will run at
//	Warp 9.
//
//	Periods above about 2000Hz should not be used, as DOS/4GW and/or Windows aren't
//	able to task switch fast enough.  Values up to about 500Hz are usually safe.

void	SetTimerPeriod(ushort period) {
	disable();
	Int08TimerPeriod = period;
	if (!period) Int08TimerPeriod=0x10000;
	if (IntHandlerInstalled & 1) _InternalSetTimerPeriod(period);
	enable();
}

void	_InternalSetTimerPeriod(ushort period) {
	outportb(TIMER_CTRL,TIMER_COUNTER0+TIMER_LSBMSB+TIMER_SQWAVE+
		TIMER_BINARY);
	outportb(TIMER_0,(uchar)period);
	outportb(TIMER_0,period>>8);
}

