#ifndef	VGAPAINT_LISTBOX_H
#define	VGAPAINT_LISTBOX_H

#include <a/types.h>
#include <a/bits.h>

#include "video.h"
#include "list.h"
#include "sprite.h"
#include "tools.h"

#define	LB_END	(-0x8000)

struct Listbox {
	long	x,y,width,height,cwidth,cheight;
	ulong	flags;
	long	cpos;
	long	alen;
	long	sel;
	void	*seladd;
	void *(*getIndex)(struct Listbox *,int);
	void (*displayIndex)(struct Listbox *,void *,int,int);
	void (*clearIndex)(struct Listbox *,int);
	void (*scrollIndex)(struct Listbox *,int);
	void (*clearWindow)(struct Listbox *);
	struct BitMap *destbm;
	long textheight;	/* item height */
	void *listadd;		/* item list */
};

#define	FLD_NUMERIC	(0x00000001L)
#define	FLD_SYSFONT	(0x00000002L)
#define	FLD_AUTOSIZE	(0x00000004L)

// autosize:	cwidth = width/sysfontwidth
//		height = sysfontheight

#define	FLD_FLASHTIME	(1)

struct Field {
	long x,y,width,height;
	struct BitMap *destbm;
	char	*buf;
	int	flen;
	int	cpos;
	int	buflen;
	int	cwidth;
	color	fore,back;
	ulong	flags;
	long	low;
	long	high;
	long	lval;
};

typedef	void	(*Sld_Upd)(struct Slider *);

#define	SLD_HITID	(1)

/* REQSLD_ONEHIT: Wait for release on knock hits */
#define	REQSLD_ONEHIT	(0x00000001L)

/* SLD_ARROWxHIT: Wait for release on arrow hits */
#define	SLD_ARROW1HIT	(0x00000002L)
#define	SLD_ARROW2HIT	(0x00000004L)

/* SLD_LISTBOX: Allow this slider to control a listbox */
#define	SLD_LISTBOX	(0x00000008L)

/* SLD_ARROWx: Rendering direction for arrow x */
#define	SLD_ARROW1	(0x00000010L)
#define	SLD_ARROW1MASK	(0x000000F0L)
#define	SLD_ARROW2	(0x00000100L)
#define	SLD_ARROW2MASK	(0x00000F00L)

/* SLD_OUTLINE3D: Render 3d motif around bounding box */
#define	SLD_OUTLINE3D	(0x00001000L)
#define	SLD_ATTACHX	(0x00002000L)
#define	SLD_ATTACHY	(0x00004000L)

struct Slider {
	struct	Slider	*next;		/* for the requesters */
	struct	Sprite	knobspr;
	struct	BitMap	*destbm;
	struct	Rectangle	boundbox;
	uchar	*knob;		/* can be NULL for auto-knob */
	long	knob_width;	/* knob width */
	long	knob_height;	/* knob height */
	long	min_width;	/* autoknob minimum width -or- x knob handle */
	long	min_height;	/* autoknob minimum height -or- y knob handle */
	long	xpixspc;	/* horizontal pixel space for movement */
	long	ypixspc;	/* vertical pixel space for movement */
	long	xclicks;	/* horizontal clicks */
	long	yclicks;	/* vertical clicks */
	long	xbump;		/* click bump left/right */
	long	ybump;		/* click bump up/down */
	long	pposx;		/* x pixel position */
	long	pposy;		/* y pixel position */
	long	cposx;		/* x click position */
	long	cposy;		/* y click position */
	long	flags;		/* flags longword */
	Sld_Upd	updater;	/* number updater routine */
	struct	Rectangle	arrow1;		/* 1st built-in arrow */
	struct	Rectangle	arrow2;		/* 2nd built-in arrow */
	long	arrow1tapx;	/* taps for the arrows */
	long	arrow1tapy;
	long	arrow2tapx;
	long	arrow2tapy;
	struct	Listbox		*lboxctrl;	/* for controlling a listbox */
};

struct TextboxNode {
	struct Node node;
	char	*string;
};

struct Textbox {
	struct Listbox *lb;
	struct List text;
	char **table;
};

boolean	SliderInit(struct Slider *sld);
void	SliderRedraw(struct Slider *sld);
void	SliderDrawArrow(struct Slider *sld,struct Rectangle *r,short dir);
void	SliderRecompute(struct Slider *sld);
void	SliderDraw(struct Slider *sld);
void	SliderMove(struct Slider *sld,long x,long y);
void	SliderAbs(struct Slider *sld,long x,long y);
void	SliderDrag(struct Slider *sld);
void	SliderClose(struct Slider *sld);
ulong	SliderHandle(struct Slider *sld,long x,long y);

///////// Field Functions //////////

void NumIntoField(struct Field *fld,long num);
void DisplayField(struct Field *fld);
void DisplayFieldCursor(struct Field *fld,coord x);
void ActivateField(struct Field *fld,coord x);

///////// Listbox Functions //////////

void	DrawListbox(struct Listbox *);
void	DisplayListbox(struct Listbox *,int,int,int);
void	SelectListbox(struct Listbox *,int);
void	ScrollListbox(struct Listbox *,int);

/******** Textbox functions ********/

void	InitTextbox(struct Textbox *,struct Listbox *);
boolean	TextboxAdd(struct Textbox *,char *);
boolean	TextboxPrintf(struct Textbox *,char *, ...);
char **	TextboxGlue(struct Textbox *);
void	TextboxDissolve(struct Textbox *);
void	TextboxCreate(struct Textbox *,boolean sysfont);
void *	TextboxGetIndex(struct Listbox *lb,int index);
void	TextboxDisplayIndex(struct Listbox *lb,void *entry,int dindex,int highlight);
void	TextboxDisplayIndex2(struct Listbox *lb,void *entry,int dindex,int highlight);
void	TextboxClearIndex(struct Listbox *lb,int dindex);
void	TextboxScrollIndex(struct Listbox *lb,int dy);
void	TextboxClearWindow(struct Listbox *lb);

#endif
