//////////////////////////////////////////////////////////////
//
//	VGAPaint 386 - PCX.c
//
//	Routines for reading and writing picture files, including
//	PCX format files.
//
//	Lossy-LZ compression is in lossylz.c and Anim compression
//	is in animio.c.
//

#define	VGAPAINT_PCX_C

#include <a/bits.h>

#include "vgapaint.h"
#include "average.h"
#include "pcx.h"
#include "picture.h"
#include "io.h"

/////////////////////////////////////

// from pcxcom.asm

extern uchar com_initialize[],com_initialize2[];
extern uchar com_pdecomp_none[],com_pdecomp_6bit[];
extern uchar com_ddecomp_none[],com_ddecomp_pcx[];
extern uchar com_termin_key[];
extern uchar com_cleanup[];

extern ushort com_palette_offset,com_picture_len;
extern ushort com_palette_len;

/////////////////////////////////////

struct Color {
	uchar red,grn,blu;
};

struct PCXHeader {
	uchar			magic;		// 00
	uchar			version;	// 01
	char			encoding;	// 02
	char			bitsPerPixel;	// 03
	short			x1,y1,x2,y2;	// 04
	short			xdpi,ydpi;	// 0c
	struct	Color		palette[16];	// 10
	uchar			videomode; /* reserved */
	char			planes;
	short			bytesPerRow;
	short			palinfo;
	short			xsize;
	short			ysize;
	char			filler[54];
};
/* WATCOM C needs to be ANSI C compliant!!!!
#if sizeof(struct PCXHeader)!=128
#error Error in size of PCX header
#endif
*/
#define	PCXH_MAGIC	(0x0A)
#define	PCXH_RLL	(0x01)
#define	PCXH_PALCOLOR	(0x00)
#define PCXH_PALBW	(0x01)
#define	PCXH_PALGREY	(0x02)

enum PictureIOErrors PCXWrite(char *name,struct BitMap *bm,uchar *pal) {
	FILE *file;
	struct PCXHeader pcxh;
	long y;
	uchar *vmem=bm->vm;
	uchar *pckmem;	/* packing memory */

	if (bm->width>384) {
		if (!(pckmem=(uchar *)malloc(bm->width<<1))) {
			return PIOERR_MEMORY;
		};
	}	else	pckmem=packmem;

	syncbitmap();

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"wb"))) return PIOERR_OPEN;

	memset(&pcxh,0,sizeof pcxh);
	pcxh.magic=		PCXH_MAGIC;
	pcxh.version=		5;
	pcxh.encoding=		PCXH_RLL;
	pcxh.bitsPerPixel=	8;
	pcxh.x2=		bm->width-1;
	pcxh.y2=		bm->height-1;
	pcxh.xdpi=		bm->xdpi;	//80;
	pcxh.ydpi=		bm->ydpi;	//72;
	pcxh.videomode=		0x13;
	pcxh.planes=		1;
//	pcxh.bytesPerRow=	(bm->width+1)&0xfffffffe;
	pcxh.bytesPerRow=	bm->width;
	pcxh.palinfo=		PCXH_PALCOLOR;
	pcxh.xsize=		bm->width;
	pcxh.ysize=		bm->height;

	fwrite(&pcxh,sizeof pcxh,1,file);

	for(y=0;y<bm->height;y++) {
//		SetTitle("Packing row %d...",y);
		if (1!=fwrite(pckmem,pack_PCX_row(pckmem,vmem,bm->width),1,file)) {
			fclose(file);
			return	PIOERR_WRITE;
		}
		vmem+=bm->bpr;
	};

	fputc(0x0c,file);
	if (1!=fwrite(pal,768,1,file)) {
		fclose(file);
		return PIOERR_WRITE;
	}

	fclose(file);

	if (pckmem!=packmem) free(pckmem);

//	mouse_hide();
//	repick_colors();
//	restore_screen();
//	mouse_show();

	PictureSetName(pict_cur,name);

	return PIOERR_NONE;
}

enum PictureIOErrors PCXWrite4(char *name,struct BitMap *bm) {
	FILE *file;
	struct PCXHeader pcxh;
	long y;
	uchar p;
	uchar *vmem=bm->vm;
	uchar *pckmem;	/* packing memory */
	uchar *cmpmem;

	if (bm->width>2048) {
		if (!(pckmem=(uchar *)malloc(bm->width>>1))) {
			return PIOERR_MEMORY;
		};
	}	else	pckmem=packmem;

	cmpmem = pckmem + (((bm->width+15)>>3)&0xFFFFFFFE);

	syncbitmap();

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"wb"))) return PIOERR_OPEN;

	memset(&pcxh,0,sizeof pcxh);
	memcpy(pcxh.palette,bm->pal,48);
	pcxh.magic=		PCXH_MAGIC;
	pcxh.version=		5;
	pcxh.encoding=		PCXH_RLL;
	pcxh.bitsPerPixel=	1;
	pcxh.x2=		bm->width-1;
	pcxh.y2=		bm->height-1;
	pcxh.xdpi=		bm->xdpi;	//80;
	pcxh.ydpi=		bm->ydpi;	//80;
	pcxh.videomode=		0x12;
	pcxh.planes=		4;
	pcxh.bytesPerRow=	((bm->width+15)>>3)&0xFFFFFFFE;
	pcxh.palinfo=		PCXH_PALCOLOR;
	pcxh.xsize=		bm->width;
	pcxh.ysize=		bm->height;

	fwrite(&pcxh,sizeof pcxh,1,file);

	for(y=0;y<bm->height;y++) {
		SetTitle("Packing row %d...",y);
		for(p=1;p<16;p<<=1) {
			ChunkyToPlanar(pckmem,vmem,pcxh.bytesPerRow,p);
			if (1!=fwrite(cmpmem,pack_PCX_row(cmpmem,pckmem,pcxh.bytesPerRow),1,file)) {
				fclose(file);
				return	PIOERR_WRITE;
			}
		}
		vmem+=bm->bpr;
	};

	fclose(file);

	if (pckmem!=packmem) free(pckmem);

	mouse_hide();
	repick_colors();
	restore_screen();
	mouse_show();

	PictureSetName(pict_cur,name);

	return PIOERR_NONE;
}

enum PictureIOErrors PCXWrite24(char *name,struct BitMap *bm,uchar *red,uchar *grn,uchar *blu) {
	FILE *file;
	struct PCXHeader pcxh;
	long y;
//	uchar *vmem=bm->vm;
	ulong	voff=0;
	uchar *pckmem;	/* packing memory */

	if (bm->width>384) {
		if (!(pckmem=(uchar *)malloc(bm->width<<1))) {
			return PIOERR_MEMORY;
		};
	}	else	pckmem=packmem;

	syncbitmap();

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"wb"))) return PIOERR_OPEN;

	memset(&pcxh,0,sizeof pcxh);
	pcxh.magic=		PCXH_MAGIC;
	pcxh.version=		5;
	pcxh.encoding=		PCXH_RLL;
	pcxh.bitsPerPixel=	8;
	pcxh.x2=		bm->width-1;
	pcxh.y2=		bm->height-1;
	pcxh.xdpi=		80;
	pcxh.ydpi=		72;
	pcxh.videomode=		0x13;
	pcxh.planes=		3;
	pcxh.bytesPerRow=	bm->bpr;
	pcxh.palinfo=		PCXH_PALCOLOR;
	pcxh.xsize=		bm->width;
	pcxh.ysize=		bm->height;

	fwrite(&pcxh,sizeof pcxh,1,file);

	for(y=0;y<bm->height;y++) {
		SetTitle("Packing row %d...",y);
		if (1!=fwrite(pckmem,pack_PCX_row24(pckmem,red+voff,bm->width),1,file)) {
			fclose(file);
			return	PIOERR_WRITE;
		}
		if (1!=fwrite(pckmem,pack_PCX_row24(pckmem,grn+voff,bm->width),1,file)) {
			fclose(file);
			return	PIOERR_WRITE;
		}
		if (1!=fwrite(pckmem,pack_PCX_row24(pckmem,blu+voff,bm->width),1,file)) {
			fclose(file);
			return	PIOERR_WRITE;
		}
		voff+=bm->bpr;
	};

	fputc(0x0c,file);
	if (1!=fwrite(bm->pal,768,1,file)) {
		fclose(file);
		return PIOERR_WRITE;
	}

	fclose(file);

	if (pckmem!=packmem) free(pckmem);

	mouse_hide();
	repick_colors();
	restore_screen();
	mouse_show();

	PictureSetName(pict_cur,name);

	return PIOERR_NONE;
}

/////////////////////////////////////

static uchar sfx_routine1[]={
	0xe8,0,0,0,0,			// 00 call $+5
	0x58,				// 05 pop eax
	0x83,0xe8,0x15,			// 06 sub eax,15h
	0x60,				// 09 pushad
	0x8d,0xb0,0,0,0,0,		// 0a lea esi,[eax+xxxxxxxxh]
	0x8d,0xb8,0,0,0,0,		// 10 lea edi,[eax+xxxxxxxxh]
	0xb9,0,0,0,0,			// 16 mov ecx,xxxxxxxxh
	0xfd,				// 1b std
	0xf3,0xa4,			// 1c rep movsb
	0xfc,				// 1e cld

	0x8d,0x77,0x01,			// 1f lea esi,[edi+1]
	0x8b,0xf8,			// 22 mov edi,eax

	0xe9,0,0,0,0,			// 24 jmp xxxxxxxxh
					// 29
};

static uchar sfx_routine2[]={
	0x8d,0x9f,0,0,0,0,		// lea ebx,[edi+xxxxxxxxh]
	0xac,				// $2:lodsb
	0x3c,0xc0,			// cmp al,0c0h
	0xb1,0x01,			// mov cl,1
	0x72,0x05,			// jb $1
	0x2c,0xc0,			// sub al,0c0h
	0x8a,0xc8,			// mov cl,al
	0xac,				// lodsb
	0xf3,0xaa,			// $1: rep stosb
	0x3b,0xfb,			// cmp edi,ebx
	0x72,0xee,			// jb $2
	0x61,				// popad
	0xc3				// ret
};

enum PictureIOErrors PCXWriteSFX(char *name,struct BitMap *bm) {
	FILE *file;
	struct PCXHeader pcxh;
	long y;
	uchar *vmem=bm->vm;
	uchar *pckmem;	/* packing memory */
	ulong	imagesize = 0;
	long	s;

	if (bm->width>384) {
		if (!(pckmem=(uchar *)malloc(bm->width<<1))) {
			return PIOERR_MEMORY;
		};
	}	else	pckmem=packmem;

	syncbitmap();

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"wb"))) return PIOERR_OPEN;

	memset(&pcxh,0,sizeof pcxh);
	pcxh.magic=		0xeb;
	pcxh.version=		0x0e;
	pcxh.encoding=		PCXH_RLL;
	pcxh.bitsPerPixel=	8;
	pcxh.x2=		bm->width-1;
	pcxh.y2=		bm->height-1;
	pcxh.xdpi=		bm->xdpi;	//80
	pcxh.ydpi=		bm->ydpi;	//72
	pcxh.videomode=		sizeof sfx_routine2;	// heh heh heh
	pcxh.planes=		1;
	pcxh.bytesPerRow=	bm->width;
	pcxh.palinfo=		PCXH_PALCOLOR;
	pcxh.xsize=		bm->width;
	pcxh.ysize=		bm->height;

	fwrite(&pcxh,sizeof pcxh,1,file);	// just take up space... we need to backpatch

	for(y=0;y<bm->height;y++) {
		if (!(y&8)) SetTitle("Packing row %d...",y);

		if (1!=fwrite(pckmem,s = pack_PCX_row(pckmem,vmem,bm->width),1,file))
			goto write_error;

		imagesize += s;
		vmem+=bm->bpr;
	};

	// embed the inline routine in here
	*(ulong *)(sfx_routine2+0x02) = bm->width*bm->height;		// size of uncompressed
	if (1!=fwrite(sfx_routine2,sizeof sfx_routine2,1,file)) goto write_error;

	// write extended palette
	fputc(0x0c,file);
	if (1!=fwrite(bm->pal,768,1,file)) goto write_error;

	// prepare the sfx routine for entry
	// the memory buffer should be the bitmapsize+1024 for safety (sfx2+decompression)

	*(ulong *)(sfx_routine1+0x0c) = imagesize+sizeof sfx_routine2+128-1;		// offset from "pop eax" to end of header - 1 (rep movsb down)
	*(ulong *)(sfx_routine1+0x12) = bm->width*bm->height+1024-1;	// end of buffer - 1 (rep movsb down)
	*(ulong *)(sfx_routine1+0x17) = imagesize+sizeof sfx_routine2;				// size of image
	*(ulong *)(sfx_routine1+0x25) = bm->width*bm->height+1024-sizeof sfx_routine2-0x39;

	// backtrack and write the sfx routine
	// the first part of the sfx routine is buried in the 16-color palette area

	memcpy((uchar *)pcxh.palette,sfx_routine1,sizeof sfx_routine1);
	if (fseek(file,0,SEEK_SET)) goto write_error;
	if (1!=fwrite(&pcxh,sizeof pcxh,1,file)) goto write_error;

	fclose(file);

	if (pckmem!=packmem) free(pckmem);

	mouse_hide();
	repick_colors();
	restore_screen();
	mouse_show();

	PictureSetName(pict_cur,name);

	return PIOERR_NONE;

write_error:
	fclose(file);
	return PIOERR_WRITE;
}

/////////////////////

long pack_PCX_row(uchar *dst,uchar *src,long left) {
	uchar data,new,count=0;
	uchar *start=dst;

	data=*src;

	while(left--) {
		new=*src++;
		if (count==63 || (new!=data && count)) {
			if (count>=2 || data>=0xc0)
				*dst++=count+0xc0;
			*dst++=data;
			data=new;
			count=1;
		} else data=new,++count;
	};
	if (count) {
		if (count>=2 || data>=0xc0) *dst++=count+0xc0;
		*dst++=data;
	};
	return dst-start;
}

long pack_PCXcom_row(uchar *dst,uchar *src,long left,ulong *codeptr) {
	uchar data,new,count=0;
	uchar *start=dst;
	ulong codes=0;

	data=*src;

	while(left--) {
		new=*src++;
		if (count==63 || (new!=data && count)) {
			if (count>=2 || data>=0xc0)
				*dst++=count+0xc0;
			*dst++=data;
			data=new;
			count=1;
			++codes;
		} else data=new,++count;
	};
	if (count) {
		if (count>=2 || data>=0xc0) *dst++=count+0xc0;
		*dst++=data;
		++codes;
	};
	*codeptr += codes;
	return dst-start;
}

long pack_PCX_row24(uchar *dst,uchar *src,long left) {
	uchar data,new,count=0;
	uchar *start=dst;

	data=*src<<2;

	while(left--) {
		new=*src++<<2;
		if (count==63 || (new!=data && count)) {
			if (count>=2 || data>=0xc0)
				*dst++=count+0xc0;
			*dst++=data;
			data=new;
			count=1;
		} else data=new,++count;
	};
	if (count) {
		if (count>=2 || data>=0xc0) *dst++=count+0xc0;
		*dst++=data;
	};
	return dst-start;
}

int	PCXCom_Write_Chunk(char *c,FILE *f) {
	return fwrite(c,*(ulong *)&c[-4],1,f);
}

#define	PCXC_SIZ(x)	(*(ulong*)&(x)[-4])

enum PictureIOErrors PCXCom_Write(char *name,struct BitMap *bm) {
	FILE *file;
	long y;
	long stmp=0;
	uchar *vmem=picmem;
	uchar *pic_decompressor;
	uchar *pal_decompressor;
	ulong codes=0;

	syncbitmap();

	if (picbm->width!=320 || picbm->height!=200) {
		return PIOERR_BADSIZE;
	};

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"wb"))) return PIOERR_OPEN;


	vmem=picmem;
	for(y=0;y<200;y++) {
		if (!(y & 7)) SetTitle("Testing compressability: %d",y);
		stmp+=pack_PCXcom_row(packmem,vmem,320,&codes);
		vmem+=320;
	};

	if (stmp>=64000) {
		pic_decompressor = com_ddecomp_none;
		codes = 32000;
	} else {
		pic_decompressor = com_ddecomp_pcx;
	};

	/////////////////

	if (com_palette_len > (PCXC_SIZ(com_pdecomp_6bit)<<2)) {
		pal_decompressor = com_pdecomp_6bit;
		com_palette_len = 192;
	} else {
		pal_decompressor = com_pdecomp_none;
		com_palette_len = 768;
	}

	com_palette_offset = 	0x0100
				+ PCXC_SIZ(com_initialize)
				+ PCXC_SIZ(pal_decompressor)
				+ PCXC_SIZ(com_initialize2)
				+ PCXC_SIZ(pic_decompressor)
				+ PCXC_SIZ(com_termin_key)
				+ PCXC_SIZ(com_cleanup);
	com_picture_len = codes;

	/////////////////

	if (1!=PCXCom_Write_Chunk(com_initialize	,file)) goto err;
	if (1!=PCXCom_Write_Chunk(com_pdecomp_none	,file)) goto err;
	if (1!=PCXCom_Write_Chunk(com_initialize2	,file)) goto err;
	if (1!=PCXCom_Write_Chunk(pic_decompressor	,file)) goto err;
	if (1!=PCXCom_Write_Chunk(com_termin_key	,file)) goto err;
	if (1!=PCXCom_Write_Chunk(com_cleanup		,file)) goto err;

	memcpy(packmem,piccol,768);
	for(y=0;y<768;y++) packmem[y]>>=2;
	if (pal_decompressor == com_pdecomp_6bit) {
		uchar *fr,*to;
		uchar a,b,c;

		fr = packmem;
		to = packmem;

		for(y=768>>2;y;y--) {
			a =  fr[0]     | (fr[1]<<6);
			b = (fr[1]>>2) | (fr[2]<<4);
			c = (fr[2]>>4) | (fr[3]<<2);
			to[0]=a;
			to[1]=b;
			to[2]=c;
			to+=3; fr+=4;
		}
		if (1!=fwrite(packmem,192*3,1,file)) goto err;
	} else
		if (3!=fwrite(packmem,256,3,file)) goto err;

	if (stmp>=64000) {
		if (1!=fwrite(picmem,64000,1,file)) goto err;
	} else {
		vmem=picmem;
		for(y=0;y<200;y++) {
			mouse_hide();
			if (!(y & 7)) SetTitle("Packing row %d...",y);
			mouse_show();
			if (1!=fwrite(packmem,pack_PCX_row(packmem,vmem,320),1,file)) goto err;
			vmem+=320;
		};
	};


	fclose(file);
	mouse_hide();
	restore_screen();
	mouse_show();

	PictureSetName(pict_cur,name);

	return PIOERR_NONE;

err:
	fclose(file);
	return PIOERR_WRITE;
}

////////////////////////////////////////////////////////////////////

void divideby4(uchar *,ulong);
#pragma aux divideby4=		\
	"x:",			\
	"mov bl,[eax+edx-1]"	\
	"shr bl,2",		\
	"dec edx",		\
	"mov [eax+edx],bl",	\
	"jne x"			\
	parm [eax] [edx] modify exact [edx]


int PCXRead(char *name,int interactive) {
	struct PCXHeader pcxh;
	FILE *file;
	long y;
	uchar *dmem;
	uchar *temp;
	short xsize,ysize,xpicsize,xskip;
	ulong packsize;		/* packing memory size required */
	boolean	trueColor=FALSE;	// decompressing a 24-bit PCX

	uchar	*decomp_mem=NULL;
	uchar	*decomp_ptr;

	if (!*name) return PIOERR_NONAME;
	if (!(file=VPfopen(name,"rb"))) {
		if (interactive>=2)
			Message("Error opening file: %s",name);
		return PIOERR_OPEN;
	};
	if (1!=fread(&pcxh,sizeof pcxh,1,file)) goto abort;

	// NON-STANDARD: A start of '0xeb 0x0e' indicates a VP386 SFX pcx file

	if (pcxh.magic!=PCXH_MAGIC && (pcxh.magic!=0xeb || pcxh.version!=0x0e)) {
		if (interactive>=2)
			Message("Not a PCX file!");
		return PIOERR_BADFORMAT;
	};

	xsize=pcxh.x2+1-pcxh.x1;
	ysize=pcxh.y2+1-pcxh.y1;

	if (xsize!=picbm->width || ysize!=picbm->height) {
		switch(RequestBoolean("PCX Size Mismatch","Resize & Load","Load & Clip","Abort Load",
				"Picture is %ld,%ld -- PCX is %ld,%ld",picbm->width,picbm->height,xsize,ysize)) {
		case 2:	resize_picture(xsize,ysize,0);	/* junk existing picture */
			break;
		case 1:	break;		/* that was easy */
		case 0:	goto abort_user;
		};
	};

	/* make sure we don't get pestered */

	rectmodified=FALSE;

	/* pointer must be reset */

	dmem=picmem;

	if (pcxh.encoding != PCXH_RLL) {
		if (interactive>=2)
			Message("PCX file must be RLL encoded");
		goto abort;
	};

	/* Next statement is a bit odd... let me explain...
	** the expression x&(x-1) deletes the rightmost
	** (least significant) bit in x. The next command
	** determines quickly if the value is a power of 2
	** by determining if only one bit is on. */

	if (	(pcxh.bitsPerPixel & (pcxh.bitsPerPixel-1))!=0 ||
		!pcxh.bitsPerPixel ||
		pcxh.bitsPerPixel > 8) {
		if (interactive>=2)
			Message("PCX file has odd bits/pixel value: %d",pcxh.bitsPerPixel);
		goto abort;
	};

	/* 8 planes is quite unusual for PCX files, but if I get any
	** PCX files from an Amiga (unlikely), I'll be able to decompress
	** them... */
#if 0
	if (	!pcxh.planes ||
		(pcxh.planes & (pcxh.planes-1)) ||
		pcxh.planes>8) {
		if (interactive>=2)
			Message("PCX file has odd planes value: %d",pcxh.planes);
		goto abort;
	};
#endif
	// Technically, PCX files should be either byte-per-pixel, or have 8 pixels per
	// byte and use 1, 2, or 4 planes.  This prevents VGAPaint 386 from trying
	// to load true color PCX pictures (3 plane byte-per-pixel).
	//
	// New as of V1.1: allow 3 planes and 8 bits per pixel (24-bit PCX)

	trueColor = pcxh.planes==3 && pcxh.bitsPerPixel==8;

	if (pcxh.planes * pcxh.bitsPerPixel > 8 && !trueColor) {
		if (interactive>=2)
			Message("PCX file depth too great: %d bits",pcxh.planes * pcxh.bitsPerPixel);
		goto abort;
	};

	/* determine memory required for decompression */

	packsize=pcxh.bytesPerRow*pcxh.planes;

/* PCX file seems alright (for now)... begin decompression */

	vd->set(picbm,0);

/* If packsize is larger than 768, allocate a row line and copy to pict.
** Otherwise, decompress to packmem and copy to pict.
*/

	y=min(ysize,picbm->height);
	xpicsize=min(picbm->width,xsize);

//	xskip=(pcxh.bytesPerRow & 1);
	xskip = (pcxh.bytesPerRow*8 - xpicsize*pcxh.bitsPerPixel)>>3;

//	if (xsize>picbm->width) xskip+=(xsize-picbm->width);

	if (!(decomp_mem=malloc(packsize))) goto abort;
	decomp_ptr = decomp_mem;

	if (trueColor) {
		uchar *sepmem;

		// allocate 24-bit memory (sepred, sepgrn, sepblu)

		if (!(sepmem = malloc(picbm->size*3))) goto abort;
		dmem = sepmem;

		while(y--) {
			if (unpack_PCX_row(decomp_ptr,file,packsize)) {
				free(sepmem);
				goto abort;
			}
			memcpy(dmem		 ,decomp_ptr			 ,pcxh.bytesPerRow);
			memcpy(dmem+picbm->size  ,decomp_ptr+pcxh.bytesPerRow  ,pcxh.bytesPerRow);
			memcpy(dmem+picbm->size*2,decomp_ptr+pcxh.bytesPerRow*2,pcxh.bytesPerRow);
			dmem+=picbm->bpr;
		}

//		divideby4(sepmem,picbm->size*3);

		ObtainPalette2(sepmem,sepmem+picbm->size,sepmem+picbm->size*2,interactive>=3,reduce_maxcol);
		Separate_Remap2(sepmem,sepmem+picbm->size,sepmem+picbm->size*2,interactive>=3);
		free(sepmem);

		if (ysize>picbm->height)
			for (y=ysize-picbm->height;y;y--)
				if (unpack_PCX_row(decomp_ptr,file,packsize)) goto abort;

	} else {
		while(y--) {
			if (unpack_PCX_row(decomp_ptr,file,packsize)) goto abort;
			PCX_Demultiplex(dmem,decomp_ptr,xpicsize,pcxh.bitsPerPixel,pcxh.planes,xskip);
			dmem+=picbm->bpr;
		}
		if (ysize>picbm->height)
			for (y=ysize-picbm->height;y;y--)
				if (unpack_PCX_row(decomp_ptr,file,packsize)) goto abort;
	}

// if we have allocated decompression memory, free it

	if (decomp_mem) {
		free(decomp_mem);
		decomp_mem = NULL;
	}

// if the file is a VP386 self-extracting PCX, skip the 2nd half of the decompressor
// the length is held in the "videomode" field (reserved for version 5 PCX files)

	if (pcxh.magic == 0xeb || pcxh.version == 0x0e) {
		if (fseek(file,pcxh.videomode,SEEK_CUR))
			goto abort;
	}

/* Check for extended palette & use it if so */

/* NOTE: Windows Paintbrush does not save palettes properly: in 16-color modes,
**	 it saves the palette in the header, and in 256-color modes, it attaches
**	 an extended palette to the file.  This is correct, EXCEPT that Paintbrush
**	 has the annoying habit of saving palettes out all black.  To avoid this,
**	 holding down SHIFT during loading forces VGAPaint 386 to override the
**	 palette and keep the current palette instead.  Since Paintbrush really
**	 only allows you to use the lower 16 colors, the standard palette should
**	 suffice.  Surprisingly, this bug hasn't been fixed in Windows 3.11.
**
**	 Even better is the solution in Windows 95: Win95 Paint won't save PCX at all.
*/

	if (!trueColor) {	// ignore extpal for truecolor pictures
		uchar *dstpal = piccol;

//		if (pict_cur->current->frame_pal)
//			dstpal = pict_cur->current->frame_pal->palette;
//		else
//			dstpal = pict_cur->palette;

		if (!(kbd_shift() & KBD_SHIFT))	{	/* SHIFT key overrides palette */
			if (fgetc(file)==0x0c) {

/* Hurray! Read in all 768 bytes and complain like a banshee on an error */

				if (1!=fread(dstpal,768,1,file)) goto abort;

				y=768;
			} else {
				memcpy(dstpal,pcxh.palette,48);
				y=48;

				// Check palette mode, premake palette if not color

				switch(pcxh.palinfo) {
				case PCXH_PALBW:
					dstpal[0]=dstpal[1]=dstpal[2]=0;
					dstpal[3]=dstpal[4]=dstpal[5]=255;
					pict_cur->pal_bits[0]=3;
					memset(pict_cur->pal_bits+1,0,31);
					break;
				case PCXH_PALGREY:
					{
						int i;
						int p = pcxh.planes * pcxh.bitsPerPixel;
						int l = 1L<<p;

						for(i=0;i<l;i++) {
							dstpal[i*3  ]=(i*255L)/(l-1);
							dstpal[i*3+1]=(i*255L)/(l-1);
							dstpal[i*3+2]=(i*255L)/(l-1);
						}

						if (p>=3)
							memset(pict_cur->pal_bits,0xff,p>>3);
						else {
							memset(pict_cur->pal_bits,0,32);
							pict_cur->pal_bits[0]=(1L<<(1L<<p))-1;
						}
					}
					break;
				}

			}

/* The CTRL key overrides the adjustment of the PCX palette to the VGA's
** palette range.  Old versions of VGAPaint (we'll call them VGAPaint 286)
** did not rescale the palette properly, resulting in a dark image.
** Holding down CTRL while loading solves this problem.
*/

			if (kbd_shift() & KBD_CTRL) {
				temp=dstpal;
				do *temp++>>=2; while(--y);
			}
		}
	}

	fclose(file);
	PictureMarkDirty(pict_cur);
	PictureSetName(pict_cur,name);
	if (interactive>=1) {
		repick_colors();
		restore_screen();
	}
	return PIOERR_NONE;

abort:
	if (interactive>=2)
		Message("Error reading file: %s",name);
abort_user:
	if (decomp_mem) free(decomp_mem);
	fclose(file);

	return PIOERR_UNKNOWN;
}

boolean unpack_PCX_row(uchar *dmem,FILE *file,long xs) {
	long left=xs;
	long c,c2;
	while(left) {
		if (-1==(c=fgetc(file))) return TRUE;
		if (c>=0xC0) {
			if (-1==(c2=fgetc(file))) return TRUE;
			if ((left-=(c-0xC0))<0) return TRUE;
			for(c-=0xC0;c;c--) *dmem++=(uchar)c2;
		} else {
			*dmem++=c;
			--left;
		};
	};
	return FALSE;
}

void RawHam_Extract(FILE *f) {
	ushort t;
	short x,y,p,b;
	uchar amipal[48],*ap;
	uchar *sr,*sg,*sb;
	uchar c;
	uchar cr,cg,cb;

 	Separate_Free();
	if (!(sr=sepred=(uchar *)malloc(64000))) goto xit;
	if (!(sg=sepgrn=(uchar *)malloc(64000))) goto xit;
	if (!(sb=sepblu=(uchar *)malloc(64000))) goto xit;
	qsetl((ulong *)sepred,0,16000);
	qsetl((ulong *)sepgrn,0,16000);
	qsetl((ulong *)sepblu,0,16000);

	ap=amipal;
	for(x=0;x<16;x++) {
		fread(&t,2,1,f);
		*ap++=(short)(((t    )&15)*17)>>2;
		*ap++=(short)(((t>>12)&15)*17)>>2;
		*ap++=(short)(((t>>8 )&15)*17)>>2;
	};

	for(y=0;y<143;y++) {
		memset(packmem,0,320);
		for (p=1;p<64;p<<=1) {
			ap=packmem;
			for(x=0;x<40;x++) {
				c=getc(f);
				for(b=8;b;b--) {
					if ((signed char)c<0) *ap+=p;
					++ap;
					c<<=1;
				};
			};
		};
		cr=amipal[0];
		cg=amipal[1];
		cb=amipal[2];
		ap=packmem;
		for(x=0;x<320;x++) {
			c=*ap++;
			switch(c & 48) {
			case 0:
				c=c*3;
				cr=amipal[c];
				cg=amipal[c+1];
				cb=amipal[c+2];
				break;
			case 32: cr=((c & 15)*17)>>2; break;
			case 48: cg=((c & 15)*17)>>2; break;
			case 16: cb=((c & 15)*17)>>2; break;
			};
			*sr++=cr;
			*sg++=cg;
			*sb++=cb;
		};
	};
	PCXWrite24("__temp.pcx",picbm,sepred,sepgrn,sepblu);
	if (ObtainPalette()) Separate_Remap_FS();
xit:
	Separate_Free();
}

void Raw256_Extract(char *name) {
	FILE *f;
	uchar *mem=picmem;
	short x;

	if (!*name) return;
	if (!(f=VPfopen(name,"rb"))) {
		Message("Unable to open file.");
		goto xit;
	};
	if ((kbd_shift() & (KBD_LSHIFT+KBD_CTRL+KBD_ALT))==(KBD_LSHIFT+KBD_CTRL+KBD_ALT)) {
		RawHam_Extract(f);
		goto xit;
	};
		
	for (x=0;x<picbm->height;x++) {
		if (1!=fread(mem,picbm->width,1,f)) break;
		mem+=picbm->bpr;
	};
	if ((x=fgetc(f))==EOF)
		for (x=0;x<64;x++) {
			piccol[12+x*3]=x;
			piccol[13+x*3]=x;
			piccol[14+x*3]=x;
		}
	else {
		piccol[0]=(uchar)x;
		fread(piccol+1,767,1,f);
	};

	if (kbd_shift() & KBD_CTRL)
		for(x=0;x<768;x++) piccol[x]=(piccol[x]*85L)/21;

xit:
	vid_loadRGB6(0,piccol,256);
	fclose(f);
	mouse_hide();
	PictureMarkDirty(pict_cur);
	PictureSetName(pict_cur,name);
	repick_colors();
	restore_screen();
	mouse_show();
}

void	PCX_Demultiplex(uchar *d,uchar *s,ulong width,uchar bpp,uchar planes,ulong byteskip) {
	ulong wtemp;
	uchar acc,bmask;
	char pix;
	char preshift=0;
	uchar *d2;

	if (bpp==8) {	/* bpp*planes always <= 8 */
		memcpy(d,s,width);
		return;
	};
	memset(d,0,width);

	bmask=(1<<bpp)-1;

	do {
		wtemp=width;
		pix=bpp;
		d2=d;
		do {
			if (!(pix-=bpp)) {
				pix=8;
				acc=brotl(*s++,preshift);
			};
			acc=brotl(acc,bpp);
			*d2++ += (acc & bmask);
		} while(--wtemp);
		bmask <<= bpp;
		preshift+=bpp;
		s+=byteskip;
	} while(--planes);
}


enum PictureIOErrors Pak_Dump(char *name,struct BitMap *bm) {
	FILE *f;
	int l;
	uchar *pp;

	if (!name || !*name) return PIOERR_NONAME;
	if (!(f=VPfopen(name,"wb"))) return PIOERR_OPEN;

// don't dump out junk at eol
//	if (1!=fwrite(picmem,picbm->size,1,f)) {
//		fclose(f);
//		return PIOERR_WRITE;
//	}

	pp=picmem;
	for(l=picbm->height;l;l--) {
		if (1!=fwrite(pp,picbm->width,1,f)) goto abort;
		pp += picbm->bpr;
	}

	if (1!=fwrite(piccol,768,1,f)) goto abort;

	fclose(f);
	return PIOERR_NONE;
abort:
	fclose(f);
	return PIOERR_WRITE;
}

/******** IFF routines ********/

struct	Text IFFReq_Texts[]={
	{ "Colors"	,  4, 14,St0,St2 },
	{ "2"		, 61, 14,St0,St2 },
	{ "4"		, 76, 14,St0,St2 },
	{ "8"		, 91, 14,St0,St2 },
	{ "16"		,103, 14,St0,St2 },
	{ "32"		,118, 14,St0,St2 },
	{ "64"		,133, 14,St0,St2 },
	{ "256"		,148, 14,St0,St2 },
	{ "HAM"		,169, 14,St0,St2 },

	{ "Palette"	,  4, 26,St0,St2 },
	{ "4-bit"	, 58, 26,St0,St2 },
	{ "6-bit"	, 91, 26,St0,St2 },
	{ "Reduce"	,163, 26,St0,St2 },

	{ "Type"	,  4, 38,St0,St2 },

	{ "Ok"		, 30, 58,St0,St2 },
	{ "Cancel"	,150, 58,St0,St2 },
	{NULL}
};

struct	Zone	IFFReq_Z1[]={
	{  56, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{  71, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{  86, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{ 101, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{ 116, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{ 131, 12, 15, 11,ZN_3D			,NULL,NULL,NULL },
	{ 146, 12, 21, 11,ZN_3D+ZN_INVERTED	,NULL,NULL,NULL },
	{ 167, 12, 33, 11,ZN_3D			,NULL,NULL,NULL	},
	{ -1,-1,-1,-1 }
};

struct Zone	IFFReq_Z2[]={
	{  56, 24, 33, 11,ZN_3D+ZN_INVERTED	,NULL,NULL,NULL },
	{  89, 24, 33, 11,ZN_3D			,NULL,NULL,NULL },
	{ -1,-1,-1,-1 }
};

struct Zone	IFFReq_Z3[]={
	{ 161, 24, 39, 11,ZN_3D+ZN_INVERTED	,NULL,NULL,NULL },
	{ -1,-1,-1,-1 }
};

struct Zone	IFFReq_Z4[]={
	{   4, 50, 64, 24,ZN_3D			,NULL,NULL,NULL },
	{ 136, 50, 64, 24,ZN_3D			,NULL,NULL,NULL },
	{ -1,-1,-1,-1 }
};

struct	Zone	IFFReq_Z5[]={
	{  56, 36, 30, 11,ZN_3D+ZN_INVERTED+ZN_TEXTRENDER	,NULL,"ILBM",NULL },
	{  86, 36, 30, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"ACBM",NULL },
	{ 116, 36, 30, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"PBM ",NULL },
	{ -1,-1,-1,-1 }
};

struct	ZoneRange IFFReq_ZR[]={
	{ &IFFReq_ZR[1]	,IFFReq_Z1,&IFFReq_Z1[6], 6,ZR_CHOICE,0,0,0,NULL,FALSE },
	{ &IFFReq_ZR[2]	,IFFReq_Z2,&IFFReq_Z2[0], 0,ZR_CHOICE,0,0,0,NULL,FALSE },
	{ &IFFReq_ZR[3]	,IFFReq_Z3,NULL		,-1,ZR_TOGGLE,0,0,0,NULL,FALSE },
	{ &IFFReq_ZR[4]	,IFFReq_Z4,NULL		,-1,ZR_TAP   ,0,0,0,NULL,FALSE },
	{ NULL		,IFFReq_Z5,&IFFReq_Z5[0], 0,ZR_CHOICE,0,0,0,NULL,FALSE },
};

static struct Requester IFFReq={
	0,0,204,78,
	REQ_BORDER,
	"IFF saving options",
	St2,
	NULL,
	NULL,
	IFFReq_ZR,
	NULL,
	NULL,
	NULL,
	IFFReq_Texts
};

#define	IFF_FORMAT_ILBM	0
#define	IFF_FORMAT_ACBM	1
#define	IFF_FORMAT_PBM	2

static struct IFFRequestReturn {
	short	planes;		/* 12 for HAM-4 */
	flag	reduce:1,
		pal6bit:1;
	short	format;
} iffreq;

boolean	IFFRequest(void) {
	mouse_coord x,y,b;
	ulong msg;
	boolean quit=FALSE,okay=FALSE;

	if (!OpenRequester(&IFFReq,0,0)) return FALSE;

	do {
		mouse_read2(&x,&y,&b);
		switch(msg=CheckRequester(&IFFReq,x,y,b)) {
		case RZoneMsg(3,0):
			okay=TRUE;
		case RZoneMsg(3,1):
			quit=TRUE;
			break;
		};
	} while(!quit);
	CloseRequester(&IFFReq);

	iffreq.planes="\x1\x2\x3\x4\x5\x6\x8\xc"[IFFReq_ZR[0].highindex];
	iffreq.reduce=((IFFReq_Z3[0].flags & ZN_INVERTED)!=0);
	iffreq.pal6bit=(IFFReq_ZR[2].highindex == 1);
	iffreq.format=IFFReq_ZR[4].highindex;

	return okay;
};

//	boolean IFF_Line_Decompress(uchar *output,FILE *i,long length);
//
//	Reads bytes from the input stream and decompresses bytes until <length> bytes
//	have been written to the output pointer.  Returns TRUE if successful; returns FALSE
//	if failed (i.e. error in input stream or incorrect number of bytes during
//	decompression).

boolean	IFF_Line_Decompress(uchar *output,FILE *i,long length) {
	short c,d;
	do {
		if ((c=getc(i))==EOF) return FALSE;
		if (c!=128) {
			if (((signed char)c)<0) {
				if ((d=getc(i))==EOF) return FALSE;
				c=-((signed char)c)+1;
				if ((length-=c)<0) return FALSE;
				do {
					*output++=d;
				} while(--c);
			} else {
				++c;
				if ((length-=c)<0) return FALSE;
				if (1!=fread(output,c,1,i)) return FALSE;
				output+=c;
			};
		};
	} while(length);
	return TRUE;
}

boolean	IFF_Line_DecompressMemory(uchar *output,uchar *input,long length) {
	short c,d;
	do {
		c=*input++;
		if (((signed char)c)<0) {
			d=*input++;
			c=-c+1;
			if ((length-=c)<0) return FALSE;
			do {
				*output++=d;
			} while(--c);
		} else {
			++c;
			if ((length-=c)<0) return FALSE;
			memcpy(output,input,c);
			output+=c;
			input+=c;
		};
	} while(length);
	return TRUE;
}

//	ulong IFF_Line_Compress(uchar *output,uchar *input,long length);
//
//	Compresses <length> bytes from the input memory area to the output area.
//	Values 0x81-0xFF indicate a run, where 0xFF is 2 bytes and 0x81 is 128 bytes.
//	Values 0x00-0x7F indicate a literal string, where 0x00 is a single byte and
//	0x7F is 128 bytes.  For true IFF compatibility the value 0x80 is not used (it
//	is defined as a no-op).  The function returns the number of bytes compressed.
//	To determine the maximum output size, use IFF_Line_Max().

#define	IFFLC_LITERAL	(0)
#define	IFFLC_REPEAT	(1)
#define	IFFLC_NONE	(2)

ulong	IFF_Line_Compress(uchar *output,uchar *input,long length) {
	uchar *old_output;
	uchar *key;
	char mode=IFFLC_REPEAT;
	uchar x,y;
	int cnt;

	old_output=output;

	y=*input++;
	--length;
	while(length) {
		if (mode==IFFLC_REPEAT) {
			x=y;
			y=*input++;
			if (!--length) {
				mode=IFFLC_LITERAL;
				break;
			};
		};
		if (x==y) {
			mode=IFFLC_REPEAT;
			cnt=2;
			key=output++;
			*output++=x;
			for(;;) {
				y=*input++;
				if (!--length) {
					if (x==y && cnt<128) ++cnt;
					else {
						*output++=0;
						*output++=y;
					};
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x!=y) break;
				++cnt;
			};
			*key=1-cnt;
		} else {
			mode=IFFLC_LITERAL;
			cnt=1;
			key=output++;
			for(;;) {
				*output++=x;
				x=y;
				y=*input++;
				if (!--length) {
					if (cnt<127) cnt+=2;
					else {
						*output++=1;
					};
					*output++=x;
					*output++=y;
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x==y) {

					/* if we have ??xx?? include xx in literal run */

					if (*input==x)	break;
				};
				++cnt;
			};
			*key=cnt-1;
		};
	};

	if (mode==IFFLC_LITERAL) {
		if (x==y) {
			*output++=0xFF;		/* repeat of 2 */
		} else {
			*output++=0x01;		/* literal 2 */
			*output++=x;
		};
		*output++=y;
	} else if (mode==IFFLC_REPEAT) {
		*output++=0x00;	/* literal 1 */
		*output++=y;
	};

	return output-old_output;
}

//	ulong IFF_Line_Compress_Cautious(uchar *output,uchar *input,long length,long maxlen);
//
//	Compresses <length> bytes from the input memory area to the output area.
//	Values 0x81-0xFF indicate a run, where 0xFF is 2 bytes and 0x81 is 128 bytes.
//	Values 0x00-0x7F indicate a literal string, where 0x00 is a single byte and
//	0x7F is 128 bytes.  For true IFF compatibility the value 0x80 is not used (it
//	is defined as a no-op).  The function returns the number of bytes compressed.
//
//	The function aborts prematurely if the output exceeds maxlen.  Since the check
//	slows down compression, use IFF_Line_Compress() for unchecked compression.
//	Returns 0 if limit exceeded.

#define	CHECK	if (output>=limit) return 0

ulong	IFF_Line_Compress_Cautious(uchar *output,uchar *input,long length,long maxlen) {
	uchar *old_output,*limit;
	uchar *key;
	char mode=IFFLC_REPEAT;
	uchar x,y;
	int cnt;

	old_output=output;
	limit = output + maxlen;

	y=*input++;
	--length;
	while(length) {
		if (mode==IFFLC_REPEAT) {
			x=y;
			y=*input++;
			if (!--length) {
				mode=IFFLC_LITERAL;
				break;
			};
		};
		if (x==y) {
			mode=IFFLC_REPEAT;
			cnt=2;
			CHECK; key=output++;
			CHECK; *output++=x;
			for(;;) {
				y=*input++;
				if (!--length) {
					if (x==y && cnt<128) ++cnt;
					else {
						CHECK; *output++=0;
						CHECK; *output++=y;
					};
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x!=y) break;
				++cnt;
			};
			*key=1-cnt;
		} else {
			mode=IFFLC_LITERAL;
			cnt=1;
			CHECK; key=output++;
			for(;;) {
				CHECK; *output++=x;
				x=y;
				y=*input++;
				if (!--length) {
					if (cnt<127) cnt+=2;
					else {
						CHECK; *output++=1;
					};
					CHECK; *output++=x;
					CHECK; *output++=y;
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x==y) {

					/* if we have ??xx?? include xx in literal run */

					if (*input==x)	break;
				};
				++cnt;
			};
			*key=cnt-1;
		};
	};

	if (mode==IFFLC_LITERAL) {
		if (x==y) {
			CHECK; *output++=0xFF;		/* repeat of 2 */
		} else {
			CHECK; *output++=0x01;		/* literal 2 */
			CHECK; *output++=x;
		};
		CHECK; *output++=y;
	} else if (mode==IFFLC_REPEAT) {
		CHECK; *output++=0x00;	/* literal 1 */
		CHECK; *output++=y;
	};

	return output-old_output;
}

#undef	CHECK

//	ulong IFF_Line_Compress_Predict(uchar *input,long length);
//
//	Predicts the compression of the input memory area in bytes.
//	Values 0x81-0xFF indicate a run, where 0xFF is 2 bytes and 0x81 is 128 bytes.
//	Values 0x00-0x7F indicate a literal string, where 0x00 is a single byte and
//	0x7F is 128 bytes.  For true IFF compatibility the value 0x80 is not used (it
//	is defined as a no-op).  The function returns the size of the compressed object.

ulong	IFF_Line_Compress_Predict(uchar *input,long length) {
	char mode=IFFLC_REPEAT;
	uchar x,y;
	int	cnt;
	ulong	size=0;

	y=*input++;
	--length;
	while(length) {
		if (mode==IFFLC_REPEAT) {
			x=y;
			y=*input++;
			if (!--length) {
				mode=IFFLC_LITERAL;
				break;
			};
		};
		if (x==y) {
			mode=IFFLC_REPEAT;
			cnt=2;
			size+=2;
			for(;;) {
				y=*input++;
				if (!--length) {
					if (x==y && cnt<128) ++cnt;
					else size += 2;
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x!=y) break;
				++cnt;
			};
		} else {
			mode=IFFLC_LITERAL;
			cnt=1;
			++size;
			for(;;) {
				++size;
				x=y;
				y=*input++;
				if (!--length) {
					if (cnt<127) cnt+=2;
					else ++size;
					size += 2;
					mode=IFFLC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x==y) {

					/* if we have ??xx?? include xx in literal run */

					if (*input==x)	break;
				};
				++cnt;
			};
		};
	};

	if (mode==IFFLC_LITERAL) {
		if (x==y) ++size;		/* repeat of 2 */
		else	  size+=2;		/* literal 2 */
		++size;
	} else if (mode==IFFLC_REPEAT) {
		size+=2;	/* literal 1 */
	};

	return size;
}

//	void	MemFlip(void *dest,ulong length);
//
//	Swap bytes of a memory block as to completely reverse its byte order.

void asm_memflip(void *s1,void *s2,ulong halflength);
#pragma aux asm_memflip="l:","mov al,[esi]","xchg [edi],al","mov [esi],al","dec edi","inc esi","dec ecx","jne l"	\
	parm [esi] [edi] [ecx] modify exact [al esi edi ecx];

void	MemFlip(void *dest,ulong length) {
	uchar *s1,*s2;
	s1=dest;
	s2=s1+length-1;
	if (!(length>>=1)) return;
	asm_memflip(s1,s2,length);
}

//	All structures are here since I don't expect to need to read IFF anywhere else
//	in VGAPaint 386.  One word of caution:  IFF words are opposite Intel words.  The
//	way I get around this is by creating req'd structures backwards and using a memory
//	flip routine.

typedef	uchar		Masking;
typedef	uchar		Compression;

#define	mskNone			0
#define	mskHasMask		1
#define	mskHasTransparentColor	2
#define	mskLasso		3

#define	cmpNone			0
#define	cmpByteRun1		1

struct	IFF_BMHD {	// flip this structure
	short		pageHeight,pageWidth;
	uchar		yAspect,xAspect;
	ushort		transparentColor;
	uchar		pad1;
	Compression	compression;
	Masking		masking;
	uchar		nPlanes;
	short		y,x;
	ushort		h,w;
};

struct	IFF_CMAP {	// no flipping
	uchar	red,green,blue;
};

long IFF_Backwards(long);
#pragma aux IFF_Backwards="xchg ah,al","rol eax,16","xchg ah,al" parm [eax] value [eax] modify exact [eax] nomemory;

boolean	IFF_Write_Structure(FILE *f,void *strc,ulong siz,char *tag) {
	ulong siz2;
	siz2=IFF_Backwards(siz);
	if (1!=fwrite(tag,4,1,f)) return FALSE;
	if (1!=fwrite(&siz2,4,1,f)) return FALSE;
	if (1!=fwrite(strc,siz,1,f)) return FALSE;
	if (siz&1) fputc(0,f);
	return TRUE;
}

boolean	IFF_Write_StructureRev(FILE *f,void *strc,ulong siz,char *tag) {
	MemFlip(strc,siz);
	return IFF_Write_Structure(f,strc,siz,tag);
};

boolean	IFF_Write_LongRev(FILE *f,long val) {
	val=IFF_Backwards(val);
	return 1==fwrite(&val,4,1,f);
}

static char IFF_Palette_Colors[]={
	1,		/* 0 planes 1 color */
	2,		/* 1 plane  2 colors */
	4,		/* 2 planes 4 colors */
	8,		/* 3 planes 8 colors */
	16,		/* 4 planes 16 colors */
	32,		/* 5 planes 32 colors */
	64,		/* 6 planes 32 colors Extra-Halfbrite */
	128,		/* 7 planes 128 colors */
};

enum PictureIOErrors IFFWrite(char *output,struct BitMap *bm) {
	FILE *f;
	ulong pos;
	ulong siz;
	ulong csiz;
	short p,y,wPlanes,wPlaneMax,pBpr,pSkip;
	uchar *s;
	uchar *packArea;
	uchar *planeArea;
	ulong packSize;
	struct IFF_BMHD bmhd={
		200,320,72,80,0,0,cmpByteRun1,mskNone,8,0,0,200,320
	};

	if (!IFFRequest()) return PIOERR_NONAME;

	if (!output) return PIOERR_NONAME;

	bmhd.w=bm->width;
	bmhd.h=bm->height;
	bmhd.nPlanes=	/* LISP time! */
		(iffreq.format==IFF_FORMAT_PBM
			? 1
			: (iffreq.planes==12	/* save as HAM? */
				? (iffreq.pal6bit
					? 8	/* save as HAM-8 if using 6-bit palette (AGA only) */
					: 6	/* save as HAM-4 if using 4-bit palette (All Amigas) */
				  )
				: iffreq.planes	/* 2/4/8/16/32/64/256 */
			  )
		);
	bmhd.pageWidth=picbm->width;
	bmhd.pageHeight=picbm->height;
	bmhd.xAspect=80;
	bmhd.yAspect=72;

	if (iffreq.planes==12) wPlanes=iffreq.reduce ? 8 : 6; else wPlanes=iffreq.planes;
	wPlaneMax=1<<wPlanes;
	pBpr=(picbm->width+7)>>3;
	pSkip=pBpr & 1;

	if (iffreq.reduce)
		if (iffreq.planes != 8)	{	/* do we have to reduce the number of colors in the picture? */

		/* if we are using HAM-4 we need to reduce to 16 colors
		** if we are using HAM-8 we need to reduce to 64 colors
		** if we are using EHB we need to reduce to 32 colors
		*/

			memcpy(packmem+256,piccol,512);
			ReducePalette(
				iffreq.planes == 12 ? (iffreq.pal6bit ? 64 : 16) : IFF_Palette_Colors[iffreq.planes],
				packmem,
				packmem+256	/* since we'll end up with max 64 colors, only 192 bytes needed */
			);
			vid13_translate(picbm,0,0,picbm->width,picbm->height,packmem);
			memcpy(piccol,packmem+256,512);
		};

	packSize=((pBpr+pSkip+127)>>8)*129+pBpr+pSkip;
	if (packSize>768) {
		if (!(planeArea=malloc(packSize))) {
//			Message("Not enough memory to compress picture.");
			return PIOERR_MEMORY;
		};
	} else	planeArea=packmem;
	memset(planeArea,0,packSize);		/* clear it to avoid problems */
	packArea=planeArea+pBpr+pSkip;


	if (f=VPfopen(output,"wb")) {
		if (1!=fwrite("FORMxxxx",8,1,f)) goto writeerr;
		if (1!=fwrite(&"ILBMACBMPBM "[iffreq.format<<2],4,1,f)) goto writeerr;

		if (!IFF_Write_StructureRev(f,&bmhd,sizeof bmhd,"BMHD")) goto writeerr;
		if (!IFF_Write_Structure(f,piccol,768,"CMAP")) goto writeerr;

		if (1!=fwrite(&"BODYABITBODY"[iffreq.format<<2],4,1,f)) goto writeerr;
		pos=ftell(f);
		if (1!=fwrite("xxxx",4,1,f)) goto writeerr;
		siz=0;
		s=picmem;
		switch(iffreq.format) {
		case IFF_FORMAT_ACBM:	/* uncompressed non-interleaved planes */
			for(p=1;p<wPlaneMax;p<<=1) {
				for(y=0;y<200;y++) {
					ChunkyToPlanar(s,planeArea,pBpr,p);
					if (1!=fwrite(planeArea,pBpr+pSkip,1,f)) goto writeerr;
				};
			};
			siz+=(pBpr+pSkip)*wPlanes;
		default:
			for(y=0;y<200;y++) {
				if (iffreq.format==IFF_FORMAT_ILBM) {	/* compressed interleaved planes */
					for(p=1;p<wPlaneMax;p<<=1) {
						ChunkyToPlanar(s,planeArea,pBpr,p);
						csiz=IFF_Line_Compress(packArea,planeArea,pBpr+pSkip);
						if (1!=fwrite(packArea,csiz,1,f)) goto writeerr;
					};
				} else {	/* PBM format (compressed byte-per-pixel lines) */
					csiz=IFF_Line_Compress(packArea,s,320);
					if (1!=fwrite(packArea,csiz,1,f)) goto writeerr;
				};
				siz+=csiz;
				s+=picbm->bpr;
			};
			break;
		};
		if (siz & 1) fputc(0,f);
		csiz=ftell(f)-8;
		fseek(f,4,SEEK_SET);
		if (!IFF_Write_LongRev(f,csiz)) goto writeerr;
		fseek(f,pos,SEEK_SET);
		if (!IFF_Write_LongRev(f,siz)) goto writeerr;
		fclose(f);
	} else return PIOERR_OPEN;

	if (planeArea && planeArea!=packmem) free(planeArea);

	PictureSetName(pict_cur,output);

	return PIOERR_NONE;

writeerr:
	fclose(f);
	return PIOERR_WRITE;
}

/////////////// IFF READING ROUTINES ////////////////

enum IFFType {
	IFFT_ILBM,
	IFFT_ACBM,
	IFFT_PBM,
};

enum {
	IFF_FOUND_CMAP=0x01,
	IFF_FOUND_BMHD=0x02,
};

boolean	IFFRead(char *name) {
	FILE *f;
	enum IFFType ifft;
	uchar found_flags=0;
	ulong len;
	struct IFF_BMHD bmhd;
	uchar *dcmem=packmem;		// decompression memory
	int	width,height;

	if (!(f=VPfopen(name,"rb"))) return FALSE;

	// read the first 12 bytes of the file into memory & check them

	if (1!=fread(packmem,12,1,f)) goto readerr;
	packmem[12]=0;		// zero-terminate string

	if (*(ulong*)packmem!='MROF') { Message("Not an IFF file!"); goto close_file; }

	switch(*(ulong *)(packmem+8)) {
	case 'NBLI':	// IFF pictures from "Who Framed Roger Rabbit"(tm) have this mutilated tag
	case 'MBLI':	ifft=IFFT_ILBM; break;
	case 'MBCA':	ifft=IFFT_ACBM; break;
	case ' MBP':	ifft=IFFT_PBM; break;
	default:	Message("Invalid FORM type: %s",packmem+8); goto close_file;
	};

	// we have an ILBM/ACBM/PBM... process each chunk as it arrives
	// don't check the FORM length, since some programs don't set this correctly
	// (you see this in the "Star Trek: Next Generation Trivia Challenge" files)
	// the result is that some programs, like DPaint, refuse to load the pictures
	// it may be intentional, but I'm writing this for versatility...

	while(1==fread(packmem,8,1,f)) {
		len=IFF_Backwards(*(ulong*)(packmem+4));
		switch(*(ulong *)packmem) {
		case 'DHMB':	//BitMapHeader
			if (len>sizeof packmem) { Message("BMHD structure too long"); return FALSE; };	// too long for buffer (768b), must be bad
			if (1!=fread(packmem,len,1,f)) goto readerr;
			// flip the data structure into the <bmhd> variable
			memcpy(&bmhd,packmem,sizeof bmhd);
			MemFlip(&bmhd,sizeof bmhd);

			// digest the data in the BMHD and see if it is all valid

			if (bmhd.w==0 || bmhd.h==0 || bmhd.w>0x7FF4 || bmhd.h>0x7FF4)
				{ Message("Invalid image size: %d,%d",bmhd.w,bmhd.h); goto close_file; }
			if (bmhd.nPlanes<1 || bmhd.nPlanes>8)
				{ Message("Invalid plane value: %d planes (%d colors)",bmhd.nPlanes,1L<<bmhd.nPlanes); goto close_file; }

			// check current picture size, and request size change if needed

			if (bmhd.w != picbm->width || bmhd.h != picbm->height)
				switch(RequestBoolean("IFF Size Mismatch","Resize & Load","Load & Clip","Abort Load",
						"Picture is %ld,%ld -- IFF is %ld,%ld",picbm->width,picbm->height,bmhd.w,bmhd.h)) {
				case 2:	resize_picture(bmhd.w,bmhd.h,0);
					break;
				case 1:	break;
				case 0: goto close_file;
				};

			// check to see if <packmem> is sufficient to decompress

			{
				ulong needmem;

				switch(ifft) {
				case IFFT_ILBM:
					needmem=(((bmhd.w+15)>>3)&0xFFFFFFFE)*(bmhd.nPlanes+(bmhd.masking == mskHasMask)) + ((bmhd.w+1)&0xFFFFFFFE);
					break;
				case IFFT_ACBM:
					needmem=(((bmhd.w+15)>>3)&0xFFFFFFFE)*2;
					break;
				case IFFT_PBM:
					needmem=bmhd.w+(bmhd.w&1);
					break;
				};

				if (needmem>sizeof packmem)
					if (!(dcmem=malloc(needmem))) { Message("Not enough memory to decompress picture."); goto close_file; }
			}

			width = min(picbm->width,bmhd.w);
			height = min(picbm->height,bmhd.h);

			// check & set the found flag

			if (found_flags & IFF_FOUND_BMHD) Message("WARNING: Multiple BMHD's found in file.");

			found_flags |= IFF_FOUND_BMHD;

			break;

		case 'PAMC':	//ColorMap
			if (len>768) Message("WARNING: Too many colors in CMAP chunk (%d)",len/3);
			if (1!=fread(piccol,min(len,768),1,f)) goto readerr;

			// if we have extra bytes, skip them

			if (len>768) fseek(f,len-768,SEEK_CUR);

			found_flags |= IFF_FOUND_CMAP;

			break;

		case 'YDOB':	// BODY chunk, ILBM and PBM formats
		case 'TIBA':	// ABIT chunk, ACBM format

			// check to see if we have mismatching formats

			packmem[4]=0;
			if (	(*(ulong*)packmem == 'YDOB' && ifft == IFFT_ACBM) ||
				(*(ulong*)packmem == 'TIBA' && ifft != IFFT_ACBM))
				if (RequestBoolean("VP386 Warning","Load",NULL,"Abort","Data chunk (%s) does not match IFF type",packmem)!=2)
					return FALSE;

			// check to see if the proper chunk(s) have been found

			if (!(found_flags & IFF_FOUND_BMHD)) {
				Message("BMHD chunk must proceed ABIT/BODY chunk.");
				return FALSE;
			};

			// decompress the picture line-by-line

			UndoInvalidate();
			switch(ifft) {
			case IFFT_ILBM: {
				ushort y,p,maxplane;
				ulong linewid;
				uchar *pp,*dc,*dc2;

				linewid=((bmhd.w+15)>>3)&0xFFFFFFFE;
//				linewid=((bmhd.w+7)>>3);
				maxplane=1L<<bmhd.nPlanes;
				dc2=dcmem + linewid*bmhd.nPlanes;

//				Message("w%ld linewid:%ld  maxplane:%ld  comp:%ld",bmhd.w,linewid,maxplane,bmhd.compression);

//				vd->set(picbm,0);

				pp=picbm->vm;

				for(y=height;y;y--) {
//					Message("attempting (%d) fpos (%08lx)",y,ftell(f));
					if (bmhd.compression) {
						dc=dcmem;
						for(p=0;p<bmhd.nPlanes;p++) {
							if (!IFF_Line_Decompress(dc,f,linewid)) goto readerr;
							dc += linewid;
						};

						// if we have a mask, decompress it

						if (bmhd.masking == mskHasMask)
							if (!IFF_Line_Decompress(dc,f,linewid)) goto readerr;

					} else if (1!=fread(dcmem,linewid*(bmhd.nPlanes+(bmhd.masking == mskHasMask)),1,f)) goto readerr;
					memset(dc2,0,bmhd.w);
					dc=dcmem;
					for(p=1;p<maxplane;p<<=1) {
						PlanarToChunky(dc2,dc,linewid,p);
						dc += linewid;
					};
					memcpy(pp,dc2,width);
					pp+=picbm->bpr;
				};
				};break;
			case IFFT_PBM: {
				ushort y;
				uchar *pp;

				pp=picbm->vm;

				for(y=height;y;y--) {
					if (bmhd.compression)	if (!IFF_Line_Decompress(dcmem,f,bmhd.w+(bmhd.w&1))) goto readerr;
					memcpy(pp,dcmem,width);
					pp+=picbm->bpr;
				};
				
				}; break;
			case IFFT_ACBM:
				Message("ACBM decompression not implemented");
				break;
			}

			if (len&1) fgetc(f);

			break;
		default:
			// skip over the unknown chunk
			fseek(f,len,SEEK_CUR);
			break;
		};
	};

	if (dcmem != packmem) free(dcmem);
	fclose(f);
	PictureMarkDirty(pict_cur);
	PictureSetName(pict_cur,name);
	repick_colors();
	restore_screen();
	return TRUE;

readerr:
	Message("Error reading ILBM file!");
close_file:
	fclose(f);
	if (dcmem && dcmem != packmem) free(dcmem);
	return FALSE;
};


/////////////////// BMP WRITING ROUTINES ///////////////////////

//	int error = BMPWrite(char *fname, struct BitMap *bitmap);
//
//	A rather cheap & dirty way to write out an uncompressed BMP file.
//
//	This routine is extra lazy and writes out a file with BI_RGB compression
//	(read uncompressed) and furthermore says the picture uses all 256 colors -
//	256 IMPORTANT colors.  Ought to be interesting viewing in Windows.

int	BMPWrite(char *fn,struct BitMap *bm) {
	FILE *f;
	struct BMPFileHeader bmpfh = { 0x4D42,0,0,0 };
	struct BMPInfoHeader bmpih;
	uchar pal[1024];
	int i;
	uchar *s,*t;

	if (!fn) return PIOERR_NONAME;
	if (!*fn) return PIOERR_NONAME;
	if (!(f=VPfopen(fn,"wb"))) return PIOERR_OPEN;

	bmpfh.bfSize		= bm->bpr * bm->height + sizeof(struct BMPFileHeader) + sizeof(struct BMPInfoHeader) + 1024;
	bmpfh.bfOffBits	= sizeof(struct BMPFileHeader) + sizeof(struct BMPInfoHeader) + 1024;

	memset(&bmpih,0,sizeof bmpih);	// clear BMPInfoHeader

	bmpih.biSize		= sizeof(struct BMPInfoHeader);
	bmpih.biWidth		= bm->bpr;	// we're lucky that the VP386 video library pads to 32-bit row sizes
	bmpih.biHeight		= bm->height;
	bmpih.biPlanes		= 1;
	bmpih.biBitCount  	= 8;
	bmpih.biCompression	= 0;		// good 'ole zero
	bmpih.biSizeImage	= bm->bpr * bm->height;		// stupid DEUTEX1B can't handle zero
	bmpih.biXPelsPerMeter	= 80;
	bmpih.biYPelsPerMeter	= 72;
	bmpih.biClrUsed   	= 256;		// all 256 colors used???
	bmpih.biClrImportant	= 256;		// all 256 colors important!?!?

	{
		s=pal;
		t=piccol;

		for(i=256;i;i--) {
			s[0]=t[2]; s[1]=t[1]; s[2]=t[0]; s[3]=0;
			s+=4; t+=3;
		}
	}

	do {	// sneaky catch/throw
		if (1!=fwrite(&bmpfh,sizeof bmpfh,1,f)) break;
		if (1!=fwrite(&bmpih,sizeof bmpih,1,f)) break;
		if (1!=fwrite(pal,sizeof pal,1,f)) break;

		// write out the rows of the picture from bottom to top (looney format)

		s=bm->vm + bm->bpr*bm->height;
		for(i=bm->height;i;i--) {
			s-=bm->bpr;
			if (1!=fwrite(s,bm->bpr,1,f)) goto write_err;
		}

		fclose(f);
		PictureSetName(pict_cur,fn);
		return 0;
	}while(0);

write_err:
	fclose(f);
	return PIOERR_WRITE;
}

//	int error = BMPRead(char *fname);
//
//	Read a Windows/OS2 .BMP file into memory.
//
//	Does not support OS/2 compressed BMP's.

int	BMPRead(char *fname) {
	FILE *f;
	struct BMPFileHeader bmpfh;
	struct BMPInfoHeader bmpih;
	uchar pal[1024];
	uchar *dp,*sp;
	int y,i,xs,ys;

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;
	if (!(f=VPfopen(fname,"rb"))) return PIOERR_NONAME;

	// attempt to read in BMP file header

	if (1!=fread(&bmpfh,sizeof bmpfh,1,f)) goto abort;
	if (1!=fread(&bmpih,sizeof bmpih,1,f)) goto abort;

	if (bmpfh.bfType != 0x4d42) goto bad_format;
	if (bmpih.biSize < sizeof(struct BMPInfoHeader)) goto bad_format;
	if (bmpih.biPlanes != 1) goto bad_format;
	if (bmpih.biBitCount != 8) goto bad_format;

	if (fseek(f,bmpih.biSize - sizeof(struct BMPInfoHeader),SEEK_CUR)) goto abort;
	if (4!=fread(pal,min(bmpih.biClrUsed ? bmpih.biClrUsed : (1L<<bmpih.biBitCount),256),4,f)) goto abort;
	if (bmpih.biWidth!=picbm->width || bmpih.biHeight!=picbm->height) {
		switch(RequestBoolean("BMP Size Mismatch","Resize & Load","Load & Clip","Abort Load",
				"Picture is %ld,%ld -- BMP is %ld,%ld",picbm->width,picbm->height,bmpih.biWidth,bmpih.biHeight)) {
		case 2:	resize_picture(bmpih.biWidth,bmpih.biHeight,0);	/* junk existing picture */
			break;
		case 1:	break;		/* that was easy */
		case 0:	goto abort_user;
		};
	};

	// seek to start of picture
	// if clipping, seek to start of load region

	dp = picbm->vm + picbm->size - picbm->bpr;
	xs = min(picbm->bpr,(bmpih.biWidth+3)&0xfffffffc);
	ys = min(picbm->height,bmpih.biHeight);

	if (fseek(f,bmpfh.bfOffBits + (bmpih.biHeight-ys)*((bmpih.biWidth+3)&~3),SEEK_SET)) goto abort;

	// read the picture in, bottom to top

	for(y=ys;y;y--) {
		if (1!=fread(dp,xs,1,f)) goto abort;
		if (picbm->bpr < bmpih.biWidth)
			if (fseek(f,bmpih.biWidth-picbm->bpr,SEEK_CUR)) goto abort;
		dp-=picbm->bpr;
	}

	// decode colors

	sp = pal;
	dp = piccol;

	i=0;

	// some programs give really weird color-use counts like 257

	y = 1L<<bmpih.biBitCount;
	if (bmpih.biClrUsed) if (y>bmpih.biClrUsed) y=bmpih.biClrUsed;

	for(;y;y--) {
		dp[0]=sp[2];
		dp[1]=sp[1];
		dp[2]=sp[0];
		sp+=4;
		dp+=3;

		pict_cur->pal_bits[i>>3] |= (1<<(i&7));
		++i;
	}

	fclose(f);
	PictureMarkDirty(pict_cur);
	PictureSetName(pict_cur,fname);
	repick_colors();
	restore_screen();
	return PIOERR_NONE;
abort:
	fclose(f);
	return PIOERR_READ;
bad_format:
	fclose(f);
	return PIOERR_BADFORMAT;
abort_user:
	fclose(f);
	return PIOERR_ABORT_USER;
}



//////////////////////////////// PAL ROUTINES /////////////////////////////////////


int	PALWrite(char *fname,struct BitMap *bm) {
	FILE *f;

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;

	if (!(f=VPfopen(fname,"wb"))) return PIOERR_OPEN;

	if (1!=fwrite(bm->pal,768,1,f)) {
		fclose(f);
		return PIOERR_WRITE;
	}

	fclose(f);
	return PIOERR_NONE;
}


int	PALRead(char *fname) {
	FILE *f;

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;

	if (!(f=VPfopen(fname,"rb"))) return PIOERR_OPEN;
	if (1!=fread(piccol,768,1,f)) {
		fclose(f);
		return PIOERR_READ;
	}
	fclose(f);

	return PIOERR_NONE;
}

/////////////////////// HP-48G 8-COLOR GROB ROUTINES ////////////////////////////

enum PictureIOErrors HP48Write3(char *name,struct BitMap *bm) {
	FILE *f;
	int p,y,x;
	uchar c,*dm1,*dm2;

	// a PC # that looks like abcde
	// appears as edcba
	// and is encoded as debc-a

	static char idstr[]={
		'H','P','H','P','4','8','-','E',
//		0x02b1e
//		6543
//		192
//		131
//		e1b20 f8910 0c000 38000
//		1e 2b f0 98 01 c0 00 30 08 00

		0x1e,0x2b,0xf0,0x98,0x01,0xc0,0x00,0x30,0x08,0x00

	};

	if (!name) return PIOERR_NONAME;
	if (!*name) return PIOERR_NONAME;

	if (bm->width < 131) return PIOERR_BADSIZE;
	if (bm->height< 64) return PIOERR_BADSIZE;

	if (!(f=VPfopen(name,"wb"))) return PIOERR_OPEN;

	if (1!=fwrite(idstr,18,1,f)) goto writerr;

	// write out grob frames one by one

	for(p=4;p;p>>=1) {
		dm1 = bm->vm;
		for(y=0;y<64;y++) {
			dm2=dm1; dm1+=bm->bpr;
			for(x=0;x<16;x++) {	// process 16 bytes
				c=255;
				if (dm2[0]&p) c-=1;
				if (dm2[1]&p) c-=2;
				if (dm2[2]&p) c-=4;
				if (dm2[3]&p) c-=8;
				if (dm2[4]&p) c-=16;
				if (dm2[5]&p) c-=32;
				if (dm2[6]&p) c-=64;
				if (dm2[7]&p) c-=128;
				dm2+=8;
				fputc(c,f);
			}
			// add extra 3 pixels
			c=0;
			if (dm2[0]&p) c+=1;
			if (dm2[1]&p) c+=2;
			if (dm2[2]&p) c+=4;
			fputc(c,f);
		}
	}
	fclose(f);
	return PIOERR_NONE;
writerr:
	fclose(f);
	return PIOERR_WRITE;
}


/////////////// AUXILIARY ROUTINES /////////////////

static int RP_Compare(ulong *s,ulong *t) {
	ulong s1=*s,t1=*t;
	return (s1<t1 ? -1 : (s1>t1 ? 1 : 0));
}

void	ReducePalette(short dstcols,uchar *remap,uchar *newpal) {
	ulong histomem[512];
	short x;
	ulong *t,*s;
	uchar *u,*v;

	Histogram(histomem,picbm);

	s=histomem+256;
	t=histomem+512;
	u=piccol+765;
	for(x=256;x;x--) {
		*--t=*(ulong *)u;
		*--t=*--s;
		u-=3;
	};
	qsort(histomem,256,8,RP_Compare);

	v=(uchar *)(histomem+1);
	u=(uchar *)histomem;
	for(x=dstcols;x;x--) {
		*v++=*u++;
		*v++=*u++;
		*v++=*u++;
		v+=5;
	};

	vid_make_remap(piccol,(uchar *)histomem,remap,256,dstcols);
	memcpy(newpal,(uchar *)histomem,dstcols*3);
}

//	void ChunkyToPlanar(uchar *destmem,uchar *srcmem,ulong bytes,short plane);
//
//	Extracts a single bit position out of every byte of a chunky byte-per-pixel
//	memory area into a planar bit-per-pixel memory area.

void	ChunkyToPlanar(uchar *destmem,uchar *srcmem,ulong bytes,short plane) {
	uchar c;
	do {
		c=0;
		if (*srcmem++ & plane) c+=128;
		if (*srcmem++ & plane) c+=64;
		if (*srcmem++ & plane) c+=32;
		if (*srcmem++ & plane) c+=16;
		if (*srcmem++ & plane) c+=8;
		if (*srcmem++ & plane) c+=4;
		if (*srcmem++ & plane) c+=2;
		if (*srcmem++ & plane) c+=1;
		*destmem++=c;
	} while(--bytes);
}

//	void PlanarToChunky(uchar *destmem,uchar *srcmem,ulong bytes,short plane);
//
//	Inserts all bits of a single line-plane into a certain bit position of
//	a byte-per-pixel memory area.

void	PlanarToChunky(uchar *destmem,uchar *srcmem,ulong bytes,short plane) {
	uchar c;
	do {
		c=*srcmem++;
		if (c & 128) destmem[0]|=plane;
		if (c &  64) destmem[1]|=plane;
		if (c &  32) destmem[2]|=plane;
		if (c &  16) destmem[3]|=plane;
		if (c &   8) destmem[4]|=plane;
		if (c &   4) destmem[5]|=plane;
		if (c &   2) destmem[6]|=plane;
		if (c &   1) destmem[7]|=plane;
		destmem+=8;
	} while(--bytes);
}

