#include <malloc.h>
#include <string.h>

#include "picture.h"
#include "video.h"
#include "pcx.h"
#include "request.h"

static int picture_new_number = 1;

// PictureInit: Initialize Picture structure

boolean	PictureInit(struct Picture *pict, struct BitMap *bm, long width, long height) {

	// try to allocate 1st frame, palette, and palbits structures, and image memory

	if (pict->frame_buf = malloc(vid13_rassize(width,height))) {
		if (pict->first = malloc(sizeof(struct Frame))) {
			if (pict->palette = malloc(768+32)) {
				vid_initBitMap(bm,pict->frame_buf,0,0,width,height,vid13_bpr(width),pict->palette);
				pict->bm		= bm;
				pict->last		= pict->first;
				pict->current		= pict->first;
				pict->current_num	= 1;
				pict->total_frames	= 1;
				pict->playback_speed	= 0x1111;	// very close to 15fps
				pict->pal_bits		= pict->palette + 768;
				pict->cur_pal		= pict->palette;
				pict->pack_buf		= NULL;
				pict->maxframe		= width * height;
				pict->flags		= NULL;
				pict->palettes		= NULL;
				pict->palsets		= NULL;
				pict->current_palset	= PALSET_DEFAULT;
				pict->dirty		= TRUE;		// this frame has no memory
				pict->inited		= TRUE;
				pict->first->next	= NULL;
				pict->first->prev	= NULL;
				pict->first->frame_mem	= NULL;
				pict->first->frame_pal	= NULL;
				pict->first->compressed	= FALSE;
				pict->name_ptr		= NULL;

				memset(pict->pal_bits,0xff,32);
				memset(pict->cyclerange,0,sizeof(struct CycleRange [8]));

				strcpy(pict->name,"Unknown");
				itoa(picture_new_number++,pict->name+7,10);

				return TRUE;
			}
			free(pict->first);
		}
		free(pict->frame_buf);
	}
	return FALSE;
}

//	void	PictureDispose(struct Picture *pict);
//
//	Free all memory belonging to a picture, but not the picture structure itself.

void	PictureDispose(struct Picture *pict) {
	struct Frame *frame,*next;

	free(pict->palette);
	free(pict->frame_buf);

	next=pict->first;
	while(frame=next) {
		next = frame->next;
		free(frame->frame_mem);
		if (frame->frame_pal) DetachFramePalette(pict,frame);
		free(frame);
	}

	pict->inited=FALSE;
}


//	boolean	success = PictureReinit(struct Picture *pict,long width,long height);
//
//	Frees all memory belonging to a picture and re-initializes it as a different size.
//
//	Returns TRUE for success, FALSE for failure.

boolean	PictureReinit(struct Picture *pict,long width,long height) {
	struct BitMap *bmsave;
	uchar *palette;

	bmsave = pict->bm;
	palette= pict->palette; pict->palette=NULL;
	PictureDispose(pict);
	if (!PictureInit(pict,bmsave,width,height)) {
		free(palette);
		return FALSE;
	}
	memcpy(pict->palette,palette,768);
	free(palette);
	return TRUE;
}

//	void	PictureSetName(struct Picture *pict,char *name);
//
//	Set a picture's name.

void	PictureSetName(struct Picture *pict,char *name) {
	char *slash = pict->name;
	char c,*s = pict->name;

	strcpy(pict->name,name);
	while(c=*s++) if (c==':' || c=='\\') slash=s;
	pict->name_ptr = slash;
	pict->flags |= PICT_HASNAME;
}


//	boolean success = PictureFlush(struct Picture *pict);
//
//	Checks to see if a picture needs to have its current frame flushed and
//	flushes it if it needs to.
//
//	Returns FALSE if the flush fails, most likely due to a lack of memory.

boolean	PictureFlush(struct Picture *pict) {
	ulong	size;
	void	*newmem;

	if (!pict->dirty) return TRUE;		// doesn't need to be flushed

	if (!pict->pack_buf) {
		size = PredictCompressedFrame(pict->frame_buf,pict->bm->width,pict->bm->height,pict->bm->bpr);
		if (!(newmem = realloc(pict->current->frame_mem,(size>=pict->maxframe) ? pict->maxframe : size))) return FALSE;
		if (size>=pict->maxframe) {
			CopyToFrame(newmem,pict->frame_buf,pict->bm->width,pict->bm->height,pict->bm->bpr);
			pict->current->compressed = FALSE;
			pict->current->frame_siz = pict->maxframe;
		} else {
			CompressFrame(newmem,pict->frame_buf,pict->bm->width,pict->bm->height,pict->bm->bpr,pict->maxframe);
			pict->current->compressed = TRUE;
			pict->current->frame_siz = size;
		}
	} else {
		size = CompressFrame(pict->pack_buf,pict->frame_buf,pict->bm->width,pict->bm->height,pict->bm->bpr,pict->maxframe);
		if (!size) {
			if (!(newmem = realloc(pict->current->frame_mem,pict->maxframe))) return FALSE;
			CopyToFrame(newmem,pict->frame_buf,pict->bm->width,pict->bm->height,pict->bm->bpr);
			pict->current->compressed = FALSE;
			pict->current->frame_siz = pict->maxframe;
		} else {
			if (!(newmem = realloc(pict->current->frame_mem,size))) return FALSE;
			memcpy(newmem,pict->pack_buf,size);
			pict->current->compressed = TRUE;
			pict->current->frame_siz = size;
		}
	}
	pict->current->frame_mem = newmem;
	pict->dirty = FALSE;
	return TRUE;
}

//	void PictureDecompress(struct Picture *pict);
//
//	Decompress the current frame of a picture into the picture's image
//	buffer.

void	PictureDecompress(struct Picture *pict) {
//	Message("PictureDecompress start");
	PictureDecompressTo(pict,pict->current,pict->frame_buf);
//	Message("PictureDecompress end");
}

//	void PictureDecompressTo(struct Picture *pict,struct Frame *frame,uchar *buff);
//
//	Decompress the current frame of a picture into the given image buffer
//
//	WARNING: Be sure that the frame buffer matches the current frame before
//		 resuming normal paint operation or the frame may be corrupted.

void	PictureDecompressTo(struct Picture *pict,struct Frame *frame,uchar *buff) {
	if (pict->current->compressed) {
		Frame_Decompress(buff,frame->frame_mem,pict->bm->width,pict->bm->height,pict->bm->bpr);
//		Frame_Decompress(buff,frame->frame_mem,pict->bm->width,512,pict->bm->bpr);
	} else
		CopyFromFrame(buff,frame->frame_mem,pict->bm->width,pict->bm->height,pict->bm->bpr);
}

//	boolean	success = PictureGotoFrame(struct Picture *pict, struct Frame *frame, int frame_num);
//
//	Switch a picture to a particular frame.
//
//	Returns FALSE if the switch was unsuccessful.
//
//	The gods will spite you if you use a frame from another picture...

boolean	PictureGotoFrame(struct Picture *pict, struct Frame *frame, int frame_num) {
	if (!PictureFlush(pict)) return FALSE;
	pict->current		= frame;
	pict->current_num	= frame_num;
	pict->cur_pal		= pict->current->frame_pal ? pict->current->frame_pal->palette : pict->palette;
	PictureDecompress(pict);
	return TRUE;
}

//	boolean success = PictureNextFrame(struct Picture *pict);
//
//	Advance a picture to the next frame, looping back to #1 if we're at the end.

boolean	PictureNextFrame(struct Picture *pict) {
	if (!PictureFlush(pict)) return FALSE;

	if (pict->current->next) {
		pict->current		= pict->current->next;
		pict->current_num++;
	} else {
		pict->current		= pict->first;
		pict->current_num	= 1;
	}
	pict->cur_pal		= pict->current->frame_pal ? pict->current->frame_pal->palette : pict->palette;
	PictureDecompress(pict);
	return TRUE;
}

//	boolean success = PicturePrevFrame(struct Picture *pict);
//
//	Advance a picture to the previous frame, looping to the last frame if
//	we're at the start.

boolean	PicturePrevFrame(struct Picture *pict) {
	if (!PictureFlush(pict)) return FALSE;

	if (pict->current->prev) {
		pict->current		= pict->current->prev;
		pict->current_num--;
	} else {
		pict->current		= pict->last;
		pict->current_num	= pict->total_frames;
	}
	pict->cur_pal		= pict->current->frame_pal ? pict->current->frame_pal->palette : pict->palette;
	PictureDecompress(pict);
	return TRUE;
}

//	void PictureMarkDirty(struct Picture *pict);
//
//	Mark the current frame of a picture as dirty.

void	PictureMarkDirty(struct Picture *pict) {
	pict->dirty = TRUE;
}

//	void PictureInvalidate(struct Picture *pict);
//
//	Mark the current frame of a picture as invalidated (not dirty).  This fools the Picture
//	routines in believing that the frame has not been modified, so that there is no reason
//	to recompress the frame.
//
//	This means that the frame's compressed memory is valid but the frame buffer is not.
//	You had better have data in the frame's memory before calling this, or PictureDecompress()
//	will bomb when you try to switch back to this frame.
//
//	This function cannot fail.

void	PictureInvalidate(struct Picture *pict) {
	pict->dirty = FALSE;
}

///////////////////////////////////////////////////////////////////////////////////////
//
//	frame format:
//
//	array [1...height] {
//		either:
//			80,nn	fill the line with color <nn>
//		or array of {		(this is IFF compression)
//			81-ff	run
//			00-7f	literal
//		} until the entire line is filled
//	}

ulong	PredictCompressedFrame(uchar *mem,long width,long height,long bpr) {
 	uchar *t;
	long	x;
	ulong size=0;

	while(height--) {
		t=mem;		mem+=bpr;
		for(x=1;x<width;x++)
			if (t[x]!=t[0]) break;
		if (x==width) size+=2;
		else size+=IFF_Line_Compress_Predict(t,width);
	}
	return size;
}

//	ulong size = CompressFrame(uchar *pack_buf, uchar *frame, long width, long height, long bpr, long maxsize);
//
//	Compresses a frame to a pack buffer, returning the size of the compressed
//	frame, or zero if the frame could not be compressed.

ulong	CompressFrame(uchar *pack_buf,uchar *frame,long width,long height,long bpr, long maxsize) {
 	uchar *t;
	long	x;
	ulong	size=maxsize;
	ulong	danger_pt = IFF_Line_Max(width);

	while(height--) {
		t=frame;   	frame+=bpr;
		for(x=1;x<width;x++)
			if (t[x]!=t[0]) break;
		if (x==width) {
			if (size<2) return 0;
			size -= 2;
			*pack_buf++ = 0x80;
			*pack_buf++ = t[0];
		} else {
			if (size<danger_pt) {
				x = IFF_Line_Compress_Cautious(pack_buf,t,width,size);
				if (!x) return 0;
				size -= x;
				pack_buf += x;
			} else {
				x = IFF_Line_Compress(pack_buf,t,width);
				size -= x;
				pack_buf += x;
			}
		}
	}

	return maxsize - size;
}

//	void CopyToFrame(uchar *dest, uchar *src, long width, long height, long bpr);
//
//	Copies an image to a frame, deleting out the extra pad bytes in
//	the bitmap.

void	CopyToFrame(uchar *dest, uchar *src, long width, long height, long bpr) {
	while(height--) {
		memcpy(dest,src,width);
		dest += width;
		src  += bpr;
	}
}

//	void CopyFromFrame(uchar *dest, uchar *src, long width, long height, long bpr);
//
//	Copies an image from a frame, inserting the extra pad bytes into
//	the bitmap.

void	CopyFromFrame(uchar *dest, uchar *src, long width, long height, long bpr) {
	while(height--) {
		memcpy(dest,src,width);
		dest += bpr;
		src  += width;
	}
}

//	boolean InsertBeforeFrame(struct Picture *pict);
//
//	Insert a new frame immediately before the current one.
//	Also switches to the new frame.
//
//	Returns TRUE if successful, FALSE if error.

boolean	InsertBeforeFrame(struct Picture *pict) {
	struct Frame *fr;

	if (!PictureFlush(pict)) return FALSE;

	if (!(fr=malloc(sizeof(struct Frame)))) return FALSE;

	fr->next		= pict->current;
	fr->prev		= pict->current->prev;
	if (pict->current->prev) {	// there is a previous frame
		pict->current->prev->next = fr;
	} else {			// this is the first frame
		pict->first	= fr;
	}
	pict->current->prev	= fr;
	pict->current		= fr;
	pict->dirty		= TRUE;		// this frame needs to be created
	pict->cur_pal		= pict->palette;

	fr->frame_mem		= NULL;
	fr->frame_pal		= NULL;

	++pict->total_frames;

	return TRUE;
}

//	boolean InsertAfterFrame(struct Picture *pict);
//
//	Insert a new frame immediately before the current one.
//	Also switches to the new frame.
//
//	Returns TRUE if successful, FALSE if error.

boolean	InsertAfterFrame(struct Picture *pict) {
	struct Frame *fr;

	if (!PictureFlush(pict)) return FALSE;

	if (!(fr=malloc(sizeof(struct Frame)))) return FALSE;

	fr->prev		= pict->current;
	fr->next		= pict->current->next;
	if (pict->current->next) {	// there is a next frame
		pict->current->next->prev = fr;
	} else {			// this is the first frame
		pict->last	= fr;
	}
	pict->current->next	= fr;
	pict->current		= fr;
	pict->dirty		= TRUE;		// this frame needs to be created
	pict->cur_pal		= pict->palette;

	fr->frame_mem		= NULL;
	fr->frame_pal		= NULL;

	++pict->total_frames;
	++pict->current_num;

	return TRUE;
}

//	boolean	DeleteCurrentFrame(struct Picture *pict);
//
//	Deletes the current frame and switches to a new frame.
//
//	If the last frame is deleted, the next-last frame is loaded.
//	Else the next frame that drops into the same # (the next one)
//	is loaded.
//
//	Returns TRUE for success, FALSE for failure.  Will fail if only
//	one frame exists.

boolean	DeleteCurrentFrame(struct Picture *pict) {
	struct Frame *frame;

	// switch to the new frame

	frame = pict->current;
	if (pict->current_num == pict->total_frames) {
		if (!PicturePrevFrame(pict)) return FALSE;
	} else {
		if (!PictureNextFrame(pict)) return FALSE;
		--pict->current_num;
	}

	// unlink frame

	if (frame->next) frame->next->prev = frame->prev; else pict->last = frame->prev;
	if (frame->prev) frame->prev->next = frame->next; else pict->first= frame->next;
	free(frame->frame_mem);
	free(frame->frame_pal);		// free(0) is safe
	free(frame);

	if (!--pict->total_frames) {
		pict->total_frames = 1;
		PictureMarkDirty(pict);
	}

	return TRUE;
}

//	struct FramePalette *CreateFramePalette(void);
//
//	Create a brand new framepalette.

struct FramePalette *CreateFramePalette(void) {
	struct FramePalette *fpal;

	if (!(fpal = malloc(sizeof(struct FramePalette)))) return NULL;

	fpal->usecount = 0;
	fpal->next = fpal->prev = NULL;

	return fpal;
}

//	void AttachPaletteToPicture(struct Picture *pict,struct FramePalette *fpal);
//
//	Attach a framepalette to a picture.

void AttachPaletteToPicture(struct Picture *pict,struct FramePalette *fpal) {
	fpal->next = pict->palettes;
	fpal->prev = NULL;
	fpal->next->prev = fpal;
	pict->palettes = fpal;
}

//	void DetachPaletteFromPicture(struct Picture *pict,struct FramePalette *fpal);
//
//	Detach a framepalette from a picture.

void DetachPaletteFromPicture(struct Picture *pict,struct FramePalette *fpal) {
	if (fpal->prev)	fpal->prev->next = fpal->next;
	else 		pict->palettes = fpal->next;
	if (fpal->next)	fpal->next->prev = fpal->prev;
}

//	boolean success = AttachPaletteToFrame(struct Picture *pict,struct Frame *frame,struct FramePalette *pal);
//
//	Allocate memory for a frame palette and attach it to the indicated frame.
//
//	Returns TRUE for success, FALSE for failure.

boolean AttachPaletteToFrame(struct Picture *pict,struct Frame *frame,struct FramePalette *pal) {
	if (pal) {
		if (frame->frame_pal == pal) return TRUE;
		++pal->usecount;
	} else {
		if (!(pal = malloc(sizeof(struct FramePalette)))) return FALSE;
		pal->usecount = 1;
		memcpy(pal->palette,pict->palette,768);
		pal->next = pict->palettes;
		pal->prev = NULL;
		pal->next->prev = pal;
		pict->palettes = pal;
	}
	frame->frame_pal = pal;
	return TRUE;
}

//	void DetachFramePalette(struct Picture *pict,struct Frame *frame);
//
//	Detaches a frame from a palette.  If the palette's usecount drops to
//	zero, the palette is freed.
//
//	Safe to call with a frame that has no palette.

void	DetachFramePalette(struct Picture *pict,struct Frame *frame) {
	if (!frame->frame_pal) return;
	if (!--frame->frame_pal->usecount) {
		if (frame->frame_pal->prev)	frame->frame_pal->prev->next = frame->frame_pal->next;
		else				pict->palettes = frame->frame_pal->next;
		if (frame->frame_pal->next)	frame->frame_pal->next->prev = frame->frame_pal->prev;
		free(frame->frame_pal);
	}
	frame->frame_pal = NULL;
}

//////////////////////////
//
//	void LoadFramePalette(struct Picture *pict,struct Frame *frame);
//
//	Loads the palette from a frame into the VGA.
//

void	LoadFramePalette(struct Picture *pict,struct Frame *frame) {
	if (frame->frame_pal)
		vid_loadRGB6(0,frame->frame_pal->palette,256);
	else
		vid_loadRGB6(0,pict->palette,256);
}

//	boolean AllocateFrameBuffer(struct Frame *frame,ulong siz,boolean compressed);
//
//	Resize the compressed memory for a frame so that compressed/uncompressed data
//	can be loaded directly into a frame.
//
//	If the frame is the current frame, make sure to either call PictureDecompress()
//	to decompress it to the frame buffer or call PictureInvalidate() to invalidate
//	the frame buffer.
//
//	The original frame memory is unchanged upon failure.
//
//	Returns TRUE for success and FALSE upon failure.

boolean AllocateFrameBuffer(struct Frame *frame,ulong siz,boolean compressed) {
	void *newmem;

	if (!(newmem = realloc(frame->frame_mem,siz))) return FALSE;

	frame->frame_mem = newmem;
	frame->frame_siz = siz;
	frame->compressed = compressed;

	return TRUE;
}

//	struct Frame *GetFrameAddress(struct Picture *pict,int frame_num);
//
//	Find the numbered frame in the picture and return its address.
//
//	Returns NULL if the frame does not exist.

struct Frame *GetFrameAddress(struct Picture *pict,int frame_num) {
	struct Frame *fp;

	if (frame_num<=0 || frame_num>pict->total_frames) return NULL;

	if (frame_num==1) return pict->first;

	fp = pict->first;
	while(--frame_num)
		fp=fp->next;

	return fp;
}


////////////////////

//	boolean PictureProcessFrames(struct Picture *pict,boolean (*func)(struct Picture *pict));
//
//	Sequentially process all frames in a picture.
//
//	Returns TRUE if successful, FALSE upon failure.

boolean	PictureProcessFrames(struct Picture *pict,boolean (*func)(struct Picture *pict),char *sb_title) {
	int i;
	struct StatusBar *sb = NULL;

	if (sb_title) sb=CreateStatusBar(pict->total_frames,sb_title);

	for(i=0;i<pict->total_frames;i++) {
		if (!func(pict)) return FALSE;
		PictureMarkDirty(pict);
		PictureNextFrame(pict);
		ModifyStatusBar(sb,i+1);
	}
	DestroyStatusBar(sb);
	return TRUE;
}

