#include <math.h>
#include <assert.h>

#include "print.h"
#include "request.h"
#include "listbox.h"

#include "printini.c"

static void Printer_Recalc_Sizes(void);
static void PrintRequestUpdater(struct Slider *sld);
static void PrintRequestUpdater2(struct Slider *sld);
static void Printer_Make_Gamma(uchar *xlat,double gamma);
boolean	Print_Write(FILE *f,uchar *buf,ulong bytes);

///////////// CONFIGURATION VARIABLES ///////////////

#define	PRINT_GREY	0
#define	PRINT_COLOR	1

#define	PRINT_DITHER_NONE	0
#define	PRINT_DITHER_ORDERED	1
#define	PRINT_DITHER_FS		2

static short print_spectrum=PRINT_GREY;		// spectrum
static short print_density=0;			// density
static float print_gamma=1.000;			// gamma correction
static uchar print_gamma_translate[256];	// gamma translation
static short print_dither=0;			// dithering mode

static long	print_width;			// width of printed bitmap in printer pixels
static long	print_height;			// height of printed bitmap in printer pixels

static struct	IniItem	*print_iniden;		// IniItems for current density

static struct Rectangle print_pos[]={
	{  4, 86, 42,13 },
	{ 66, 86, 42,13 },
	{  2,102,196,13 },
	{RECT_END}
};

static struct Rectangle print_neg[]={
	{ 70, 60,120,11 },
	{  5, 87, 40,11 },
	{ 67, 87, 40,11 },
	{  3,103,194,11 },
	{ RECT_END }
};

static struct Text print_text[]={
	{ "Spectrum"	,  4, 14,St0,St2 },
	{ "Dithering"	,  4, 26,St0,St2 },
	{ "Position"	,  4, 38,St0,St2 },
	{ "by"		, 50, 89,St0,St2 },
	{ NULL }
};

static struct Zone print_zn1[]={
	{   4,135, 48, 23,ZN_3D+ZN_TEXTRENDER			,NULL,"Print!"	,NULL },
	{ 148,135, 48, 23,ZN_3D+ZN_TEXTRENDER			,NULL,"Cancel"	,NULL },
	{ -1,-1,-1,-1 }
};

static struct Zone print_zn2[]={
	{  64, 12, 30, 11,ZN_3D+ZN_TEXTRENDER+ZN_INVERTED	,NULL,"Grey"	,NULL },
	{  94, 12, 36, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"Color"	,NULL },
	{-1,-1,-1,-1}
};

static struct Zone print_zn3[]={
	{  64, 24, 30, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"None"	,NULL },
	{  94, 24, 48, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"Ordered"	,NULL },
	{ 142, 24, 24, 11,ZN_3D+ZN_TEXTRENDER+ZN_INVERTED	,NULL,"F-S"	,NULL },
	{-1,-1,-1,-1}
};

static struct Zone print_zn4[]={
	{  64, 36, 30, 11,ZN_3D+ZN_TEXTRENDER+ZN_INVERTED	,NULL,"Left"	,NULL },
	{  94, 36, 42, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"Center"	,NULL },
	{ 136, 36, 36, 11,ZN_3D+ZN_TEXTRENDER			,NULL,"Right"	,NULL },
	{-1,-1,-1,-1}
};

static struct ZoneRange print_zr[]={
	{ &print_zr[1]	,print_zn1,NULL		,-1,ZR_TAP	},
	{ &print_zr[2]	,print_zn2,&print_zn2[0], 0,ZR_CHOICE	},
	{ &print_zr[3]	,print_zn3,&print_zn3[2], 2,ZR_CHOICE	},
	{ NULL		,print_zn4,&print_zn4[0], 0,ZR_CHOICE	},
};

static struct Slider print_sld[]={
	{
		&print_sld[1],{NULL},NULL,{ 17,62,28,8 },
		NULL,0,0,4,4,
		0,0,7,0,
		1,1,0,0,0,0,
		REQSLD_ONEHIT+SLD_ARROW1HIT+SLD_ARROW2HIT+SLD_ARROW1*4+SLD_ARROW2*6+SLD_OUTLINE3D,
		&PrintRequestUpdater,
		{  4,60,11,12 },
		{ 47,60,11,12 },
		-1,0,1,0,NULL
	},{
		NULL,{NULL},NULL,{ 17,74,28,8 },
		NULL,0,0,4,4,
		0,0,512,0,
		16,16,0,0,0,0,
		REQSLD_ONEHIT+SLD_ARROW1HIT+SLD_ARROW2HIT+SLD_ARROW1*4+SLD_ARROW2*6+SLD_OUTLINE3D,
		&PrintRequestUpdater2,
		{  4,72,11,12 },
		{ 47,72,11,12 },
		-1,0,1,0,NULL
	}
};

static char print_fld_width_buf[7];
static char print_fld_height_buf[7];
static char print_output_name[160]="lpt1out";

static struct Field print_fld[]={
	{
		7,89,36,8,NULL,print_fld_width_buf,
		0,0,sizeof print_fld_width_buf-1,6,St0,St2,FLD_NUMERIC,1,32767,640
	},{
		69,89,36,8,NULL,print_fld_height_buf,
		0,0,sizeof print_fld_height_buf-1,6,St0,St2,FLD_NUMERIC,1,32767,480
	},{
		4,104,192,8,NULL,print_output_name,
		0,0,sizeof print_output_name-1,32,St0,St2,NULL,0,0,0
	},{	-1,-1,-1,-1 }
};

static struct Requester print_req={
	0,0,200,160,
	REQ_BORDER+REQ_POPCEN,
	"Printing options",
	St2,
	NULL,		// listboxes
	print_fld,	// fields
	print_zr,	// zoneranges
	print_sld,	// sliders
	print_pos,	// pos boxes
	print_neg,	// neg boxes
	print_text,	// texts
	NULL
};

boolean	PrintRequest(void) {
	mouse_coord mx,my,mb;
	boolean exit=FALSE;
	boolean result;

	print_iniden=print_ini_density[print_density];
	Printer_Recalc_Sizes();

	if (!OpenRequester(&print_req,0,0)) return FALSE;

	SliderAbs(&print_sld[1],256,0);

	while(!exit) {
		mouse_read2(&mx,&my,&mb);
		if (mb)	{
			ulong msg;

			switch(msg=CheckRequester(&print_req,mx,my,mb)) {
			case RZoneMsg(0,0):
				result=TRUE;
				exit=TRUE;
				break;
			case RZoneMsg(0,1):
				result=FALSE;
				exit=TRUE;
				break;
			default:
				if ((msg & 0xFFFF0000)==REQM_SLIDER) {
					print_iniden=print_ini_density[print_sld[0].cposx];
					Printer_Recalc_Sizes();
					DisplayField(&print_fld[0]);
					DisplayField(&print_fld[1]);
				};
				break;
			};
		};
	};
	mouse_wait_unkey();
	CloseRequester(&print_req);

	print_spectrum	= print_zr[1].highindex;
	print_dither	= print_zr[2].highindex;
	print_density	= print_sld[0].cposx;

	print_width	= atoi(print_fld_width_buf);
	print_height	= atoi(print_fld_height_buf);

	Printer_Make_Gamma(print_gamma_translate,print_gamma);

	return result;
}

static void Printer_Recalc_Sizes(void) {
	print_width=(long)print_iniden[PRINT_INIDEN_TOTALWIDTH].ivalue;
//	print_height=((long)print_iniden[PRINT_INIDEN_RESHEIGHT].ivalue*picbm->height)/200;

//	<total printer width> * <picture height> * <vert resolution> / <picture width> / <horiz resolution>

	print_height=(
		  (long)print_iniden[PRINT_INIDEN_TOTALWIDTH].ivalue
		* (long)print_iniden[PRINT_INIDEN_RESHEIGHT].ivalue
		* picbm->height)
		/ ((long)print_iniden[PRINT_INIDEN_RESWIDTH].ivalue
		  * picbm->width);
	ltoa(print_width ,print_fld_width_buf ,10);
	ltoa(print_height,print_fld_height_buf,10);
}

static void PrintRequestUpdater(struct Slider *sld) {
	vd->bar(&print_req.reqbm,print_neg[0].x+1,print_neg[0].y+1,print_neg[0].width-2,print_neg[0].height-2,print_req.backfill);
	vd->text6(&print_req.reqbm,print_neg[0].x+2,print_neg[0].y+2,(char*)print_ini_density[sld->cposx][PRINT_INIDEN_NAME].ivalue,ST0,ST2);
}

static void PrintRequestUpdater2(struct Slider *sld) {
	print_gamma=pow(2.0,2.0-(sld->cposx/128.0));
	vid_printf6(&print_req.reqbm,72,74,ST0,ST2,"%5.3f gamma corr.",print_gamma);
}

static void Printer_Make_Gamma(uchar *xlat,double gamma) {
	double x,xt;
	short y;

	for(x=0;x<=1;x+=(1.0/255.0)) {
		xt=x;
		if (xt<0) xt=0; else if (xt>=256) xt=255.9;
		y=ceil(pow(x,gamma)*255.0+0.5);
		if (y<0) y=0; else if (y>255) y=255;
		*xlat++=y;
	};
}

//	void Print_StepScale(void);
//
//	Continue scaling the current picture for a single destination row.  This routine
//	does a hybrid scale/step operation in a single step; that way, the printer can
//	print better pictures than the monitor (this can happen with a high-resolution
//	printer and interpolative scaling).
//
//	Code shamelessly stolen from video.c, since it's my code...

static long	print_yacc;			// step scaling y accumulator
static long	print_scwidth;			// scaling width (width unclipped to printer boundaries)
static boolean	print_compress_h;		// obvious
static boolean	print_compress_v;
static boolean	print_expand_h;
static boolean	print_expand_v;
static uchar *	print_scptr;			// step scaling pointer to picture bitmap
static uchar *	print_palette=NULL;		// optimized palette (1024 bytes instead of 768)
static char	print_pins;			// pins to use (either 8 or 24)

static uchar *	print_redmem=NULL;		// red buffer memory
static uchar *	print_grnmem=NULL;		// green buffer memory
static uchar *	print_blumem=NULL;		// blue buffer memory
static uchar *	print_blkmem=NULL;		// black buffer memory;

static uchar *	print_redptr;			// red buffer pointer
static uchar *	print_grnptr;			// green buffer pointer
static uchar *	print_bluptr;			// blue buffer pointer
static uchar *	print_blkptr;			// black buffer pointer

static uchar *	print_outmem=NULL;		// output buffer

static short *	print_fsmem=NULL;		// memory for Floyd-Steinberg dithering
static short *	print_fseven[3];		// Floyd-Steinberg current row pointers
static short *	print_fsodd[3];			// Floyd-Steinberg next row pointers
static uchar	print_ordered[16][16];		// Ordered array

static FILE *	print_file;			// file ptr

static struct StatusBar *print_sb;		// status bar

static void Print_StepScale(uchar *red,uchar *grn,uchar *blu) {
	long xacc,xc;
	uchar *src;
	short	c;

	if (print_compress_v) {	/* contract */
		while(print_yacc>0) {
			print_yacc-=print_height;
			print_scptr+=picbm->bpr;
		};
	};
	src=print_scptr;
	xc=print_scwidth;
	xacc=0;
	if (print_expand_h)	/* scale up */
		do {
			if (xacc>=0) {
				c=(short)(*src++)<<2;
				xacc-=print_scwidth;
			};
			xacc+=picbm->width;
			*red++=print_palette[c];
			*grn++=print_palette[c+1];
			*blu++=print_palette[c+2];
//			*blk++=print_palette[c+3];
		} while(--xc);
	else if (print_compress_h)	/* scale down */
		do {
			while(xacc>0) {
				xacc-=print_scwidth;
				++src;
			};
			c=((short)*src)<<2;
			*red++=print_palette[c];
			*grn++=print_palette[c+1];
			*blu++=print_palette[c+2];
//			*blk++=print_palette[c+3];
			xacc+=picbm->width;
		} while(--xc);
	else	do { 
			c=((short)*src++)<<2;
			*red++=print_palette[c];
			*grn++=print_palette[c+1];
			*blu++=print_palette[c+2];
//			*blk++=print_palette[c+3];
		} while(--xc);
	print_yacc+=picbm->height;
	if (print_expand_v) {	/* expand */
		while(print_yacc>=print_height) {
			print_yacc-=print_height;
			print_scptr+=picbm->bpr;
		};
	};
}

////////////////////////////////////////////////////////////////////
//	void	Print_Make_Palette(void);
//
//	Makes an appropriate 1K printer palette from a 768 byte picture palette.

void Print_Make_Palette(void) {
	long x;
	uchar *src,*dst;
	uchar r,g,b,i;
	long c;

	src=piccol;
	dst=print_palette;

	if (print_spectrum==PRINT_COLOR)
		for(x=256;x;x--) {
			r=print_gamma_translate[src[0]];
			g=print_gamma_translate[src[1]];
			b=print_gamma_translate[src[2]];
			i=print_gamma_translate[((long)(30L*src[0]+59L*src[1]+11L*src[2])<<8)/6301];
			if (!print_dither) {
				r=r<128;
				g=g<128;
				b=b<128;
				i=i<128;
			};
			*dst++=r;
			*dst++=g;
			*dst++=b;
			*dst++=i;
			src+=3;
		}
	else
		for(x=256;x;x--) {
			if (!print_dither)
				c=((30L*src[0]+59L*src[1]+11L*src[2])<12800);
			else	c=print_gamma_translate[((long)(30L*src[0]+59L*src[1]+11L*src[2])<<8)/25501];
			*dst++=c;
			*dst++=c;
			*dst++=c;
			*dst++=c;
			src+=3;
		};
}

////////////////////////////////////////////////////////////////////////
//	void	Print_fprintf(FILE *f,uchar *format);
//
//	Outputs a formatted string to the file indicated.
//
//	Special formats:
//		%%			Output a single percent character to the file
//		%[value][format]	Output a formatted value to the file - note that VALUE is first!
//			Formats:
//				1	One character ASCII zero padded
//				2	Two character ASCII zero padded
//				3	Three character ASCII zero padded
//				4	Four character ASCII zero padded
//				b	One character binary
//				B	Two character binary
//				R	Reversed two character binary
//			Values:
//				w	Width
//				h	Height

void	Print_fprintf(FILE *f,uchar *format) {
	short len;
	uchar c;
	long value,v;
	uchar obuf[256],*obufptr=obuf;

	if (!format) return;

	len=*format++;
	while(len--) {
		c=*format++;
		if (c=='%') {
			len-=2;
			if (len<0) return;
			switch(c=*format++) {
			case 'w':	value=print_width;	break;
			case 'h':	value=print_height;	break;
			};
			switch(c=*format++) {
			case '4':	v=value/1000; value-=(v*1000); if (v>10) v%=10; fputc(v+'0',f);
			case '3':	v=value/ 100; value-=(v* 100); if (v>10) v%=10; fputc(v+'0',f);
			case '2':	v=value/  10; value-=(v*  10); if (v>10) v%=10; fputc(v+'0',f);
			case '1':	v=value     ; value-= v      ; if (v>10) v%=10; fputc(v+'0',f);
					break;
			case 'B':	*obufptr++=(value>>8);
			case 'b':	*obufptr++=(value&0xFF);
					break;
			case 'R':	*obufptr++=(value&0xFF);
					*obufptr++=(value>>8);
					break;
			};
		} else *obufptr++ = c;
	};
	Print_Write(f,obuf,obufptr-obuf);
}

///////////////////////////////////////////////////////////////
//	void	Print_Make_Ordered_Array(uchar (*array)[16][16]);
//
//	Produces an appropriate Ordered dithering array matrix.

void	Print_Make_Ordered_Array(uchar (*array)[16][16]) {
	short x,y,xt,yt,s;
	uchar d;

	for(y=0;y<16;y++)
		for (x=0;x<16;x++) {
			d=0;
			xt=x;
			yt=y;
			for(s=4;s;s--) {
				d=(((d<<1) | ((xt&1)^(yt&1)))<<1) | (yt&1);
				xt>>=1;
				yt>>=1;
			};
			(*array)[y][x]=d;
		};
}

////////////////////////////////////////////////////////////////
//	boolean Print_Init(void);
//
//	Print initialization routine.

boolean Print_Init(void) {
	union REGS regs;

//	print_width  =(long)print_iniden[PRINT_INIDEN_RESWIDTH].value;	// 960
	print_scwidth=(long)print_width;
//	print_height =(long)print_iniden[PRINT_INIDEN_RESHEIGHT].value;	// 1200
	print_pins   =(long)print_iniden[PRINT_INIDEN_PINCOUNT].ivalue;	// 24

	if (!(print_palette	=malloc(1024				))) return Print_Cleanup();
	if (!(print_file	=VPfopen(print_output_name,"wb"		))) return Print_Cleanup();

	setvbuf(print_file,NULL,_IONBF,0);

	regs.w.ax=0x4400;		// reset the file to binary mode
	regs.w.bx=fileno(print_file);
	int386(0x21,&regs,&regs);

	if (!(regs.x.cflag & INTR_CF))
		if (regs.w.dx & 0x0080) {	// is it a device?
			regs.w.dx |= 0x0020;	// enable binary mode
			regs.w.bx=fileno(print_file);
			regs.w.ax=0x4401;
			int386(0x21,&regs,&regs);
		};

	if (!(print_redmem	=malloc(print_width*print_pins		))) return Print_Cleanup();
	if (print_spectrum==PRINT_COLOR) {
		if (!(print_grnmem	=malloc(print_width*print_pins		))) return Print_Cleanup();
		if (!(print_blumem	=malloc(print_width*print_pins		))) return Print_Cleanup();
		if (!(print_blkmem	=malloc(print_width*print_pins		))) return Print_Cleanup();
	};
	if (!(print_outmem	=malloc(print_width*(print_pins<<3)	))) return Print_Cleanup();

	if (print_spectrum==PRINT_COLOR) {
		if (!(print_fsmem	=malloc((print_width*2+4)*sizeof(short)*3	))) return Print_Cleanup();
	} else {
		if (!(print_fsmem	=malloc((print_width*2+4)*sizeof(short)		))) return Print_Cleanup();
	};

	if (print_spectrum==PRINT_COLOR) {
		print_redptr=print_redmem;
		print_grnptr=print_grnmem;
		print_bluptr=print_blumem;
		print_blkptr=print_blkmem;
	} else {
		print_redptr=print_grnptr=print_bluptr=print_blkptr=print_redmem;
	};
	print_fseven[0]=print_fsmem;
	print_fsodd[0]= print_fsmem+(print_width+2);
	print_fseven[1]=print_fsmem+(print_width+2)*2;
	print_fsodd[1]= print_fsmem+(print_width+2)*3;
	print_fseven[2]=print_fsmem+(print_width+2)*4;
	print_fsodd[2]= print_fsmem+(print_width+2)*5;
	memset(print_fsmem,0,(print_width*2+4)*sizeof(short)*((print_spectrum==PRINT_COLOR) ? 3 : 1));

	print_scptr=picmem;

	print_compress_h=	print_width < picbm->width;
	print_compress_v=	print_height< picbm->height;
	print_expand_h=		print_width > picbm->width;
	print_expand_v=		print_height>=picbm->height;

	print_yacc=0;

	Print_Make_Palette();
	Print_Make_Ordered_Array(print_ordered);

	return TRUE;
}

///////////////////////////////////////////////////////////////////
//	boolean Print_Cleanup(void);
//
//	Printing cleanup routine.
//
//	Always returns FALSE for convienence in Print_Init().

boolean	Print_Cleanup(void) {
	if (print_sb)		{ DestroyStatusBar(print_sb); print_sb=NULL;	}
	if (print_palette) 	{ free(print_palette);	print_palette=FALSE;	}
	if (print_file)		{ fclose(print_file);	print_file=NULL;	}
	if (print_redmem)	{ free(print_redmem);	print_redmem=NULL;	}
	if (print_grnmem)	{ free(print_grnmem);	print_grnmem=NULL;	}
	if (print_blumem)	{ free(print_blumem);	print_blumem=NULL;	}
	if (print_blkmem)	{ free(print_blkmem);	print_blkmem=NULL;	}
	if (print_outmem)	{ free(print_outmem);	print_outmem=NULL;	}
	return FALSE;
}

///////////////////////////////////////////////////////////////////
//	void	Print_Write(FILE *f,uchar *buf,ulong bytes);
//
//	Mashes bytes through the given file until all bytes have been written.
//
//	returns TRUE if error occurs

boolean	Print_Write(FILE *f,uchar *buf,ulong bytes) {
	int copied;

	while(bytes>0) {
		copied=fwrite(buf,1,bytes,f);
		kbhit();
		assert(copied>=0);
		bytes -= copied;
		buf += copied;
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////
//	void Print_Output(uchar *red,uchar *grn,uchar *blu);
//
//	Output the current rows of the translated picture to the printer.
//
//	At this moment this routine works only for Epson FX-80 printers at 120x72 density only.

void Print_Output_Swing(uchar *in,uchar *out,ulong bpr);
#pragma aux Print_Output_Swing=	\
	"mov al,[esi]",		\
	"add al,al",		\
	"add al,[esi+ebx]",	\
	"add al,al",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"add al,al",		\
	"add al,[esi+ebx]",	\
	"add al,al",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"add al,al",		\
	"add al,[esi+ebx]",	\
	"add al,al",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"add al,al",		\
	"add al,[esi+ebx]",	\
	"mov [edi],al",		\
	parm [esi] [edi] [ebx] modify exact [al esi];

void Print_Output_Swing_Rev(uchar *in,uchar *out,ulong bpr);
#pragma aux Print_Output_Swing_Rev=	\
	"mov al,[esi]",		\
	"ror al,1",		\
	"add al,[esi+ebx]",	\
	"ror al,1",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"ror al,1",		\
	"add al,[esi+ebx]",	\
	"ror al,1",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"ror al,1",		\
	"add al,[esi+ebx]",	\
	"ror al,1",		\
		"lea esi,[esi+ebx*2]",	\
	"add al,[esi]",		\
	"ror al,1",		\
	"add al,[esi+ebx]",	\
	"ror al,1",		\
	"mov [edi],al",		\
	parm [esi] [edi] [ebx] modify exact [al esi];

void	Print_Output_Line(uchar *clr,uchar *out,ulong bpr) {
	switch(print_ini_general[PRINT_INI_DATAFORMAT].ivalue) {
	case 0:	/* Epson format */
		Print_Output_Swing(clr,out,bpr);
		break;
	case 1: /* ImageWriter II format */
		Print_Output_Swing_Rev(clr,out,bpr);
		break;
	};
}

void	Print_Output_Color(uchar *clr) {
	long x;
	uchar *omem=print_outmem;
	long pbpr=print_width<<3;

	if (print_pins==8) {
		for(x=print_width;x;x--) {
			Print_Output_Line(clr++,omem++,print_width);
		};
	} else {		/* 24 pin */
		for(x=print_width;x;x--) {
			Print_Output_Line(clr		,omem++,print_width);
			Print_Output_Line(clr+pbpr	,omem++,print_width);
			Print_Output_Line(clr+pbpr*2	,omem++,print_width);
			++clr;
		};
	};
	Print_fprintf(print_file,(char*)print_iniden[PRINT_INIDEN_ROWSTART].ivalue);
//	if (1!=fwrite(print_outmem,print_width*(print_pins>>3),1,print_file))
//		Message("Write failure: %d",errno);
	Print_Write(print_file,print_outmem,print_width*(print_pins>>3));
};

void	Print_Output(uchar *red,uchar *grn,uchar *blu,uchar *blk) {
	Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_LINESTART].ivalue);
	Print_fprintf(print_file,(char*)print_iniden[PRINT_INIDEN_LINESTART].ivalue);
	if (print_spectrum==PRINT_COLOR)	Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_REDPASS].ivalue);
	Print_Output_Color(red);
	if (print_spectrum==PRINT_COLOR) {
		Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_GRNPASS].ivalue);
		Print_Output_Color(grn);
		Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_BLUPASS].ivalue);
		Print_Output_Color(blu);
		if (blk) {
			Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_BLKPASS].ivalue);
			Print_Output_Color(blk);
		}
	};
	Print_fprintf(print_file,(char*)print_iniden[PRINT_INIDEN_LINEEND].ivalue);
	Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_LINEEND].ivalue);
}

////////////////////////////////////////////////////////////////////
//	void Print_FS_Dither(uchar *mem,ulong length,short **fsevenptr,short **fsoddptr,char row);
//
//	Dithers a single row of pixels using Floyd-Steinberg dithering.

void	Print_FS_Dither(uchar *mem,ulong length,short **fsevenptr,short **fsoddptr,char row) {
	short *fseven,*fsodd;
	short *even,*odd;
	char direction;
	short pxl,error,error2;

	fseven=*fsevenptr;
	fsodd=*fsoddptr;

	even=fseven+1;
	odd=fsodd+length+1;
	if (row & 1) {
		direction=-1;
		mem+=length-1;
	} else {
		direction=1;
	};

	memset(fsodd,0,length*2+4);

	do {
		pxl=*mem+((*even++ +8)>>4);
		if (pxl<128) {
			*mem=1;
			error=pxl;
		} else {
			*mem=0;
			error=pxl-255;
		};
		mem+=direction;

		odd[-2]+=error;
		error2=error;
		error+=error;
		error2+=error;
		odd[ 0]+=error2;
		error2+=error;
		odd[-1]+=error2;
		error2+=error;
		*even  +=error2;
		--odd;
	} while(--length);

	*fsevenptr=fsodd;		// swaperoo
	*fsoddptr=fseven;
}

//////////////////////////////////////////////////////////////////////////
//	void Print_Ordered_Dither(uchar *mem,ulong length,char row);
//
//	Dithers a single row of pixels using Ordered dithering.

void	Print_Ordered_Dither(uchar *mem,ulong length,char row) {
	uchar *drow;
	char coland15=0;

	drow=print_ordered[row & 15];
	do {
		*mem=(*mem<*drow++);	// it is < because RGBI planes inverted for printer
		++mem;
		if (++coland15>=16) {
			coland15=0;
			drow-=16;
		};
	} while(--length);
}

////////////////////////////////////////////////////////////////////////
//	void Print_Dither(uchar *mem,ulong length,short **fsevenptr,short **fsoddptr,char row);
//
//	Chooses an appropriate dithering function.

void	Print_Dither(uchar *mem,ulong length,short **fsevenptr,short **fsoddptr,char row) {
	switch (print_dither) {
	case PRINT_DITHER_NONE:		break;
	case PRINT_DITHER_ORDERED:	Print_Ordered_Dither(mem,length,row); break;
	case PRINT_DITHER_FS:		Print_FS_Dither(mem,length,fsevenptr,fsoddptr,row); break;
	};
}

//////////////////////////////////////////////////////////////////////////
//	void Print_Mutex(uchar *red,uchar *grn,uchar *blu,uchar *blk,ulong length)
//
//	Preforms Black-Mutually-Exclusive operation on the RGB planes by
//	preforming an And-Not operation with the Black plane on the RGB planes.

void	Print_Mutex(uchar *red,uchar *grn,uchar *blu,uchar *blk,ulong length) {
	uchar t;
	do {
		t=*blk=*red && *grn && *blu;
		if (t) *red=*grn=*blu=0;
		++red;
		++grn;
		++blu;
		++blk;
	} while(--length);
}

///////////////////////////////////////////////////////////////////////////
//	void Print(void);
//
//	The main printing routine.

boolean	Print(void) {
	long ycnt,ytmp;
	uchar *tred,*tgrn,*tblu,*tblk;
	FILE *f;
	ulong l;

	static boolean first_init=TRUE;

	print_sb=NULL;

	if (first_init) {
		first_init=FALSE;
		for(ycnt=1;ycnt<8;ycnt++) memcpy(print_ini_density[ycnt],print_ini_density[0],sizeof print_ini_density[0]);
	};

	if (!(f=VPfopen("VP:vpprint.ini","r"))) return FALSE;
	if (l=IniProcessFile(f,print_inigroups)) Message("Error in print INI file at line %ld.",l);
	fclose(f);
	if (l) return FALSE;

	if (!PrintRequest()) return FALSE;
	if (!Print_Init()) return FALSE;

	print_sb=CreateStatusBar(print_height,"Printing...");

	Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_INITIALIZE].ivalue);
	Print_fprintf(print_file,(char*)print_iniden[PRINT_INIDEN_PRINTSTART].ivalue);

	for(ycnt=print_height;ycnt>0;ycnt-=print_pins) {

//	printf("Printing row %d\n",print_height-ycnt);
//getch();
	// scale as many rows as possible into the print buffer.

		tred=print_redptr;
		tgrn=print_grnptr;
		tblu=print_bluptr;
		tblk=print_blkptr;

		for (ytmp=min(ycnt,print_pins);ytmp;ytmp--) {
			Print_StepScale(tred,tgrn,tblu);
			if (print_dither) {
				Print_Dither(tred,print_width,&print_fseven[0],&print_fsodd[0],ytmp);
				if (print_spectrum == PRINT_COLOR) {
					Print_Dither(tgrn,print_width,&print_fseven[1],&print_fsodd[1],ytmp);
					Print_Dither(tblu,print_width,&print_fseven[2],&print_fsodd[2],ytmp);
				};
			};
			tred+=print_width;
			tgrn+=print_width;
			tblu+=print_width;
		};

	// if there are any rows left they should be cleared to white (0)

		if (ycnt<print_pins) {
			memset(tred,0,print_width*(print_pins-ycnt));
			if (print_spectrum == PRINT_COLOR) {
				memset(tgrn,0,print_width*(print_pins-ycnt));
				memset(tblu,0,print_width*(print_pins-ycnt));
			};
		};

	// if Black Mutually-Exclusive enabled preform BlackMutex operation

		if (print_spectrum==PRINT_COLOR)	// no sense doing Mutex if not color!
			if (print_ini_general[PRINT_INI_BLKMUTEX].ivalue)
				Print_Mutex(print_redmem,print_grnmem,print_blumem,print_blkmem,print_width*print_pins);

	// rotate rows into output printer buffer and print them

		Print_Output(print_redmem,print_grnmem,print_blumem,print_blkmem);

	// update status bar

		ModifyStatusBar(print_sb,print_height-ycnt);

	};

	Print_fprintf(print_file,(char*)print_iniden[PRINT_INIDEN_PRINTEND].ivalue);
	Print_fprintf(print_file,(char*)print_ini_general[PRINT_INI_TERMINATE].ivalue);

	Print_Cleanup();
	return TRUE;
}

