#define	VGAPAINT_SPARE_C
#include "spare.h"
#include "average.h"
#include "picture.h"
#include "listbox.h"

boolean SpareInit(long w,long h) {
	struct Picture *pict;
	struct BitMap *bm;

	if (!(pict = malloc(sizeof(struct Picture)))) return TRUE;
	if (!(bm = malloc(sizeof(struct BitMap)))) {
		free(pict);
		return TRUE;
	}

	if (!PictureInit(pict,bm,w,h)) {
		free(bm);
		free(pict);
		return TRUE;
	}

	pict->next = pict_spare;
	pict->prev = NULL;
	if (pict_spare)	pict_spare->prev = pict;
	pict_spare = pict;

	sparebm = bm;

	memcpy(pict_spare->palette,piccol,768);
	memcpy(pict_spare->pal_bits,pict_cur->pal_bits,32);
	vd->set(sparebm,0);

	return FALSE;		
}

void SpareSwap(void) {
	void *temp;

	syncbitmap();

	if (!pict_spare) {
		if (SpareInit(picbm->width,picbm->height)) {
			Message("Not enough memory to open spare screen.");
			return;
		};
	};

//	picbm->vm=spare.picmem;
//	spare.picmem=picmem;
//	picmem=picbm->vm;

//	qswapb((void *)sparebm,(void *)picbm,sizeof(struct BitMap));

	// delink the spare, swap pictures, and link in the current pic

	pict_cur->prev = pict_spare->prev;
	pict_cur->next = pict_spare->next;
	if (pict_cur->next) pict_cur->next->prev = pict_cur;
	pict_spare->prev = pict_spare->next = NULL;
	temp = pict_cur; pict_cur=pict_spare; pict_spare=temp;

	// swap other data pointers

	piccol = pict_cur->palette;
	picmem = pict_cur->frame_buf;
	cyclerange = pict_cur->cyclerange;
	picbm = pict_cur->bm;
	sparebm = pict_spare->bm;

	repick_colors();
	restore_screen();
}

void	SpareDelete(void) {
	struct Picture *pict;

	if (pict_spare) {
		pict=pict_spare;
		if (pict->next) {
			pict->next->prev = NULL;
			sparebm = pict->next->bm;
		}
		pict_spare = pict->next;

		free(pict->bm);
		PictureDispose(pict);
	}
}

void	SpareCopy(void) {
	syncbitmap();
	if (pict_spare)
		vd->bitblt(picbm,0,0,sparebm,0,0,sparebm->width,sparebm->height);
//		qmove(picmem,spare.picmem,picbm->size);
	restore_screen();
}

void	SpareGetPalette(void) {
	if (pict_spare) {
		memcpy(pict_cur->palette,pict_spare->palette,768);
		memcpy(pict_cur->pal_bits,pict_spare->pal_bits,32);
		vid_loadRGB6(0,pict_cur->palette,256);
		repick_colors();
		restore_tools();
	};
}

void	SpareFitToPalette(void) {
	uchar remap[256];
	if (pict_spare) {
		vid_make_remap(pict_cur->palette,pict_spare->palette,remap,256,256);
		vid13_translate(picbm,0,0,picbm->width,picbm->height,remap);
		memcpy(pict_cur->palette,pict_spare->palette,768);
		repick_colors();
		restore_tools();
	};
}

void	SpareMerge(void) {
	Spare_Operations(&SpareMergeFunc);
}

void	SpareMergeFunc(uchar *d,uchar *pp,uchar *sp) {
	spare_add(d,pict_cur->frame_buf,pict_spare->frame_buf,picbm->size,pp,sp);
}

void	SpareOverlay(void) {
	Spare_Operations(&SpareOverlayFunc);
}
void	SpareOverlayFunc(uchar *d,uchar *pp,uchar *sp) {
	spare_overlay(d,pict_cur->frame_buf,pict_spare->frame_buf,picbm->size,pp,sp);
}

void	SpareUnderlay(void) {
	Spare_Operations(&SpareUnderlayFunc);
}
void	SpareUnderlayFunc(uchar *d,uchar *pp,uchar *sp) {
	spare_overlay(d,pict_spare->frame_buf,pict_cur->frame_buf,picbm->size,sp,pp);
}

void	Spare_Operations(void (*func)(uchar *,uchar *,uchar *)) {
	uchar trred[256],trgrn[256],trblu[256];
	uchar tsred[256],tsgrn[256],tsblu[256];
	uchar *tp,*tpr,*tpg,*tpb;
	int x;

//	if (!spare.picmem) return;
	if (!pict_spare) return;
	if (picbm->width != sparebm->width || picbm->height != sparebm->height) {
		Message("Current and spare pictures are different sizes!");
		return;
	}

	if (!sepred) if (!(sepred=(uchar *)malloc(picbm->size))) goto xit;
	if (!sepgrn) if (!(sepgrn=(uchar *)malloc(picbm->size))) goto xit;
	if (!sepblu) if (!(sepblu=(uchar *)malloc(picbm->size))) goto xit;

	mouse_hide();

	tp=piccol;
	tpr=trred;
	tpg=trgrn;
	tpb=trblu;
	for(x=256;x;x--) {
		*tpr++=*tp++>>2;
		*tpg++=*tp++>>2;
		*tpb++=*tp++>>2;
	};

	tp=pict_spare->palette;
	tpr=tsred;
	tpg=tsgrn;
	tpb=tsblu;
	for(x=256;x;x--) {
		*tpr++=*tp++>>2;
		*tpg++=*tp++>>2;
		*tpb++=*tp++>>2;
	};

	for(x=0;x<64;x++) vid_setRGB6(x,x<<2,0,0);
	func(sepred,trred,tsred);
	vd->rbitblt(0,0,min(vidbm.width,picbm->width),min(vidbm.height,picbm->height),sepred,picbm->bpr,TRUE);

	for(x=0;x<64;x++) vid_setRGB6(x,0,x<<2,0);
	func(sepgrn,trgrn,tsgrn);
	vd->rbitblt(0,0,min(vidbm.width,picbm->width),min(vidbm.height,picbm->height),sepgrn,picbm->bpr,TRUE);

	for(x=0;x<64;x++) vid_setRGB6(x,0,0,x<<2);
	func(sepblu,trblu,tsblu);
	vd->rbitblt(0,0,min(vidbm.width,picbm->width),min(vidbm.height,picbm->height),sepblu,picbm->bpr,TRUE);

	if (ObtainPalette()) Separate_Remap();

xit:
	Separate_Free();
}


//////////////////// SPARE PICTURE LIST ////////////////////


struct Zone SP_zn[]={
	{ 0,0,0,0,ZN_TEXTRENDER+ZN_3D,NULL,"Delete",NULL},
	{ 0,0,0,0,ZN_TEXTRENDER+ZN_3D,NULL,"SwapCur",NULL},
	{-1,-1,-1,-1}
};

struct ZoneRange SP_zr[]={
	{ NULL,SP_zn,NULL,0,ZR_TAP }
};

void	SparePictureList(void) {
	struct Picture *pict,*pict2;
	struct ListRequester *lr;
	mouse_coord	mx,my,mb;
	ulong		msg;
	boolean		cont=TRUE;
	char **listmem,**listptr,*s;
	int i;

	pict = pict_spare;	i=0;
	while(pict) {
		++i;

//		printf("#%ld: %p<%p>%p\n",i,pict->prev,pict,pict->next);

		pict=pict->next;
	}	// getch();

	if (!(listmem = (char **)malloc(sizeof(char **)*(i+2)))) return;

	listmem[0] = (char *)(i+1);

	pict = pict_spare;	listptr = listmem+1;
	*listptr++ = pict_cur->name;
	while(pict) {
		i = strlen(pict->name);
		if (i>40) *listptr++ = pict->name+i-40; else *listptr++ = pict->name;
		pict=pict->next;
	}

	if (!(lr = MakeListRequester(40,16,"Spare pictures")))
		goto abort0;

	lr->zones[1].text = "SwapSpr";
	SP_zr->next = lr->req->zones;
	lr->req->zones	= SP_zr;
	SP_zn[0].xs	= lr->zones[0].xs;
	SP_zn[0].ys	= lr->zones[0].ys;
	SP_zn[0].x	= lr->zones[0].x+(lr->zones[1].x-lr->zones[0].x)/3;
	SP_zn[0].y	= lr->zones[0].y;
	SP_zn[1].xs	= lr->zones[0].xs;
	SP_zn[1].ys	= lr->zones[0].ys;
	SP_zn[1].x	= lr->zones[0].x+((lr->zones[1].x-lr->zones[0].x)*2)/3;
	SP_zn[1].y	= lr->zones[0].y;

	lr->tbox->table = listmem;

	TextboxCreate(lr->tbox,TRUE);

	if (!OpenRequester(lr->req,0,0)) goto abort1;

	do {
		mouse_read2(&mx,&my,&mb);
		if (mb & 1) {
			msg=CheckRequester(lr->req,mx,my,mb);
			mouse_wait_unkey();
			switch(msg) {
			case RZoneMsg(0,0):	// delete
				break;
			case RZoneMsg(0,1):	// switch current
				if (lr->lbox->sel>0) {
					if (lr->lbox->sel == 1) {	// swap with spare
						// swap text strings

						s=listmem[1]; listmem[1]=listmem[2]; listmem[2]=s;

						// swap pointers

						pict2=pict_spare->next; pict_spare->next=pict_cur->next; pict_cur->next=pict2;
						pict2=pict_spare->prev; pict_spare->prev=pict_cur->prev; pict_cur->prev=pict2;
						// recorrect other pointers

						if (pict_spare->prev) pict_spare->prev->next = pict_spare;
						if (pict_spare->next) pict_spare->next->prev = pict_spare;
						if (pict_cur  ->prev) pict_cur  ->prev->next = pict_cur;
						if (pict_cur  ->next) pict_cur  ->next->prev = pict_cur;

						pict2 = pict_spare; pict_spare=pict_cur; pict_cur = pict2;

						piccol = pict_cur->palette;
						picmem = pict_cur->frame_buf;
						cyclerange = pict_cur->cyclerange;
					} else {
						// delink the selected item

						pict = pict_spare;
						for(i=lr->lbox->sel-1;i;i--) pict=pict->next;

						if (pict->next) pict->next->prev = pict->prev;
						if (pict->prev) pict->prev->next = pict->next;
						s = listmem[lr->lbox->sel+1];

						memmove(listmem+4,listmem+3,sizeof(char *)*(lr->lbox->sel-2));

						// relink pict_cur in right after pict_spare

						pict_cur->next = pict_spare->next;
						pict_cur->prev = pict_spare;
						if (pict_spare->next) pict_spare->next->prev = pict_cur;
						pict_spare->next = pict_cur;
						listmem[3] = listmem[1];

						// replace pict_cur with pict

						pict_cur = pict;
						pict->prev = pict->next = NULL;
						listmem[1] = s;
					}

					// reset pointers

					piccol = pict_cur->palette;
					picmem = pict_cur->frame_buf;
					cyclerange = pict_cur->cyclerange;

					// redraw the listbox

					DrawListbox(lr->lbox);
				}
				picbm = pict_cur->bm;
				break;
			case RZoneMsg(1,1):	// switch spare
				if (lr->lbox->sel>=0 && lr->lbox->sel!=1) {
					if (lr->lbox->sel == 0) {	// swap with pict_cur
						// swap text strings

						s=listmem[1]; listmem[1]=listmem[2]; listmem[2]=s;

						// swap pointers

						pict2=pict_spare->next; pict_spare->next=pict_cur->next; pict_cur->next=pict2;
						pict2=pict_spare->prev; pict_spare->prev=pict_cur->prev; pict_cur->prev=pict2;
						// recorrect other pointers

						if (pict_spare->prev) pict_spare->prev->next = pict_spare;
						if (pict_spare->next) pict_spare->next->prev = pict_spare;
						if (pict_cur  ->prev) pict_cur  ->prev->next = pict_cur;
						if (pict_cur  ->next) pict_cur  ->next->prev = pict_cur;

						pict2 = pict_spare; pict_spare=pict_cur; pict_cur = pict2;

						piccol = pict_cur->palette;
						picmem = pict_cur->frame_buf;
						cyclerange = pict_cur->cyclerange;
						picbm = pict_cur->bm;
					} else {

						pict = pict_spare;
						for(i=lr->lbox->sel-1;i;i--) pict=pict->next;

						// move the text pointers around

						s = listmem[lr->lbox->sel+1];
						memmove(listmem+3,listmem+2,sizeof(char *)*(lr->lbox->sel-1));
						listmem[2] = s;

						// delink pict

						if (pict->next) pict->next->prev = pict->prev;
						if (pict->prev) pict->prev->next = pict->next;

						// relink at the front of the spare list

						pict_spare->prev = pict;
						pict->next = pict_spare;
						pict->prev = NULL;
						pict_spare = pict;
					}

					DrawListbox(lr->lbox);
					sparebm = pict_spare->bm;
				}
				break;
			case RZoneMsg(1,0):	/* Ok */
				cont=FALSE;
				break;
			};
		};
	} while(cont);

	CloseRequester(lr->req);

abort1:
	DisposeListRequester(lr);
abort0:
	free(listmem);
}

