/////////////////////////////////////////////////////////////////////////////
//
//	VGAPaint 386 - The Useful Paint Program
//
//	a freeware ripoff of DPaint by Avery Lee
//
//	this module is the messiest of them all
//

#define	VGAPAINT_MAIN

#include <math.h>
#include <malloc.h>
#include <signal.h>

#include "vgapaint.h"
#include "process.h"
#include "capture.h"
#include "io.h"
#include "menu.h"
#include "inifile.h"
#include "font.h"
#include "picture.h"
#include "anim.h"
#include "draw.h"
#include "loadlib.h"
#include "polygon.h"
#include "parm.h"

#include "startup.con"

extern struct List VPALoadList;
extern struct IniItem vp_ini_autoload[];
extern struct IniItem vp_ini_options[];
extern struct IniItem vp_ini_prefs[];
extern struct IniGroup vp_inigroups[];

void	change_paint_tool(ulong tool_id);
void	draw_undo(void);
void	Paint_BlendToScreen(coord x1,coord y1,coord x2,coord y2,short blitmode);
void	VP_PrevFrame(void);
void	VP_NextFrame(void);
void	VP_InsertBeforeFrame(void);
void	VP_InsertAfterFrame(void);
void	VP_DeleteCurrentFrame(void);

#define OOPS { mouse_ptrdispose(); exit(5); }
#define BUZZ outportb(0x61,(inportb(0x61) & 0xFE) ^ 0x02)

static boolean	video_initialized=FALSE;
static boolean	dcd_e=FALSE;

static struct List autoload_extensions;
static struct CallList main_accelerators;
static struct CallList idlers;

extern struct CallList menupickers;

void vpexit(void) {
	static boolean exited=FALSE;

	if (exited) return;
	exited=TRUE;

	PrtSc_Remove();
	CCycle_Remove();
	mouse_ptrdispose();
	if (video_initialized) vd->init(VD_CLOSEDOWN,0,0);
	if (picmem) free(picmem);

	unload_extensions(&autoload_extensions);

	if (paltrnmem) {
		if (opt.nodirect_pal)
			DosMemoryFree(paltrndesc);
		else
			free(paltrnmem);
		paltrnmem=NULL;
	}

	DeinitCallbackSystem();
	if (realmem) {
		DosMemoryFree(realdsc);
		realmem=NULL;
	};
	DeinitInterruptHandler();
}

extern boolean easter_decode;
static	void (*old_break_process)(int);

static void Break_Process(int sig_no) {
	signal(SIGINT,Break_Process);
}

extern char version_str[];

void Log(long logstr_num, ...) {
	char buf[1024];
	va_list val;
	char *log_string;
	FILE *log_handle;

	if (!opt.logging) return;

	if (!(log_handle=VPfopen("VP:vp386.log","a"))) {
		puts("VP386: Cannot open log file.");
		exit(10);
	}

	if (logstr_num==-1L) {
		va_start(val,logstr_num);
		strcpy(buf,va_arg(val,char *));
		va_end(val);
	} else {
		if (!(log_string = FindResourceString(&startup_strings,logstr_num)))
			sprintf(buf,"Unknown startup string %ld",logstr_num);
		else {
			va_start(val,logstr_num);
			vsprintf(buf,log_string,val);
			va_end(val);
		}
	}

	if (fprintf(log_handle,"VP386: %s.\n",buf)<0) {
		puts("VP386: Error writing to log.");
		exit(10);
	}

	if (fclose(log_handle)) {
		puts("VP386: Error flushing log filehandle.");
		exit(10);
	}
}

void Init(int argc,char **argv) {
	uchar c;

	// assign the start path to VP: so we can find stuff

	VP_Set_Start_Path(argv[0]);

	// check for logging flag

	if (!strcmp(argv[1],"-dL")) {
		opt.logging=TRUE;
		Log(SSTR_INITIAL);
	}

	// load resources

	if (!LoadResourceFile("VP:vp386.res")) {
		puts("Error loading vp386.res.");
		exit(10);
	}

	if (!key_shortcuts) {
		puts("No keyboard shortcuts in vp386.res!");
		exit(10);
	}
	if (!startup_strings.count) Log(-1L,"No startup strings in vp386.res, cannot log startup");

	// initialize data lists

	NewList(&autoload_extensions);
	NewList(&main_accelerators.list);
	NewList(&idlers.list);
	NewList(&menupickers.list);
	NewList(&VPALoadList);
	InitializeAutoloadSystem();

	// install exit handler so we don't cream the system when the user ctrl-breaks and moves the mouse

	Log(SSTR_SIGHANDLERS);

	onexit(&vpexit);

	// install ctrl-c handler so we don't cream pictures when the user hits ctrl-c

	old_break_process = signal(SIGINT,Break_Process);

	// initialize interrupt and call chain systems

	Log(SSTR_INITINTERRUPTS);

	if (InitInterruptHandler()) exit(20);

	AddCallList(&main_accelerators,'MAcl');
	AddCallList(&idlers,'IDLE');
	AddCallList(&menupickers,'MPck');

	// allocate real mode memory

	Log(SSTR_ALLOCREALMEM);

	if (!(realmem=DosMemoryAllocate(8+48,&realdsc))) {
		printf("Unable to allocate 864 byte real-mode memory buffer.\n");
		exit(30);
	};
	realtmp = (char *)realmem+8*16;

	// initialize RM->PM callback system now that we have RM memory

	Log(SSTR_INITCALLBACKS);

	if (!InitCallbackSystem()) exit(20);

	// check to see if some lameass changed my program and if so abort

	Log(SSTR_ALTERCHECK);

	check_for_bozo_alteration();

	// load system fonts

	Log(SSTR_LOADINGFONT,"6x16.vfn");	sys6x16	 = LoadFont("VP:6x16.vfn");
	Log(SSTR_LOADINGFONT,"8x8.vfn");	sys8x8	 = LoadFont("VP:8x8.vfn");
	Log(SSTR_LOADINGFONT,"8x12.vfn");	sys8x12	 = LoadFont("VP:8x12.vfn");
	Log(SSTR_LOADINGFONT,"12x16.vfn");	sys12x16 = LoadFont("VP:12x16.vfn");

	// load help file, if we can

	Log(SSTR_LOADINGHELP);

	vp386_helpfile = OpenHelpFile("VP:vp386.hlp");

	// _VGA_Paint cannot run without a VGA card (duh)

	Log(SSTR_DETECTVIDEO);

	if ((c=vid_detect()) != 'V') {
		printf(	"VP386: The%s video card you have in your system is not suitable for\n"
			"       use with VGAPaint 386.  You must have a VGA card with at least\n"
			"       256K of video RAM.\n",

			/* reminds me of LISP */

				(c=='E' ? " EGA" :
					(c=='C' ? " CGA" :
						(c=='M' ? " MDA" : "")
					)
				)
		);
		exit(20);
	};

	// allocate memory for initial picture bitmap

	Log(SSTR_ALLOCBITMAP);

	if (!(picbm = malloc(sizeof(struct BitMap)))) {
		printf("VP386: Not enough memory to allocate BitMap structure.\n");
		exit(20);
	}

	// read VP386.INI if it exists

	Log(SSTR_LOADINIFILE);

	{
		FILE *f;
		ulong err;

		if (f=VPfopen("VP:vp386.ini","r")) {
			if (main_vp386_ini=IniReadFile(f)) {
				if (err=IniProcessFileMem(main_vp386_ini,vp_inigroups)) {
					printf("Error processing VP386.INI: line %ld\n",err);
					exit(10);
				}
			}
			fclose(f);
		} else if (!(main_vp386_ini = IniCreateFile())) {
			puts("VP386: Can't allocate memory for INI");
			exit(20);
		}
	}

	// parse parameters

	Log(SSTR_PARSEPARAM);

	ParamParse(argc,argv,(char *)vp_ini_options[4].ivalue);

	if (vp_ini_options[4].ivalue) {
		free((void *)vp_ini_options[4].ivalue);
		vp_ini_options[4].ivalue = NULL;
	}

	// analyze system for OS/2, Windows 95, and other flaws

	if (!opt.wimp_mode) {
		Log(SSTR_OSDETECT);
		SetSystemFlags();
	}

	// boot the system timer up to 72.8 times per second

	if (!opt.wimp_mode) {
		Log(SSTR_BOOSTTIMER);
		SetTimerPeriod(0x4000);
	}

	// copy values over from INI file

	Log(SSTR_INITOMENU);

	menu_set_toggle_option(menu_item_address(barmenu+1,4,2,-1),opt.reptdraw		= vp_ini_prefs[0].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,3,-1),opt.coords		= vp_ini_prefs[1].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,4,-1),opt.be_square	= vp_ini_prefs[2].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,5,-1),opt.autotransp	= vp_ini_prefs[3].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,6,-1),opt.adjustBrush	= vp_ini_prefs[4].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,7,-1),opt.brushflickerfree	= vp_ini_prefs[5].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,8,-1),opt.useHiresTools	= vp_ini_prefs[6].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,9,-1),opt.CPUhog		= vp_ini_prefs[7].ivalue);
	menu_choose(menu_item_address(barmenu+1,4,13,0,-1),0,2,scroll_mode = vp_ini_prefs[8].ivalue);
	menu_set_toggle_option(menu_item_address(barmenu+1,4,14,-1),opt.fastPreview		= vp_ini_prefs[9].ivalue);
	menu_choose(menu_item_address(barmenu+1,4,16,0,-1),3,3,(reduce_mode = vp_ini_prefs[14].ivalue)+3);
	opt.reqImageDrag	= vp_ini_options[1].ivalue;
	if (opt.releaseTest)	opt.debugOpts = FALSE;
	else			opt.debugOpts = vp_ini_options[2].ivalue;

	scroll_speed	= vp_ini_prefs[13].ivalue;

	if (vp_ini_prefs[10].type & INI_FOUND) picbm->width  = vp_ini_prefs[10].ivalue;
	if (vp_ini_prefs[11].type & INI_FOUND) picbm->height = vp_ini_prefs[11].ivalue;

	// if we're using direct RAMDAC video writes, allocate 32-bit memory, else allocate real memory

	Log(SSTR_ALLOCPALMEM);

	if (opt.nodirect_pal) {
		if (!(paltrnmem = DosMemoryAllocate(48,&paltrndesc))) exit(10);
	} else {
		if (!(paltrnmem = malloc(768))) exit(20);
	}

	// initialize video driver list

	Log(SSTR_INITBUILTINVD);

	InitBuiltinVideoDrivers();

	// autoload extensions, but only if not in wimp mode

	Log(SSTR_AUTOLOADEXT);

	if (!opt.wimp_mode) {
		struct Extension *ext;
		struct Node *n=VPALoadList.prev,*next;

		while(next=n->prev) {
			Log(SSTR_LOADINGEXT,(char *)n+sizeof(struct Node));
			if (ext=load_extension((char*)n+sizeof(struct Node))) {
				if (ext != (void *)-1L)
					AddTail(&autoload_extensions,(struct Node *)ext);
			} else {
				printf("VP386: Error loading extension \"%s\"; press any key to continue.\n",(char *)n+sizeof(struct Node));
				getch();
			};
			n=next;
		};
	}

	// initialize picture structures

	Log(SSTR_ALLOCPICTURE);

	if (!(pict_cur = malloc(sizeof(struct Picture)))) {
		printf("VP386: Not enough memory to allocate Picture structure.\n");
		exit(20);
	}
	pict_cur->inited = FALSE;

	Log(SSTR_INITPICTURE);

	if (!PictureInit(pict_cur,picbm,picbm->width,picbm->height)) {
		printf("VP386: Error allocating %ldx%ld (%ldK) picture buffer.\n",picbm->width,picbm->height
			,(vid13_rassize(picbm->width,picbm->height)+1023)>>10);
		exit(20);
	}
	piccol = pict_cur->palette;
	picmem = pict_cur->frame_buf;
	cyclerange = pict_cur->cyclerange;

	// display guru if easter

	if (easter_decode) {
		Log(SSTR_DISPLAYALERT);
		DisplayEasterAlert();
	}

	Log(SSTR_GETPATH);
	filereq_init();		/* get path */

	/* initialize video mode */

	MonitorGammaReset();

	vd = NULL;

#if 0
	if (vp_ini_prefs[12].type & INI_FOUND) {
		if (SwitchVideoDriverByName((char *)vp_ini_prefs[12].ivalue,FALSE)) {
			printf("VP386: Unable to initialize video driver \"%s\".\n",(char *)vp_ini_prefs[12].ivalue);
			exit(20);
		}
	} else {
		if (SwitchBuiltinVideoDriver(VDB_Mode13,FALSE)) {
			puts("VP386: 320x200 Video Driver initialization failed!");
			exit(20);
		}
	}
#else
	Log(SSTR_SWITCH320x200);
	if (SwitchBuiltinVideoDriver(VDB_Mode13,FALSE)) {
		puts("VP386: 320x200 Video Driver initialization failed!");
		exit(20);
	}
	if (vp_ini_prefs[12].type & INI_FOUND) {
		Log(SSTR_SWITCHDRIVER);
		if (SwitchVideoDriverByName((char *)vp_ini_prefs[12].ivalue,FALSE)) {
			printf("VP386: Unable to initialize video driver \"%s\".\n",(char *)vp_ini_prefs[12].ivalue);
			exit(20);
		}
	}
#endif

	Log(SSTR_CLEARSCREEN);

	vd->set(&vidbm,0);

	video_initialized=TRUE;

	// initialize mouse

	Log(SSTR_INITMOUSE);

	if (mouse_init()<2) {
		vid_setmode(0x03);
		printf("VP386: Failure initializing mouse driver.\n");
		exit(10);
	};

	// install harderr handler

	Log(SSTR_INITHARDERR);

	InstallHarderrHandler();

	// initialize picture structures

	Log(SSTR_MISCINIT);

	br_plotback=vd->plot;
	memcpy(piccol,stdcol,768);
	vid_loadRGB6(0,piccol,256);
	if (!(fpcol = FPick_Allocate(256))) exit(5);

	/* initialize mouse */

	mouse_bounds_reset();

	/* initialize bitmaps */

	vid_initBitMap(&invizbm,(uchar *)0xa0000,0,0,0,0,0,packmem);
	remake_clip_bitmaps();

	/* clear pictures */

	vd->refresh();
	vd->set(picbm,0);

	/* setup interrupt handlers */

	if (!opt.wimp_mode) {
		Log(SSTR_PRTSCTRAP);
		PrtSc_Install();
	}
}

//////////////////////////////////////////////////////////////

static BLOCK	balloon_block = NULL;
static coord	balloon_x,balloon_y,balloon_xs,balloon_ys;
static int	balloon_id = 0;

enum BALLOON_ID {
	BALLOON_PALETTE	= 0x00010000L,
};

static boolean	make_balloon(coord y,coord xs,coord ys,int id) {
	if (balloon_id == id && (!vp_ini_options[3].ivalue || (y-(ys>>1) == balloon_y)) || toolmode != TOOLMODE_RIGHT) return FALSE;

	hide_balloon();

	OffBrush();
	vd->refresh();

	balloon_x = toolpos.t_x1 - xs;
	balloon_y = y-(ys>>1);

	mouse_hide();
	if (balloon_block = vd->blocksave(&vidbm,balloon_x,balloon_y,xs,ys)) {
		vd->bar(&vidbm,balloon_x,balloon_y,xs,ys,St2);
		vid_draw3d(&vidbm,balloon_x,balloon_y,balloon_x+xs-1,balloon_y+ys-1);

		balloon_id = id;
	}
	mouse_show();

	return balloon_block ? TRUE : FALSE;
}

static void	hide_balloon(void) {
	if (balloon_id) {
		mouse_hide();
		vd->blockrestore(&vidbm,balloon_block);
		vd->refresh();
		mouse_show();
		balloon_id = 0;
	}
}

static brush_color_changed=FALSE;

void breakpoint(void);
#pragma aux breakpoint="int 3";

void VP_Shell_Out(void) {
	char	old_pr[256];
	boolean	have_prompt = FALSE;
	char	*s;

	DeinstallHarderrHandler();
	mouse_suspend();
	vd->init(VD_TO13,0,0);
	if (vgacm.type == VCT_ADVANCED)
	if (capture=(struct VGACapture *)malloc(sizeof(struct VGACapture)))
			opt.advanced_capture=TRUE;
	opt.scr_captured=FALSE;
	SetTimerPeriod(0);
	Capture_Install();
	vid_setmode(0x03);
	_heapshrink();
	printf( "Type EXIT to return to VGAPaint 386.\n");
	Capture_Print_Info();

	// swap the prompts

	if (s = getenv("PROMPT")) {
		have_prompt = TRUE;

		strcpy(old_pr,"PROMPT=[VP386] ");
		strcpy(old_pr+15,s);

		putenv(old_pr);
	} else {
		putenv("PROMPT=[VP386] $P$G");
	}

//	puts("about to shell out...");
//	getch();

	system("command");

	if (have_prompt) {
		memcpy(old_pr+8,"PROMPT=",7);
		putenv(old_pr+8);
	} else {
		putenv("PROMPT=");
	}

	vid_setmode(0x13);
	Capture_Deinstall();
	if (!opt.wimp_mode) SetTimerPeriod(0x4000);
	mouse_resume();
	InstallHarderrHandler();
	vd->init(VD_FROM13,0,0);
	if (vgacm.frames) {
		if (vgacm.type == VCT_SIMPLE || vgacm.type == VCT_RF || vgacm.type == VCT_RFC
				|| vgacm.type == VCT_RFX)
			Capture_SimpleToAnim();
		else if (opt.advanced_capture && opt.scr_captured) {
			CaptureDisplay();
		}
	}
	free(capture);
	Capture_FreeMemory();
	vid_loadRGB6(0,piccol,256);
	restore_screen();
}

int main(int argc,char **argv) {
	mouse_coord x,y,ox,oy,b;
	mouse_coord last_x,last_y;
	int rect;
	int key;

	Init(argc,argv);

	Log(SSTR_INITPAINT);

	clipOn();
	repick_colors();
	set_tools_mode(TOOLMODE_RIGHT);
	display_tools();
	GetStdBrush(0);

	// display modeswitch requester if asked to

	if (vp_ini_options[0].ivalue) SelectVideoDriver();

	// display special title bar

	SetTitle(version_str);
	restore_scrn = FALSE;

	Log(SSTR_FINISH);

	while(!quit) {
		Idle(last_x != x || last_y != y);

		if (ms_mpi.winstate_chgd) {
			ms_mpi.winstate_chgd = FALSE;
			SetTitle(NULL);
		}

		if (brush_color_changed) {
			brush_color_changed = FALSE;
			OffBrush();
		}

		if (restore_scrn) {
			restore_scrn=FALSE;
			restore_screen();
		};
		if (kbhit())
			if ((key=CheckKeyInterrupt())!=-1) {
			uchar kbs=kbd_shift();

			OffBrush();
			hide_balloon();

			if (!ExecuteCallList(&main_accelerators,(long)key,kbs))
				switch(key) {
			case -1: break;
			case '1': VP_PrevFrame();		break;
			case '2': VP_NextFrame();		break;
			case '3': VP_GotoFrame();		break;
			case '4': Anim_Playback(pict_cur,NULL);	break;
			case '5': Anim_Playback(pict_cur,ANIM_PLAYBACK_ONCE); break;
			case '6': Anim_Playback(pict_cur,ANIM_PLAYBACK_PINGPONG); break;
			case '!': VP_InsertBeforeFrame();	break;
			case '@': VP_InsertAfterFrame();	break;
			case '#': VP_DeleteCurrentFrame();	break;

			case '7': BrushPrevFrame();		break;
			case '8': BrushNextFrame();		break;
			case '&': BrushInsertFrameBefore();	break;
			case '*': BrushInsertFrameAfter();	break;
			case '(': BrushDeleteCurrentFrame();	break;

			case 'p': DoPalette(); break;
			case 'P':
				if (pict_cur->current->frame_pal)
					DoArrange(pict_cur->current->frame_pal->palette,pict_cur->pal_bits,"Frame ");
				else
					DoArrange(pict_cur->palette,pict_cur->pal_bits,"Global ");
				break;
			case 'u': Undo(); break;
			case '.': GetStdBrush(0); break;
			case ',': mouse_setpointer(eyedropper,0,15,col.white,col.black);
				  do
				  	;
				  while(!mouse_buttons() && ' '!=kbd_check());

				  if (mouse_buttons()) {
					  mouse_read_draw(&x,&y,&b,NULL);
					  if (vid_fastinside(winbm,x,y))
						  ((b&1)?&palette_pickColor:&palette_pickHighColor)(vd->point(winbm,x,y));
					  mouse_wait_unkey();
				  }
				  mouse_setpointer(crosshairs,7,7,col.white,col.black);
				  break;
			case 'j': SpareSwap(); break;
			case 'j'-0x60:	SpareCopy(); break;
			case 'r'-0x60:	DoRange(); break;
			case 'K': clearHandler(NULL,NULL,0,0); break;
			case 'x':	brush_flipx(); break;
			case 'y':	brush_flipy(); break;
			case 'X':	brush_double_x(); break;
			case 'Y':	brush_double_y(); break;
			case 'h':	brush_halve(); break;
			case 'H':	brush_double(); break;
			case 'z':	brush_rotright(); break;
			case 'Z':	brush_scale(); break;
			case 'x'-0x60:	brush_handle_cycle(); break;
			case 0x1B:	quit=(RequestBoolean("VP386 Warning","Yes",NULL,"No","Quit this great program?")==2);
					break;
			case 0x124:	// J without scancode
				if (kbs & KBD_ALT)
					SparePictureList();
				break;
			case 0x142:
			case 0x242:
				    if (!(kbs & KBD_SHIFT)) {
						ToneAdjust();
						break;
				    };
			case 0x15b:
			case 0x25b: ToneAdjust2();
				    break;
			case 0x167: case 0x267:
			case 0x15d: case 0x25d:
			case 0x244:
			case 0x144: if (kbs & KBD_CTRL) {
					if (!vd->init(VD_GETLINECOMPARE,0,0)) {
						if (vd->init(VD_SETLINECOMPARE,vidbm.height-(toolpos.b_height+toolpos.t_height),0)) {
							vd->init(VD_SETSTART,0,toolpos.b_height+toolpos.t_height);
							mouse_set_split(vidbm.height - (toolpos.b_height+toolpos.t_height),toolpos.b_height+toolpos.t_height);
						}
					} else {
						vd->init(VD_SETLINECOMPARE,0,0);
						vd->init(VD_SETSTART,0,0);
						mouse_set_split(0,0);
					}
				    } else if (kbs & KBD_SHIFT) {
					set_tools_mode(toolmode ^ (TOOLMODE_RIGHT^TOOLMODE_TOP));
				    } else {
				    	if (toolgrid) erase_tools(); else display_tools();
				    }
				    break;
			case 0x184:
			case 0x284:
			case 0x149:
			case 0x249:
				SetPrevName(pict_cur->name);

				// convert KBD_LSHIFT and KBD_RSHIFT to KBD_SHIFT

				if (kbs & KBD_SHIFT) kbs |= KBD_SHIFT;

				switch(kbs & (KBD_SHIFT | KBD_CTRL | KBD_ALT)) {
				case KBD_ALT:
					FileIORespond(BMPWrite(GetFileNameSave("Save BMP file","bmp"),picbm));
					break;
				case KBD_SHIFT | KBD_CTRL:
					FileIORespond(IFFWrite(GetFileNameSave("Save IFF file","lbm"),picbm));
					break;
				case KBD_SHIFT:
					FileIORespond(Pak_Dump(GetFileNameSave("Save Raw data","pak"),picbm)); 
					break;
				case KBD_CTRL:
					FileIORespond(PCXCom_Write(GetFileNameSave("Save .COM file","com"),picbm));
					break;
				case NULL:
					FileIORespond(PCXWrite(GetFileNameSave("Save PCX file","pcx"),picbm,pict_cur->cur_pal));
					break;
				}
				break;
			case 0x176:	case 0x276:
			case 0x151:	case 0x251:
				if (pict_cur->flags & PICT_HASNAME) SetPrevName(pict_cur->name);

				// convert KBD_LSHIFT and KBD_RSHIFT to KBD_SHIFT

				if (kbs & KBD_SHIFT) kbs |= KBD_SHIFT;

				switch(kbs & (KBD_SHIFT | KBD_CTRL)) {
				case KBD_SHIFT | KBD_CTRL:
					IFFRead(GetFileName("Read IFF file","lbm"));
					break;
				case KBD_SHIFT:
					Raw256_Extract(GetFileName("Extract RAW data","pak"));
					break;
				case NULL:
//					PCXRead(GetFileName("Read PCX file","pcx"),2);
					PictureLoad();
					break;
				}
				break;
			case 0x148:	scroll_viewport(0,kbs & KBD_SHIFT ? -1 : -8); break;
			case 0x14b:	scroll_viewport(kbs & KBD_SHIFT ? -1 : -8,0); break;
			case 0x14d:	scroll_viewport(kbs & KBD_SHIFT ? 1 : 8,0); break;
			case 0x150:	scroll_viewport(0,kbs & KBD_SHIFT ? 1 : 8); break;
			case 'r':	change_paint_tool(DTOOL_RECT); break;
			case 'R':	change_paint_tool(DTOOL_RECT + DTOOL_DBL); break;
			case 'c':	change_paint_tool(DTOOL_CIRCLE); break;
			case 'C':	change_paint_tool(DTOOL_CIRCLE + DTOOL_DBL); break;
			case 'e':	change_paint_tool(DTOOL_ELLIPSE); break;
			case 'E':	change_paint_tool(DTOOL_ELLIPSE + DTOOL_DBL); break;
			case 'w':	change_paint_tool(DTOOL_POLYGON); break;
			case 'W':	change_paint_tool(DTOOL_POLYGON + DTOOL_DBL); break;
			case 'q':	change_paint_tool(DTOOL_CURVE); break;
			case 'v':	change_paint_tool(DTOOL_LINE); break;
			case 'b':	change_paint_tool(DTOOL_BRUSH); break;
			case 'm':	if (zoomstate) ZoomOff(); else ZoomOn(vposx+x-winbm->x1,vposy+y-winbm->y1); break;
			case 'M':	change_paint_tool(DTOOL_MAGNIFY); break;
			case 't':	change_paint_tool(DTOOL_TEXT); break;
			case 'g':	change_paint_tool(DTOOL_GRID); break;
			case 's':	change_paint_tool(DTOOL_DOTDRAW); break;
			case 'd':	change_paint_tool(DTOOL_LINEDRAW); break;
			case 'a':	change_paint_tool(DTOOL_SPRAY); break;
			case 'f':	change_paint_tool(DTOOL_FILL); break;
			case '<':	ZSetHandler2(); break;
			case '>':	ZSetHandler1(); break;
			case 'T':
				if (opt.debugOpts) {
//					vid13_htrapezoid(&vidbm,80,240,50,160,160,150,7);
//					mouse_wait_key();
//					mouse_wait_unkey();
//					PickAbstractColor();

					FILE *fi;
					if (fi=VPfopen("dump.bin","rb")) {
						LoadLossyLZFile(fi);
						fclose(fi);
					}
				}
				break;
			case '`':
				if (opt.debugOpts) {
					FILE *fo;
					if (fo=VPfopen("dump.bin","wb")) {
						LossyLZCompress(fo,picmem,picbm->size,2);
						fclose(fo);
					}

				}
				break;
			};
		};
		if (easter_decode && dcd_e) {dcd_e=FALSE; DecodeEasterPicture(); }

		last_x = x; last_y = y;
		mouse_read_draw(&x,&y,&b,MRD_SETZOOMFLG+MRD_USESHIFT);

	/* check to see if pointer has strayed outside boundaries;
	** if so, scroll the viewport to see more of the picture */

		check_scroll();


		if (b) {
			Idle(TRUE);

			OffBrush();
			hide_balloon();
			if (toolgrid && -1!=(rect=rectSearch(ToolRects,x,y))) {
				if (-1!=ZoneRangeAct(&ZRTool1[0],x,y))
					;
				else if (rect==0) {
					if (b & 2) {
						DisplayMenuBar(barmenu);
						if (kbd_shift() & KBD_SHIFT)
							MenuPickHelp();
						else
							MenuPick();
						ms_xscroll=ms_yscroll=0;
					}
				} else if (rect==2) {
					rectInvert(&ToolRects[2]);
					mouse_wait_unkey();
					rectInvert(&ToolRects[2]);

					// right or left mouse button?

					if (b&2) {

						// right; take screen capture

						mouse_hide();
						vd->bitblt(&vidbm,vidbm.x1,vidbm.y1,picbm,picbm->x1,picbm->y1,vidbm.width,vidbm.height);
						reblit_screen();
						mouse_show();

					} else {

						// left; shell out

						VP_Shell_Out();
					}
				}
				mouse_wait_unkey();
			} else {
				if (b & 2) {
					dodraw(0);
				} else if (b & 1) {
					dodraw(1);
				};
				pict_cur->dirty = TRUE;
			};
		} else {
			if (((x>=toolpos.t_x1 && x<toolpos.t_x1+toolpos.t_width && y<toolpos.t_y1+toolpos.t_height)
						|| y<toolpos.b_height) && toolgrid) {
				OffBrush();
				hide_balloon();
				if (flip_colors)
					if (!colors_reset) reset_colors();
			} else if (x>=ZTool4[0].x && y>=ZTool4[0].y && x<ZTool4[0].x+ZTool4[0].xs && y<ZTool4[0].y+ZTool4[0].ys && toolgrid) {
				int cl;
				char tmp1[9],tmp2[9];

				cl = palset[palette_coordToColor(x,y)];

				sprintf(tmp1,"%3d (%02x)",cl,cl);
				sprintf(tmp2,"%02x %02x %02x",piccol[3*cl+0],piccol[3*cl+1],piccol[3*cl+2]);

				OffBrush();
				if (make_balloon(y,max(strlen(tmp1),strlen(tmp2))*sysfontwidth+4*XPOINT,2*sysfontheight+4*YPOINT-1,cl + BALLOON_PALETTE)) {
					mouse_hide();
					SysText(&vidbm,balloon_x+2*XPOINT,balloon_y+2*YPOINT,tmp1,St0,St2);
					SysText(&vidbm,balloon_x+2*XPOINT,balloon_y+2*YPOINT+sysfontheight,tmp2,St0,St2);
					vd->refresh();
					mouse_show();
				}
			} else if (ZRTool1[0].highindex<10) { // don't display brush with brush or text tool
				brush_sprite.colr=(uchar)ZRTool1[3].highindex;
				if (opt.brushflickerfree) WaitVBL();
				ox=x; oy=y;
				snapToGrid(&ox,&oy,vposx-winbm->x1,vposy-winbm->y1);
				sprite_quikmove2(&brush_sprite,ox-brush_hx,oy-brush_hy);
				hide_balloon();
				OnBrush();
				if (colors_reset) restore_colors();
			};
		};
	};

	vpexit();
	vid_setmode(0x03);

	print_copyright();

	return EXIT_SUCCESS;
}


int CheckKeyInterrupt(void) {
	int key;

	if (!kbhit()) goto xit;	/* refresh the screen, please... */
	key=kbd_check();
	if (key) {
		mouse_hide();
		switch(key) {
		case ' ': dabort=TRUE;
			  mouse_show();
			  return key;
		case '|': menu_pick(4,3,-1); break;
		case '[': palette_pickColor((uchar)(fcolorpos-1)); break;
		case ']': palette_pickColor((uchar)(fcolorpos+1)); break;
		case '{': palette_pickHighColor((uchar)(bcolorpos-1)); break;
		case '}': palette_pickHighColor((uchar)(bcolorpos+1)); break;
		case 0x13B: BrushMode(0); break;
		case 0x13C: BrushMode(1); break;
		case 0x13D: BrushMode(2); break;
		case 0x09: CCycle_Toggle();
			   break;
		default:	mouse_show();
				return key;
		};
		mouse_show();

		brush_sprite.colr=fcolor;
		dcolr=dcolrp ? fcolor : bcolor;
		if (-1!=dinvert) dinvert=dcolr;
		brush_color_changed = TRUE;
	};
xit:
	vd->refresh();
	return -1;
}


///////////////////////////////// POINT ARITHMETIC /////////////////////////////

typedef void (*VPplotptr)(BMP bm,coord x,coord y,color c);
static VPplotptr VPplot,VPbrush;
static uchar VPplot_translate_table[256];

void	VPplot_transp(BMP bm,coord x,coord y,color c) {
	uchar	oc;

	if (x<bm->x1 || x>=bm->x2 || y<bm->y1 || y>=bm->y2) return;

	oc = bm->vm[x+y*bm->bpr];
	bm->vm[x+y*bm->bpr]=FPick_Choose(fpcol,256
		,((ulong)piccol[oc*3+0] + (ulong)piccol[c*3+0])>>1
		,((ulong)piccol[oc*3+1] + (ulong)piccol[c*3+1])>>1
		,((ulong)piccol[oc*3+2] + (ulong)piccol[c*3+2])>>1
		);
}

void	VPplot_add(BMP bm,coord x,coord y,color c) {
	uchar	oc;

	if (x<bm->x1 || x>=bm->x2 || y<bm->y1 || y>=bm->y2) return;

	oc = bm->vm[x+y*bm->bpr];
	bm->vm[x+y*bm->bpr]=FPick_Choose(fpcol,256
		,min(255,(ulong)piccol[oc*3+0] + (ulong)piccol[c*3+0])
		,min(255,(ulong)piccol[oc*3+1] + (ulong)piccol[c*3+1])
		,min(255,(ulong)piccol[oc*3+2] + (ulong)piccol[c*3+2])
		);
}

void	VPplot_subtract(BMP bm,coord x,coord y,color c) {
	uchar	oc;

	if (x<bm->x1 || x>=bm->x2 || y<bm->y1 || y>=bm->y2) return;

	oc = bm->vm[x+y*bm->bpr];
	bm->vm[x+y*bm->bpr]=FPick_Choose(fpcol,256
		,max(0,(long)piccol[oc*3+0] - (long)piccol[c*3+0])
		,max(0,(long)piccol[oc*3+1] - (long)piccol[c*3+1])
		,max(0,(long)piccol[oc*3+2] - (long)piccol[c*3+2])
		);
}

void	VPplot_translate(BMP bm,coord x,coord y,coord c) {
	if (x<bm->x1 || x>=bm->x2 || y<bm->y1 || y>=bm->y2) return;

	bm->vm[x+y*bm->bpr] = VPplot_translate_table[bm->vm[x+y*bm->bpr]];
}

VPplotptr VPplotter_list[][2]={
	NULL,			NULL,
	VPplot_subtract,	VPplot_add,
	VPplot_transp,		VPplot_transp,
	VPplot_translate,	VPplot_translate,
};

void	VPpastebrush(BMP bm,coord x,coord y,coord c) {
	coord	xst,xs,xt;
	coord	ys;
	uchar	*dp;

	x -= brush_hx;
	y -= brush_hy;

	if (c>=0) {
		TAP_RECT(x,y);
		TAP_RECT(x+brush_sprite.width-1,y+brush_sprite.height-1);
	}

	if (!blitmode) {
		PasteBrush(bm,x,y);
		return;
	}

	xs = brush_sprite.width;
	ys = brush_sprite.height;
	dp = brush_sprite.imgbm.vm;

	if (x<=bm->x1-xs || y<=bm->y1-ys || x>=bm->x2 || y>=bm->y2) return;

	if (x<bm->x1) {
		dp += (bm->x1 - x);
		xs -= (bm->x1 - x);
		x = bm->x1;
	}
	if (x+xs > bm->x2)
		xs = bm->x2 - x;
	if (y<bm->y1) {
		dp += (bm->y1 - y)*brush_sprite.imgbm.bpr;
		ys -= (bm->y1 - y);
		y = bm->y1;
	}
	if (y+ys > bm->y2)
		ys = bm->y2 - y;

	if (!xs || !ys) return;

	do {
		xst=xs;
		xt = x;
		switch(brush_sprite.flags & SPR_DRAWMODE) {
		case SPR_MATTE:
			do {
				if (*dp!=brush_sprite.trans_colr) VPplot(bm,xt,y,*dp);
				++xt;
				++dp;
			} while(--xst);
			break;
		case SPR_NOPATTERN:
			do {
				if (*dp++ != brush_sprite.trans_colr) VPplot(bm,xt,y,brush_sprite.colr);
				++xt;
			} while(--xst);
			break;
		case SPR_REPLACE:
			do
				VPplot(bm,xt++,y,*dp++);
			while(--xst);
			break;
		case SPR_NOIMAGE:
			dp += xst;
			do
				VPplot(bm,xt++,y,brush_sprite.colr);
			while(--xst);
			break;
		}
		dp += (brush_sprite.imgbm.bpr - xs);
		++y;
	} while(--ys);
}

void	VPhline(BMP bm,coord x1,coord x2,coord y,color c) {
	TAP_RECT(x1,y);
	TAP_RECT(x2,y);

	while(x1<=x2)
		VPplot(bm,x1++,y,c);
}

void	VPvline(BMP bm,coord x,coord y1,coord y2,color c) {
	TAP_RECT(x,y1);
	TAP_RECT(x,y2);

	while(y1<=y2)
		VPplot(bm,x,y1++,c);
}

///////////////////////////////////////////////////////

TOOL	doDotDraw,doContinuous,doLine,doSpray,doRect,doCircle,
	doEllipse,doBrush,doFillRect,doFillCircle,doFillEllipse;

TOOLDRG	dorept,dodrawline,dodrag,dobrush,docurve,dofill,dorept2,dopoly,dotext;

void	doCurve3pt(coord x1,coord y1,coord x2,coord y2,coord x3,coord y3,color color,boolean preview);

////////

struct drawVectors {
	uchar taps_rect;
	TOOLDRG *drawIt;
	TOOL *draw,*draw_dbl;
} drawVectors[]={
	{ TRUE	,&dorept    ,&doDotDraw			,NULL },
	{ TRUE	,&dodrawline,&doContinuous		,NULL },
	{ TRUE	,&dodrag    ,&doLine			,NULL },
	{ FALSE	,&docurve   ,(TOOL *)&doCurve3pt	,NULL },
	{ FALSE	,&dofill    ,NULL			,NULL },
	{ FALSE ,&dorept2   ,&doSpray			,NULL },
	{ TRUE	,&dodrag    ,&doRect			,&doFillRect },
	{ TRUE	,&dodrag    ,&doCircle			,&doFillCircle },
	{ TRUE	,&dodrag    ,&doEllipse			,&doFillEllipse },
	{ FALSE	,&dopoly    ,NULL			,(void *)1 },
	{ 2	,&dobrush   ,&doBrush			,NULL },
	{ FALSE	,&dotext    ,NULL			,NULL },
};

void dodraw(uchar colr) {
	struct drawVectors *dv;
	boolean reptsave,reptsaved=FALSE;
	ulong	brush_save_flags;

	if (!(VPplot = VPplotter_list[blitmode][!!colr])) VPplot = vd->plot;
	VPbrush = VPpastebrush;

	// create translate table for shade, etc.

	switch(blitmode) {
	case BMODE_SHADE:
		CCycle_Make_Translate_Table(VPplot_translate_table,!colr);
		break;
	}

	UndoInvalidate();
	OffBrush();

	/* remember to save current position */

	rectposx=vposx;
	rectposy=vposy;

	dabort=FALSE;
	dcolr=colr ? fcolor : bcolor;
	dcolrp=colr;
	brush_sprite.colr=dcolr;	// (uchar)ZRTool1[3].highindex;

	// check the color requested; if it's the background color, force
	// any existing custom brush to the COLOR mode for this operation

	if (!colr) {
		brush_save_flags = brush_sprite.flags & SPR_DRAWMODE;
		switch(brush_save_flags) {
		case SPR_NOPATTERN:	break;
		case SPR_MATTE:		BrushMode(1);	break;
		case SPR_REPLACE:	BrushMode(3);	break;
		}
		OffBrush();
	}

	if (ZRTool1[0].highindex==10) {
		dinvert=-1;
		reptsave=opt.are_repting;
		opt.are_repting=FALSE;
		reptsaved=TRUE;
	} else dinvert =(opt.are_repting=((!opt.reptdraw)^!(kbd_shift() & KBD_CTRL)))? dcolr : -1;

//	br_plotback=vd->plot;
//	vd->plot=&PasteBrush2;
	dv=&drawVectors[ZRTool1[0].highindex];

	if (dv->taps_rect == 2) {	// is it the brush_pickup?
		dcolr=-1;
	};

	(*dv->drawIt)(ZRTool1[0].doubleClick ? dv->draw_dbl : dv->draw,opt.fastPreview);
	mouse_wait_unkey();

	switch(dv->taps_rect) {
		case 2: break;
		case FALSE:
			rectx1=winbm->x1;
			recty1=winbm->y1;
			rectx2=winbm->x2-1;
			recty2=winbm->y2-1;
		case TRUE:
			rectmodified=TRUE;
	};
//	vd->plot=br_plotback;

	if (reptsaved) opt.are_repting=reptsave;

	ms_xscroll=ms_yscroll=0;

	vid_tap_rect2(winbm,rectx1,recty1,rectx2,recty2);
	UndoSave();
//	vd->refresh();

	// restore brush mode

	if (!colr) {
#if 0
		brush_sprite.flags &= ~SPR_DRAWMODE;
		switch(brush_save_flags) {
		case SPR_MATTE:		BrushMode(0); break;
		case SPR_NOPATTERN:	BrushMode(1); break;
		case SPR_REPLACE:	BrushMode(2); break;
		case SPR_NOIMAGE:	BrushMode(3); break;
		}
#else
		brush_sprite.flags = (brush_sprite.flags & ~SPR_DRAWMODE) | brush_save_flags;
#endif
	}
	if (dv->taps_rect == 2)
		brush_sprite.flags = (brush_sprite.flags & ~SPR_DRAWMODE) | SPR_MATTE;
}

void dodrag(TOOL *dfunc,boolean preview) {
	mouse_coord x1,y1,x,y,b,ox,oy;

	if (!dfunc) { mouse_wait_unkey(); return; };

	mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);

	x1=x;
	y1=y;
	mouse_hide();
	do {
		ox=x;
		oy=y;
		(*dfunc)(x1,y1,x,y,dcolr,preview);
		vd->refresh();
		mouse_show();
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && b && !kbhit());
		mouse_hide();

		if (!opt.are_repting) draw_undo();
		else draw_color_shift();
		CheckKeyInterrupt();
		if (dabort) goto abort;
	} while(b);
	(*dfunc)(x1,y1,x,y,dcolr==-1 ? 0 : dcolr,FALSE);
abort:
	mouse_show();
}

void dobrush(TOOL *dfunc,boolean preview) {
	mouse_coord x1,y1,x,y,b,ox,oy;

	if (!dfunc) { mouse_wait_unkey(); return; };

	mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);

	x1=x;
	y1=y;
	mouse_hide();
	do {
		ox=x;
		oy=y;

		(*dfunc)(x1,y1,x,y,dcolrp,TRUE);

		vd->refresh();
		mouse_show();
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && b && !kbhit());
		mouse_hide();

		if (!opt.are_repting) draw_undo();
		else draw_color_shift();
		CheckKeyInterrupt();
		if (dabort) goto abort;
	} while(b);

	(*dfunc)(x1,y1,x,y,dcolrp,FALSE);
abort:
	mouse_show();
}

void dodrawline(TOOL *dfunc,boolean preview) {
	mouse_coord x,y,b,ox,oy;

	if (!dfunc) { do_filled_outline(dcolr); return; };

	mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);

	do {
		ox=x; oy=y;
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && b && !kbhit());
		mouse_hide();
		(*dfunc)(ox,oy,x,y,dcolr,FALSE);
		mouse_show();
		draw_color_shift();
		CheckKeyInterrupt();
		if (dabort) return;
	} while(b);
}

void dorept(TOOL *dfunc,boolean preview) {
	mouse_coord x,y,b;
	coord ox=-0x8000,oy=-0x8000;
	boolean is_first = TRUE;

	if (!dfunc) { mouse_wait_unkey(); return; };

	do {
		mouse_hide();
		if (!is_first)
			(*dfunc)(x,y,x,y,dcolr,FALSE);
		mouse_show();
		is_first = FALSE;
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while(ox==x && oy==y && b && !kbhit());
		ox=x; oy=y;
		draw_color_shift();
		CheckKeyInterrupt();
		if (dabort) return;
	} while(b);
}

void dorept2(TOOL *dfunc,boolean preview) {
	mouse_coord x,y,b;

	if (!dfunc) { mouse_wait_unkey(); return; };

	do {
		mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		(*dfunc)(x,y,x,y,dcolr,FALSE);
		draw_color_shift();
		CheckKeyInterrupt();
		if (dabort) return;
	} while(b);
}

//////////////////////////////////

struct VPPointList {
	struct VPPointList *next;
	coord	x,y;
};

void dopoly(TOOL *dfunc,boolean preview) {
	struct VPPointList *vppl = NULL;
	struct VPPointList *vpp;
	coord x1,y1,x2,y2,ox,oy;
	mouse_coord x,y,b;
	coord key;
	boolean oflag;
	int	point_count = 1;

	/////////

	do {
		mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
	} while(b);
	x1=x2=x; y1=y2=y;

	mouse_hide();

	// add the point (x,y) to the polygon list

	if (dfunc) {
		if (!(vpp = malloc(sizeof(struct VPPointList)))) goto abort1;
		vpp->next = NULL;
		vpp->x = x;
		vpp->y = y;
		vppl = vpp;
	}

	mouse_wait_unkey();

	for(;;) {
		vd->line(winbm,x2,y2,x,y,dfunc ? -1 : dinvert);
		mouse_show();
		if (opt.are_repting) draw_color_shift();
		ox=x; oy=y;
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
			key = CheckKeyInterrupt();
		} while(ox==x && oy==y && !b && key!=' ' && key!=0x1b);
		mouse_hide();
		if (b) mouse_wait_unkey();
		oflag=abs(x1-x)<4*XPOINT && abs(y1-y)<4*YPOINT;
//		if (dfunc ? (!b || oflag || key==' ' || key==0x1b) : !opt.are_repting) vd->line(winbm,x2,y2,ox,oy,-1);

		if (dfunc || !opt.are_repting) vd->line(winbm,x2,y2,ox,oy,-1);

		if (key==' ' || key==0x1b) {
			if (key==' ' && !dfunc)
				doLine2(x2,y2,x1,y1,dcolr,TRUE,FALSE);
			break;
		}

		if (b) {
			if (oflag && dfunc) break;
			if (oflag) { x=x1; y=y1; };
			if (dfunc) {

				// draw line

				vd->line(winbm,x2,y2,x,y,-1);

				// add vertex to list

				if (!(vpp = malloc(sizeof(struct VPPointList)))) goto abort1;

				vpp->next	= vppl;
				vpp->x		= x;
				vpp->y		= y;
				vppl		= vpp;

				++point_count;
			} else
				doLine2(x2,y2,x,y,dcolr,TRUE,FALSE);
			x2=x; y2=y;
			if (oflag) break;
		};
	}
	if (dfunc) {
		struct PolyVertex *point_table,*ptp;
		int i;

		// change point list into a point table

		if (!(point_table = malloc(sizeof(struct PolyVertex)*point_count))) goto abort1;

		// be sure to copy the points in REVERSE order

		ptp = point_table + point_count - 1;
		vpp = vppl;
		while(vpp) {
			ptp->x = vpp->x;
			ptp->y = vpp->y;
			--ptp;
			vpp=vpp->next;
		}

		// free point list

		while(vpp=vppl) {
			vppl = vppl->next;
			free(vpp);
		}

		// erase the lines in backward order

		for(i=0;i<point_count-1;i++)
			vd->line(winbm,point_table[i].x,point_table[i].y,point_table[i+1].x,point_table[i+1].y,-1);

		// render!!!!

		VPpolygon_convex(winbm,point_table,point_count,dcolr);

		free(point_table);
	};
	mouse_show();
	mouse_wait_unkey();
	return;

abort1:
	// free the list

	while(vpp=vppl) {
		vppl = vppl->next;
		free(vpp);
	}

	mouse_show();
	mouse_wait_unkey();
	Message("Not enough memory!");
}

//////////////////////////////////

void dofill(TOOL *dfunc,boolean preview) {
	mouse_coord x,y,b;
	do {
		mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		CheckKeyInterrupt();
		if (dabort) return;
	} while(b);
	doFloodFill(x,y,x,y,dcolr,FALSE);
}

typedef	void	(*CURVETOOL)(coord,coord,coord,coord,coord,coord,color,boolean);

void docurve(TOOL *dfunc,boolean preview) {
	coord x1,y1,x2,y2,ox,oy;
	mouse_coord x,y,b;
	coord clicked_yet=0;

	if (!dfunc) { mouse_wait_unkey(); return; };

	mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);

	x1=x;
	y1=y;
	mouse_hide();
	do {
		ox=x;
		oy=y;
		vd->line(winbm,x1,y1,x,y,-1);
		vd->refresh();
		mouse_show();
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && b && !kbhit());
		mouse_hide();
		vd->line(winbm,x1,y1,ox,oy,-1);
		CheckKeyInterrupt();
		if (dabort) return;
	} while(b);
	x2=x;
	y2=y;
	for(;;) {
		ox=x;
		oy=y;
		(*(CURVETOOL)dfunc)(x1,y1,x,y,x2,y2,dinvert,preview);
		vd->refresh();
		mouse_show();

		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && ((clicked_yet == 0 && !b) || (clicked_yet == 1 && b)) && !kbhit());
		mouse_hide();
		if (!opt.are_repting) (*(CURVETOOL)dfunc)(x1,y1,ox,oy,x2,y2,-1,preview);
		else draw_color_shift();
		if (b) {
			if (clicked_yet==0) clicked_yet=1;
		} else if (clicked_yet==1) break;
		CheckKeyInterrupt();
		if (dabort) goto abort;
	};
	(*(CURVETOOL)dfunc)(x1,y1,x,y,x2,y2,dcolr,FALSE);
abort:
	mouse_show();
}

void drawTextBox(coord x,coord y,coord c) {
	vd->box(winbm,x,y,x+5,y+7,c);
	vd->hline(winbm,x+1,x+4,y+6,c);
}

#pragma aux easter_detection "*_"
extern uchar easter_detection[];

void dotext(TOOL *func,boolean preview) {
	coord ox,oy,x=-1,y=-1,b,key;
	boolean curs_state=FALSE;
	ulong cur_tick,last_tick=read_tick();
	char	keys[2]="\0";
	long e_state=easter_decode?0:-1,kc=0;

	mouse_hide();
	do {
		ox=x; oy=y;
		drawTextBox(x,y,-1);
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while(ox==x && oy==y && b);
		drawTextBox(ox,oy,-1);
	} while(b);
	mouse_show();

//	drawTextBox(ox=x,oy=y,-1);

	for(;;) {
		cur_tick=read_tick();
		if (cur_tick-last_tick >= 5) {
			last_tick=cur_tick;
			curs_state=!curs_state;
			mouse_hide();
			drawTextBox(x,y,-1);
			mouse_show();
		};
		vd->refresh();
		if (mouse_buttons()) break;
		switch(key=kbd_check()) {
		case '\x1b': goto double_exit;	/* where's my doublebreak! */
		case 0x148:	if (y>winbm->y1) {
					if (curs_state) {
						mouse_hide();
						drawTextBox(x,y,-1);
						mouse_show();
						curs_state=FALSE;
					};
					y=max(winbm->y1,y-8);
				};
				break;
		case '\r':
			if (curs_state) {
				mouse_hide();
				drawTextBox(x,y,-1);
				mouse_show();
				curs_state=FALSE;
			};
			x=ox;
			/* fall through and go down a line */
		case 0x150:	if (y<winbm->y2-1) {
					if (curs_state) {
						mouse_hide();
						drawTextBox(x,y,-1);
						mouse_show();
						curs_state=FALSE;
					};
					y=min(winbm->height-1,y+8);
				};
				break;
		case 0x08:	if (curs_state) {
					mouse_hide();
					drawTextBox(x,y,-1);
					mouse_show();
					curs_state=FALSE;
				};
				vd->bitblt(picbm,x-6,y,winbm,x-6,y,6,8);

				if (kc>0) --kc;

		case 0x14B:	if (x>winbm->x1) {
					if (curs_state) {
						mouse_hide();
						drawTextBox(x,y,-1);
						mouse_show();
						curs_state=FALSE;
					};
					x=max(0,x-6);
				};
				break;
		case 0x14D:	if (x<winbm->x2) {
					if (curs_state) {
						mouse_hide();
						drawTextBox(x,y,-1);
						mouse_show();
						curs_state=FALSE;
					};
					x=min(winbm->x2-1,x+6);
				};
				break;
		case -1: break;
		default:
			if (key>=0 && key<0x100) {
				if (e_state==kc) {
					if (easter_detection[kc]==key)
						if (++e_state==32) {
							dcd_e=TRUE;
							goto double_exit;
						}
				}

//				printf("e_state %d, kc %d, key %c\n",e_state,kc,easter_detection[kc]);

				++kc;
				mouse_hide();
				if (curs_state) {
					curs_state=FALSE;
					drawTextBox(x,y,-1);
				};
				keys[0]=key;
				vd->text6t(winbm,x,y,keys,dcolr);
				mouse_show();
				if ((x+=6)>=winbm->x2) {
					x=ox;
					if (y<(winbm->y2-8)) y+=8;
				};
				draw_color_shift();
			};
			break;
		};
	}

double_exit:

	if (curs_state) drawTextBox(x,y,-1);
}

TOOLDEFINE(doDotDraw) {
	VPbrush(winbm,x1,y1,colr);
	TAP_RECT(x1,y1);
}

TOOLDEFINE(doContinuous) {
//	vd->line(winbm,x1,y1,x2,y2,colr);
	doLine(x1,y1,x2,y2,colr,preview);
}

TOOLDEFINE(doSpray) {
	int i;
	static ulong tick;
	ulong tick2;
	coord x,y;

	tick2 = *(ulong *)0x46c;
	if (tick2 != tick) {
		mouse_hide();
		for (i=16;i;i--) {
			for(;;) {
				x = (((ulong)rand() * 31)>>15)-15;
				if (vpopts.airbrush.shape != VP_AIRBRUSH_LSLASH && vpopts.airbrush.shape != VP_AIRBRUSH_RSLASH)
					y = (((ulong)rand() * 31)>>15)-15;

				switch(vpopts.airbrush.shape) {
				case VP_AIRBRUSH_CIRCLE:
					// reject the point if it is too far off

					if (x*x+y*y>15*15) continue;
					VPbrush(winbm,x1+x,y1+y,colr);
					goto dot_plotted;

				case VP_AIRBRUSH_DIAMOND:

					// if the point is outside the diamond, "fold it" back on the opposite side

					if (abs(x)+abs(y) > 15) {
						if (x<0) x+=15; else x-=15;
						if (y<0) y+=15; else y-=15;
					}
					VPbrush(winbm,x1+x,y1+y,colr);
					goto dot_plotted;

				case VP_AIRBRUSH_SQUARE:
					VPbrush(winbm,x1+x,y1+y,colr);
					goto dot_plotted;

				case VP_AIRBRUSH_LSLASH:
					VPbrush(winbm,x1+x,y1+x,colr);
					goto dot_plotted;

				case VP_AIRBRUSH_RSLASH:
					VPbrush(winbm,x1+x,y1-x,colr);
					goto dot_plotted;
				}
			}
dot_plotted:
			;
		}
		mouse_show();
		tick = tick2;
	}
}

//	Warning: doLine() is called directly from dopoly()

void doLine(coord x1,coord y1,coord x2,coord y2,coord colr,boolean preview) {
	doLine2(x1,y1,x2,y2,colr,FALSE,preview);
}

void doLine2(coord x1,coord y1,coord x2,coord y2,coord colr,boolean no_first_dot,boolean preview) {
//	if (colr>=0) {
	if (!preview) {
		switch(linemode) {
		case LMODE_NORMAL:
			vid_line(winbm,x1,y1,x2,y2,colr,VPbrush,no_first_dot);
			break;
		case LMODE_ANTIALIAS:
			vid_antialias_line(winbm,x1,y1,x2,y2,colr,bcolor,FALSE);
			break;
		case LMODE_AABLEND:
			vid_antialias_line(winbm,x1,y1,x2,y2,colr,bcolor,TRUE);
			break;
		};
	} else
		vd->line(winbm,x1,y1,x2,y2,colr);
	TAP_RECT(x1,y1);
	TAP_RECT(x2,y2);
}

void doRect(coord x1,coord y1,coord x2,coord y2,coord colr,boolean preview) {
	if (preview) {
		vd->box(winbm,x1,y1,x2,y2,colr);
	} else {
		if (x1>x2) swap(coord,x1,x2);
		if (y1>y2) swap(coord,y1,y2);

		if (x1==x2) VPvline(winbm,x1,y1,y2,colr);
		else if (y1==y2) VPhline(winbm,x1,x2,y1,colr);
		else {
			VPhline(winbm,x1,x2-1,y1  ,colr);
			VPvline(winbm,x2,y1  ,y2-1,colr);
			VPhline(winbm,x1+1,x2,y2  ,colr);
			VPvline(winbm,x1,y1+1,y2,colr);
		}
	}
	TAP_RECT(x1,y1);
	TAP_RECT(x2,y2);
}

long	fixed16scale32i(long i,long s);
#pragma aux fixed16scale32i="imul ebx","shrd eax,edx,16" parm [eax] [ebx] value [eax] modify exact nomemory [eax edx];

void doCircle(coord x1,coord y1,coord x2,coord y2,coord colr,boolean preview) {
	long r;
	coord xs,ys;

	x2-=x1;
	y2-=y1;
	if (opt.be_square) y2=fixed16scale32i(y2,x_ratio);
	r=sqrt((long)x2*x2+(long)y2*y2);
	if (opt.be_square && x_ratio != 0x10000L) {
		xs = r;
		ys=fixed16scale32i(r,y_ratio);
		doEllipse(x1,y1,x1+xs,y1+ys,colr,preview);
		return;
#if 0
		if (preview)
			vd->ellipse(winbm,x1,y1,xs,ys,colr);
		else
			vid_ellipse_func(winbm,x1,y1,xs,ys,colr,VPbrush);
#endif
	} else {
		xs=ys=r;
		switch(linemode) {
		case LMODE_NORMAL:

			if (brush_sprite.width>1 || brush_sprite.height>1 || blitmode)
				vid_circle_func(winbm,x1,y1,r,colr,VPbrush);
			else
				vd->circle(winbm,x1,y1,r,colr);
			break;
		case LMODE_ANTIALIAS:	vid_antialias_circle(winbm,x1,y1,r,colr,bcolor,FALSE); break;
		case LMODE_AABLEND:	vid_antialias_circle(winbm,x1,y1,r,colr,bcolor,TRUE); break;
		}
	}

	TAP_RECT(x1-xs-brush_hx                   ,y1-ys-brush_hy                    );
	TAP_RECT(x1+xs-brush_hx+brush_sprite.width,y1+ys-brush_hy+brush_sprite.height);
}

TOOLDEFINE(doEllipse) {
	x2=abs(x2-x1);
	y2=abs(y2-y1);
	if (preview)
		vd->ellipse(winbm,x1,y1,abs(x2),abs(y2),colr);
	else {
		switch(linemode) {
		case LMODE_NORMAL:
			vid_ellipse_func(winbm,x1,y1,abs(x2),abs(y2),colr,VPbrush);
			break;
		case LMODE_ANTIALIAS:
			vid_antialias_ellipse(winbm,x1,y1,abs(x2),abs(y2),colr,bcolor,FALSE,FALSE);
			break;
		case LMODE_AABLEND:
			vid_antialias_ellipse(winbm,x1,y1,abs(x2),abs(y2),colr,bcolor,TRUE,FALSE);
			break;
		}
	}
	TAP_RECT(x1-x2-brush_hx                   ,y1-y2-brush_hy                    );
	TAP_RECT(x1+x2-brush_hx+brush_sprite.width,y1+y2-brush_hy+brush_sprite.height);
}

TOOLDEFINE(doFillRect) {
	coord y;

	if (x1>x2) swap(coord,x1,x2);
	if (y1>y2) swap(coord,y1,y2);

	TAP_RECT(x1,y1);
	TAP_RECT(x2,y2);
	if (!preview) {
		switch(fillmode) {
		case FILL_TILE:
			vd->tile(&brush_sprite.imgbm,winbm,x1,y1,x2+1-x1,y2+1-y1,x1+vposx+brush_ox,y1+vposy+brush_oy);
			return;
		case FILL_STRETCH:
			{
				y=y2+1-y1;
				if (!vsd->init(winbm,x1,y1,&brush_sprite.imgbm,x2+1-x1,y))
					return;

				do {
					vsd->rec_hline(x1,x2,y1+y-1);
				} while(--y);

				vsd->map_init();

				y=y2-y1;
				do {
					vsd->map_hline(winbm,&brush_sprite.imgbm,x1,x2,y1++);
				} while(y--);
				vsd->dispose();
			}
			return;
		default:
			if (blitmode) {
				while(y1<=y2) {
					VPhline(winbm,x1,x2,y1,colr);
					++y1;
				}
			} else
				vd->bar(winbm,x1,y1,x2+1-x1,y2+1-y1,colr);
		};
	} else	vd->bar(winbm,x1,y1,x2+1-x1,y2+1-y1,colr);
}

TOOLDEFINE(doFillCircle) {
	long rsq,rsq2;
	long rsqrt,y;
	long fmode=(colr>=0)?fillmode:FILL_NORMAL;
	long xoff,yoff;
	static boolean state=FALSE;

	x2-=x1;
	if (x2<0) x2=-x2;
	y2-=y1;
	if (y2<0) y2=-y2;

	if (opt.be_square && x_ratio != 0x10000L) {
		long r;

//		y2=(y2*10L)/9;
		y2=fixed16scale32i(y2,x_ratio);
		r=sqrt((long)x2*x2+(long)y2*y2);
//		doFillEllipse(x1,y1,x1+r,y1+(r*9L)/10,colr);
		doFillEllipse(x1,y1,x1+r,y1+fixed16scale32i(r,y_ratio),colr,preview);
		return;
	};

	rsq=(ulong)x2*x2+(ulong)y2*y2;
	rsqrt=ceil(sqrt(rsq));

	TAP_RECT(x1-rsqrt,y1-rsqrt);
	TAP_RECT(x1+rsqrt,y1+rsqrt);

	switch(linemode) {
	case LMODE_ANTIALIAS:
		vid_antialias_ellipse(winbm,x1,y1,rsqrt,rsqrt,colr,bcolor,FALSE,TRUE);
		return;
	case LMODE_AABLEND:
		vid_antialias_ellipse(winbm,x1,y1,rsqrt,rsqrt,colr,bcolor,TRUE,TRUE);
		return;
	}

	y=0;

	if (!state && fmode==FILL_STRETCH) {
		if (!vsd->init(winbm,x1-rsqrt,y1-rsqrt,&brush_sprite.imgbm,rsqrt*2+1,rsqrt*2+1))
			return;
		state=TRUE;
		doFillCircle(x1,y1,x1+x2,y1+y2,colr,preview);
		state=FALSE;
		vsd->map_init();

		y=-rsqrt;
	}

	xoff=x1+vposx+brush_ox;
	yoff=y1+vposy+brush_oy;

	for (;y<=rsqrt;y++) {
		if ((rsq2=rsq-y*y)>0) {
			rsq2=sqrt(rsq2);
			switch(fmode) {
			case FILL_TILE:
				if (y)
					vd->tile(&brush_sprite.imgbm,winbm,x1-rsq2,y1-y,rsq2*2+1,1,xoff-rsq2,yoff-y);
				vd->tile(&brush_sprite.imgbm,winbm,x1-rsq2,y1+y,rsq2*2+1,1,xoff-rsq2,yoff+y);
				break;
			case FILL_STRETCH:
				if (state) {
					if (y)
						vsd->rec_hline(x1-rsq2,x1+rsq2,y1-y);
					vsd->rec_hline(x1-rsq2,x1+rsq2,y1+y);
				} else {
					vsd->map_hline(winbm,&brush_sprite.imgbm,x1-rsq2,x1+rsq2,y1+y);
				};
				break;
			default:
				if (preview) {
					if (y)
						vd->hline(winbm,x1-rsq2,x1+rsq2,y1-y,colr);
					vd->hline(winbm,x1-rsq2,x1+rsq2,y1+y,colr);
				} else {
					if (y)
						VPhline(winbm,x1-rsq2,x1+rsq2,y1-y,colr);
					VPhline(winbm,x1-rsq2,x1+rsq2,y1+y,colr);
				}
			};
		};
	};
	if (fmode==FILL_STRETCH && !state)
		vsd->dispose();
}

/* xsq/asq+ysq/bsq=1
   xsq*bsq/asq+ysq=bsq
   y=sqrt(bsq-xsq*bsq/asq) */

TOOLDEFINE(doFillEllipse) {
	long asq,bsq;
	long x,y;
	long xoff,yoff;
	long fmode=(colr>=0)?fillmode:FILL_NORMAL;
	static boolean state=FALSE;

	x2-=x1;
	y2-=y1;
	if (x2<0) x2=-x2;
	if (y2<0) y2=-y2;

	if (!x2 || !y2) return;

	TAP_RECT(x1-x2,y1-y2);
	TAP_RECT(x1+x2,y1+y2);

	switch(linemode) {
	case LMODE_ANTIALIAS:
		vid_antialias_ellipse(winbm,x1,y1,x2,y2,colr,bcolor,FALSE,TRUE);
		return;
	case LMODE_AABLEND:
		vid_antialias_ellipse(winbm,x1,y1,x2,y2,colr,bcolor,TRUE,TRUE);
		return;
	}

	asq=(long)x2*x2;
	bsq=(long)y2*y2;

	y=0;

	if (!state && fmode==FILL_STRETCH) {
		if (!vsd->init(winbm,x1-x2,y1-y2,&brush_sprite.imgbm,x2*2+1,y2*2+1))
			return;
		state=TRUE;
		doFillEllipse(x1,y1,x1+x2,y1+y2,colr,preview);
		state=FALSE;
		vsd->map_init();

		y=-y2;
	};

	xoff=x1+vposx+brush_ox;
	yoff=y1+vposy+brush_oy;

	for (;y<=y2;y++) {
//		if ((rsq=asq-((long)y*y*asq)/bsq)>0) {
//			x=sqrt(rsq);

		if (labs(y)<=y2) {
			x=x2*sqrt(1.0-(float)((long)y*y)/(float)bsq);

			switch(fmode) {
			case FILL_TILE:
				if (y)
					vd->tile(&brush_sprite.imgbm,winbm,x1-x,y1-y,x*2+1,1,xoff-x,yoff-y);
				vd->tile(&brush_sprite.imgbm,winbm,x1-x,y1+y,x*2+1,1,xoff-x,yoff+y);
				break;
			case FILL_STRETCH:
				if (state) {
					if (y)
						vsd->rec_hline(x1-x,x1+x,y1-y);
					vsd->rec_hline(x1-x,x1+x,y1+y);
				} else {
					vsd->map_hline(winbm,&brush_sprite.imgbm,x1-x,x1+x,y1+y);
				};
				break;
			default:
				if (preview) {
					if (y)
						vd->hline(winbm,x1-x,x1+x,y1-y,colr);
					vd->hline(winbm,x1-x,x1+x,y1+y,colr);
				} else {
					if (y)
						VPhline(winbm,x1-x,x1+x,y1-y,colr);
					VPhline(winbm,x1-x,x1+x,y1+y,colr);
				}
			};
		};
	};

	if (fmode==FILL_STRETCH && !state)
		vsd->dispose();
}

void doCurve3pt(coord x1,coord y1,coord x2,coord y2,coord x3,coord y3,coord colr,boolean preview) {
	coord x,y;
	coord t2,t;
	coord co0,co1,co2;
	coord ox,oy;

	x=(x1+x3+1)>>1;	y=(y1+y3+1)>>1;
	x2=(x2<<1)-x;	y2=(y2<<1)-y;

	ox=-1;
	oy=-1;

	for (t=0;t<16384;t+=24) {
		t2=((long)t*t)>>14;
		co0=t2-t-t+16384;
		co1=(t-t2)<<1;
		co2=t2;
		x=((long)co0*x1+(long)co1*x2+(long)co2*x3+8192)>>14;
		y=((long)co0*y1+(long)co1*y2+(long)co2*y3+8192)>>14;
		if (ox!=x || oy!=y) {
			if ((abs(ox-x)>1 || abs(oy-y)>1) && t)
				vd->line(winbm,ox,oy,x,y,colr);
			else
				vd->plot(winbm,x,y,colr);
			ox=x; oy=y;
		};
	}
}

///////////////////////////////////////////////////////////

void do_filled_outline(color colr) {
	struct VPPointList *vppl=NULL,*vpp;
	struct PolyVertex *pvl,*pvlt;
	coord x1,y1;
	mouse_coord x,y,b,ox,oy;
	int vertex_count = 1,i;

	mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
	x1 = x;
	y1 = y;

	do {
		ox=x; oy=y;
		do {
			mouse_read_draw(&x,&y,&b,MRD_GRID+MRD_DOSPLIT+MRD_USEZOOMFLG+MRD_USESHIFT);
		} while (ox==x && oy==y && b);

		mouse_hide();
		vd->line(winbm,ox,oy,x,y,-1);
		vd->plot(winbm,ox,oy,-1);
		mouse_show();

		// add vertex (ox,oy)

		if (!(vpp = malloc(sizeof(struct VPPointList)))) goto abort1;

		vpp->next = vppl;
		vppl = vpp;
		vpp->x = x;
		vpp->y = y;

		++vertex_count;
	} while(b);

	// allocate memory for the point list and fill it in

	if (!(pvl = malloc(sizeof(struct PolyVertex)*vertex_count))) goto abort1;

	pvlt = pvl + vertex_count - 1;
	vpp = vppl;
	while(vpp) {
		pvlt->x = vpp->x;
		pvlt->y = vpp->y;

		--pvlt;
		vpp = vpp->next;
	}

	pvlt->x = x1;
	pvlt->y = y1;

	mouse_hide();
	for(i=0;i<vertex_count-1;i++) {
		vd->plot(winbm,pvl[i].x,pvl[i].y,-1);
		vd->line(winbm,pvl[i].x,pvl[i].y,pvl[i+1].x,pvl[i+1].y,-1);
	}

	VPpolygon_convex(winbm,pvl,vertex_count,dcolr);
	mouse_show();

	free(pvl);

	return;

abort1:
	// free point list

	while(vpp=vppl) {
		vppl = vppl->next;
		free(vpp);
	}

	mouse_wait_unkey();
	Message("Not enough memory for point list.");
}

////////////////////////////////////////////////////////////

TOOLDEFINE(doBrush) {

	// if adjustBrush is on (ExclBrush to DPaint users), adjust
	// (x2,y2) if (x2,y2)>(x1,y1) and grid is on

	if (colr) {
		if (gridmode && opt.adjustBrush) {
			if (x2>x1) --x2;
			if (y2>y1) --y2;
		}
	} else {
		x1 = x2 - brush_hx;
		y1 = y2 - brush_hy;
		x2 = x1 + brush_bm.width - 1;
		y2 = y1 + brush_bm.height- 1;
	}

	if (preview) {
		vd->box(winbm,x1,y1,x2,y2,-1);
		TAP_RECT(x1,y1);
		TAP_RECT(x2,y2);
	} else {
		if (colr)
			GetBrush(x1,y1,x2,y2);
		else {
			vd->bitblt(winbm,x1,y1,&brush_bm,0,0,x2+1-x1,y2+1-y1);
			PictureMarkDirty(&brush_picture);
		}

		SetTitle(NULL);
		change_paint_tool(DTOOL_DOTDRAW);
	};
}


////////// CHANGE PAINT TOOL


void	change_paint_tool(ulong tool_id) {
	struct Zone *z;
	struct ZoneRange *zr;
	boolean brushsave;
	boolean dblclick = FALSE;
	struct BitMap *old_zonebm;

	if (tool_id & DTOOL_DBL) {
		tool_id &= ~DTOOL_DBL;
		dblclick = TRUE;
	}

	if (tool_id >= 18) return;

	if (tool_id < 12) {
		z = ZTool1 + tool_id;
		zr = ZRTool1;
		if (tool_id == ZRTool1[0].highindex)
			if (dblclick == zr->doubleClick) return; else dblclick=0;
	} else if (tool_id < 14) {
		tool_id -= 12;
		z = ZTool2 + tool_id;
		zr = ZRTool1 + 1;
	} else {
		tool_id -= 14;
		z = ZTool3 + tool_id;
		zr = ZRTool1 + 2;
	}

	old_zonebm = zonebm;
	if (!toolgrid)	// hack the zone bitmap to avoid corruption on screen
		zonebm = &invizbm;

	brushsave = brush;
	brush=FALSE;
	ZoneHit(zr,z,tool_id,0,0);
	if (dblclick) ZoneHit(zr,z,tool_id,0,0);
	brush=brushsave;

	zonebm = old_zonebm;
}


/********* GRIDDING *********/

void snapToGrid(mouse_coord *x,mouse_coord *y,int xoffset,int yoffset) {
	if (gridmode) {
		*x=grid.xoff+((*x+(grid.xspc>>1)-grid.xoff+xoffset)/grid.xspc)*grid.xspc-xoffset;
		*y=grid.yoff+((*y+(grid.yspc>>1)-grid.yoff+yoffset)/grid.yspc)*grid.yspc-yoffset;
	};
}

/********* DRAWING COLOR MODES *********/

/* draw_color_shift: shift colors once each time an item is drawn (used for
**		     range shift modes) */

void draw_color_shift(void) {
	struct CycleRange *range = &cyclerange[dcrange];
	struct CycleStrip *strip = &range->strip[dcstrip];

	/* don't do ANY color shifting if not using foreground color... */

	if (!dcolr) return;

	switch(dcmode) {
	case DC_RANGEFORW:

		/* first, check to see if we have ANY strips... */

		if (!range->strips) return;

		/* advance color in strip, then check to see if we have overrun
		** the strip... */

		if (strip->start > strip->end) {
			dcolr--;
			if (dcolr<strip->end || dcolr==strip->start) {	/* must handle 255-0 strip */
				++strip;
				if (++dcstrip>=range->strips) {
					dcstrip=0;
					strip=&range->strip[0];
				};
				dcolr=strip->start;
			};
		} else {
			dcolr++;
			if (dcolr>strip->end || dcolr==strip->start) {	/* must handle 0-255 strip */
				++strip;
				if (++dcstrip>=range->strips) {
					dcstrip=0;
					strip=&range->strip[0];
				};
				dcolr=strip->start;
			};
		};
		break;
	case DC_RANGEBACK:
		/* first, check to see if we have ANY strips... */

		if (!range->strips) return;

		/* advance color in strip, then check to see if we have overrun
		** the strip... */

		if (strip->start > strip->end) {
			dcolr++;
			if (dcolr>strip->start || dcolr==strip->end) {	/* must handle 255-0 strip */
				--strip;
				if (--dcstrip<0) {
					dcstrip=range->strips-1;
					strip=&range->strip[dcstrip];
				};
				dcolr=strip->end;
			};
		} else {
			dcolr--;
			if (dcolr<strip->start || dcolr==strip->end) {	/* must handle 0-255 strip */
				--strip;
				if (--dcstrip<0) {
					dcstrip=range->strips-1;
					strip=&range->strip[dcstrip];
				};
				dcolr=strip->end;
			};
		};
		break;
	};
	if (opt.are_repting) dinvert=dcolr;
}


///////////////////// COORDINATE READING //////////////////////

void	mouse_read_draw(mouse_coord *x,mouse_coord *y,mouse_coord *b,int flags) {
	mouse_coord xt,yt;
	static boolean zoom_flag = FALSE;
	static boolean last_shift_flag = FALSE;
	static boolean constrain_active = FALSE;
	static boolean constrain_use_x;
	static mouse_coord constrain_coord_x,constrain_coord_y;
	boolean	inside_zoom_window;
	boolean	shift_state;

	if (flags & MRD_WAITMOVE)
		mouse_readmove(x,y,b);
	else
		mouse_read2(x,y,b);

	if (flags & MRD_USESHIFT) {
		shift_state = !!(kbd_shift() & KBD_SHIFT);

		if (shift_state) {
			if (!last_shift_flag) {
				constrain_coord_x = *x;
				constrain_coord_y = *y;
			} else {
				if (constrain_active) {
					if (constrain_use_x) *y = constrain_coord_y;
					else *x = constrain_coord_x;
				} else {
					mouse_coord dx,dy;

					dx = abs(*x - constrain_coord_x);
					dy = abs(*y - constrain_coord_y);

					if (dx>dy && dx>5) {
						constrain_active = TRUE;
						constrain_use_x = TRUE;
					} else if (dy>dx && dy>5) {
						constrain_active = TRUE;
						constrain_use_x = FALSE;
					}
				}
			}
		} else constrain_active = FALSE;

		last_shift_flag = shift_state;
	}

	if (flags & MRD_DOSPLIT)
		if (*y<ms_mpi.split_2) *y+=vidbm.height;

	if (zoomstate) {
		xt=(*x)-zoombm->x1;
		yt=(*y)-zoombm->y1;
		inside_zoom_window = (xt>=0 && yt>=0 && xt<(zoombm->width*zoom_x) && yt<(zoombm->height*zoom_y));

		if (((flags & MRD_USEZOOMFLG) && zoom_flag) || (!(flags & MRD_USEZOOMFLG) && inside_zoom_window)) {
			*x=winbm->x1+zposx+((*x-zoombm->x1)/zoom_x)-vposx;
			*y=winbm->y1+zposy+((*y-zoombm->y1)/zoom_y)-vposy;
		}

		if (flags & MRD_SETZOOMFLG) zoom_flag = inside_zoom_window;
	}

	if (flags & MRD_GRID)
		snapToGrid(x,y,vposx-winbm->x1,vposy-winbm->y1);

	if (opt.coords)
		if (toolgrid) {
#if 0
			vid_printf6(&vidbm,tool_x-19-6*9,2,ST0,ST2,"%4d,%-4d",vposx+*x-winbm->x1,vposy+*y-winbm->y1);
#else
			static coord cpx=0L,cpy=0L;
			coord	cx,cy;

			cx = vposx + *x - winbm->x1;
			cy = vposy + *y - winbm->y1;
			
			if (cx!=cpx || cy!=cpy) {
				SysTextPrintf(&vidbm,toolpos.b_width-9*sysfontwidth-2*XPOINT,2*YPOINT,ST0,ST2,"%4d,%-4d",cx,cy);
				cpx = cx;
				cpy = cy;
				vd->refresh();
			}
#endif
		}
}


////////////////////////  DRAWING ERASURE


//	draw_undo()
//
//	Repatch the drawing on screen when displaying rubberbands

void	draw_undo(void) {
	coord picx1,picy1,picxs,picys;

	if (rectx1>rectx2 || recty1>recty2) return;

	picx1=rectposx+rectx1-winbm->x1;
	picy1=rectposy+recty1-winbm->y1;
	picxs=rectx2+1-rectx1;
	picys=recty2+1-recty1;

	mouse_hide();
	vd->bitblt(picbm,picx1,picy1,winbm,rectx1,recty1,picxs,picys);
	vd->refresh();
	mouse_show();

	rectx1=recty1=0x7FFF;
	rectx2=recty2=-0x8000;
	rectmodified=FALSE;
}


////////////////////////////////////////////////////////////////////////////////
//
//	void Paint_BlendToScreen(coord x1,coord y1,coord x2,coord y2,int blitmode);
//
//	Blend the picture with the altered screen utilizing the wanted blitmode.

void	Paint_BlendToScreen(coord x1,coord y1,coord x2,coord y2,int blitmode) {
	uchar *vm,*dm;
	uchar vc,dc;
	long vmod,dmod;

	// clip (x1,y1)-(x2,y2) to winbm

	if (x2<winbm->x1 || y2<winbm->y1 || x1>=winbm->x2 || y1>=winbm->y2) return;
	if (x1<winbm->x1) x1=winbm->x1;
	if (y1<winbm->y1) y1=winbm->y1;
	if (x2>=winbm->x2) x2=winbm->x2-1;
	if (y2>=winbm->y2) y2=winbm->y2-1;

	// notify video routines that (x1,y1)-(x2,y2) on screen has been modified

	vid_tap_rect2(&vidbm,x1,y1,x2,y2);

	// calculate address of UL corner, sizes, & modulo

	vm	= vidbm.vm + x1 + y1*vidbm.bpr;
	dm	= picbm->vm + (vposx + x1 - winbm->x1) + (vposy + y1 - winbm->y1)*picbm->bpr;
	x2	= x2+1-x1;		// x size
	y2	= y2+1-y1;		// y size
	x1	= x2;			// use x1 as backup for x2
	vmod	= vidbm.bpr - x2;
	dmod	= picbm->bpr - x2;

	// do the actual merge

	mouse_hide();

	switch(blitmode) {
	case BMODE_ADD:
		do {
			x2 = x1;
			do {
				vc = *vm;
				dc = *dm;
				*vm=GETCOLOR(
					min(255,(coord)piccol[dc*3  ]+(coord)piccol[vc*3  ]),
					min(255,(coord)piccol[dc*3+1]+(coord)piccol[vc*3+1]),
					min(255,(coord)piccol[dc*3+2]+(coord)piccol[vc*3+2])
					);
				++vm,++dm;
			} while(--x2);
			vm += vmod;
			dm += dmod;
		} while(--y2);
		break;
	case BMODE_TRANSP:
		do {
			x2 = x1;
			do {
				vc = *vm;
				dc = *dm;
				if (vc != dc) *vm=GETCOLOR(
					((coord)piccol[dc*3  ]+(coord)piccol[vc*3  ])>>1,
					((coord)piccol[dc*3+1]+(coord)piccol[vc*3+1])>>1,
					((coord)piccol[dc*3+2]+(coord)piccol[vc*3+2])>>1
					);
				++vm,++dm;
			} while(--x2);
			vm += vmod;
			dm += dmod;
		} while(--y2);
		break;
	}

	mouse_show();
}

//////// IDLING

static void	Idle(boolean noidle) {
	static ulong last_tick;
	static ulong idle_tick;
	static ulong idle2_tick;
	ulong	cur_tick;

	if (opt.CPUhog) return;

	cur_tick = read_tick();
	if (abs(cur_tick-last_tick)>18 || noidle) {
		idle_tick = cur_tick + 18;
		idle2_tick = cur_tick + 91;
	} else {
		if (cur_tick >= idle2_tick) {
			int i;

			for(i=0;i<5;i++) {
				DPMI_Yield();
			}
			ExecuteCallList(&idlers);
		} else if (cur_tick >= idle_tick) {
			DPMI_Yield();
			ExecuteCallList(&idlers);
		}
	}
	last_tick = cur_tick;
}


//////////////////////////////////////////// Animation Handling

void	VP_PrevFrame(void) {
	UndoInvalidate();
	PicturePrevFrame(pict_cur);
	LoadFramePalette(pict_cur,pict_cur->current);
	reblit_screen();
	SetTitle(NULL);
}

void	VP_NextFrame(void) {
	UndoInvalidate();
	PictureNextFrame(pict_cur);
	LoadFramePalette(pict_cur,pict_cur->current);
	reblit_screen();
	SetTitle(NULL);
}

void	VP_GotoFrame(void) {
	struct Frame *f;
	int fn;

	fn = RequestSlider("Goto Frame",-1,pict_cur->total_frames-1,pict_cur->current_num);
	if (f=GetFrameAddress(pict_cur,fn)) {
		UndoInvalidate();
		PictureGotoFrame(pict_cur,f,fn);
		LoadFramePalette(pict_cur,f);
		reblit_screen();
		SetTitle(NULL);
	}
}

void	VP_InsertBeforeFrame(void) {
	UndoInvalidate();
	InsertBeforeFrame(pict_cur);
	LoadFramePalette(pict_cur,pict_cur->current);
	reblit_screen();
	SetTitle(NULL);
}

void	VP_InsertAfterFrame(void) {
	UndoInvalidate();
	InsertAfterFrame(pict_cur);
	LoadFramePalette(pict_cur,pict_cur->current);
	reblit_screen();
	SetTitle(NULL);
}

void	VP_DeleteCurrentFrame(void) {
	if (pict_cur->total_frames<=1) return;

	UndoInvalidate();
	DeleteCurrentFrame(pict_cur);
	LoadFramePalette(pict_cur,pict_cur->current);
	reblit_screen();
	SetTitle(NULL);
}

void	VP_RescaleAnimation(int new_xs,int new_ys,boolean scale) {
	struct Picture *pict;
	struct BitMap *bm;
	int i;
	boolean is_first=TRUE;

	// create a spare picture with the new size

	if (SpareInit(new_xs,new_ys)) return;

	// swap it with the current picture

	SpareSwap();

	// start copying & scaling frames over

	vd->set(pict_cur->bm,0);

	PictureGotoFrame(pict_cur,pict_cur->first,1);
	for(i=pict_spare->total_frames;i;i--) {
		if (!is_first)
			if (!InsertAfterFrame(pict_cur)) return;
		is_first = FALSE;
		if (scale)
			vid13_scale(
				pict_spare->bm,
				pict_spare->bm->x1,pict_spare->bm->y1,
				pict_spare->bm->width,pict_spare->bm->height,
				pict_cur->bm,
				pict_cur->bm->x1,pict_cur->bm->y1,
				pict_cur->bm->width,pict_cur->bm->height);
		else
			vid13_bitblt(
				pict_spare->bm,
				pict_spare->bm->x1,pict_spare->bm->y1,
				pict_cur->bm,
				pict_cur->bm->x1,pict_cur->bm->y1,
				pict_cur->bm->width,pict_cur->bm->height);

		if (pict_spare->current->frame_pal) {
			DetachPaletteFromPicture(pict_spare,pict_spare->current->frame_pal);
			AttachPaletteToPicture(pict_cur,pict_spare->current->frame_pal);
			AttachPaletteToFrame(pict_cur,pict_cur->current,pict_spare->current->frame_pal);
			DetachFramePalette(pict_spare,pict_spare->current);
		}

		DeleteCurrentFrame(pict_spare);
		PictureMarkDirty(pict_cur);
	}

	// delete spare picture

	SpareDelete();
}

void	VP_ProcessAnimation(struct Picture *pict) {
	
}


