#define	VGAPAINT_VIDEOS_C

#include <assert.h>
#include <math.h>

#include "vgapaint.h"

static struct StretchRow {
	long pixtotal;
	long pixcntr;
	long pixoffset;
} *str_h=NULL,*str_v=NULL;
static coord str_xs,str_ys,str_x1,str_y1;
static uchar *str_range=NULL;
	       
void vid13_sfill_rec_hline(coord x1,coord x2,coord y) {
	coord t;
	struct StretchRow *sr;

	x1-=str_x1;
	x2-=str_x1;
	y-=str_y1;

	if (x2<x1) { t=x1; x1=x2; x2=t; };
	if (y<0 || y>=str_ys || x2<0 || x1>=str_xs) return;
	if (x1<0) x1=0;
	if (x2>=str_xs) x2=str_xs-1;

	if(str_h) str_h[y].pixtotal+=x2+1-x1;

	if(str_v) {
		sr=&str_v[x1];
		x2-=x1;

		do {
			sr++->pixtotal++;
		} while(x2--);
	};
}

void vid13_sfill_map_init(void) {
	struct StretchRow *sr;
	coord t;

	if (str_h) {
		t=str_ys;
		sr=str_h;
		while(t--) {
			sr->pixcntr=sr->pixtotal;
			sr++;
		};
	}

	if (str_v) {
		t=str_xs;
		sr=str_v;
		while(t--) {
			sr->pixcntr=sr->pixtotal;
			sr++;
		};
	};
}

boolean	vid13_sfill_hinit(BMP dst,coord x1,coord y1,BMP src,long xs,long ys) {

	if (	(x1>=dst->x2) ||
		(y1>=dst->y2) ||
		(xs<=0)	||
		(ys<=0) ||
		(x1<dst->x1-xs) ||
		(y1<dst->y1-ys))	return FALSE;

	str_v=NULL;

	if (!(str_h=malloc(sizeof(struct StretchRow)*ys))) {
		vid13_sfill_dispose();
		return FALSE;
	};

	memset(str_h,0,sizeof(struct StretchRow)*ys);

	str_xs=xs;
	str_ys=ys;
	str_x1=x1;
	str_y1=y1;

	return TRUE;
}

boolean	vid13_sfill_vinit(BMP dst,coord x1,coord y1,BMP src,long xs,long ys) {

	if (	(x1>=dst->x2) ||
		(y1>=dst->y2) ||
		(xs<=0)	||
		(ys<=0) ||
		(x1<dst->x1-xs) ||
		(y1<dst->y1-ys))	return FALSE;

	str_h=NULL;

	if (!(str_v=malloc(sizeof(struct StretchRow)*xs))) {
		vid13_sfill_dispose();
		return FALSE;
	};

	memset(str_v,0,sizeof(struct StretchRow)*xs);

	str_xs=xs;
	str_ys=ys;
	str_x1=x1;
	str_y1=y1;

	return TRUE;
}

boolean	vid13_sfill_init(BMP dst,coord x1,coord y1,BMP src,long xs,long ys) {

	if (	(x1>=dst->x2) ||
		(y1>=dst->y2) ||
		(xs<=0)	||
		(ys<=0) ||
		(x1<dst->x1-xs) ||
		(y1<dst->y1-ys))	return FALSE;

	/* allocate horizontal and vertical counters */

	str_h=NULL;
	str_v=NULL;

	if (!(str_h=malloc(sizeof(struct StretchRow)*ys)) ||
	    !(str_v=malloc(sizeof(struct StretchRow)*xs))) {
		vid13_sfill_dispose();
		return FALSE;
	};

	memset(str_h,0,sizeof(struct StretchRow)*ys);
	memset(str_v,0,sizeof(struct StretchRow)*xs);

	str_xs=xs;
	str_ys=ys;
	str_x1=x1;
	str_y1=y1;

	return TRUE;
}

boolean	vid13_hgradient_init(BMP dst,coord x1,coord y1,BMP src,long xs,long ys) {
	if (!(str_range=CCycle_Unravel(fill_ccrange))) return FALSE;
	return vid13_sfill_hinit(dst,x1,y1,src,xs,ys);
}

boolean	vid13_vgradient_init(BMP dst,coord x1,coord y1,BMP src,long xs,long ys) {
	if (!(str_range=CCycle_Unravel(fill_ccrange))) return FALSE;
	return vid13_sfill_vinit(dst,x1,y1,src,xs,ys);
}

void vid13_sfill_dispose(void) {
	if (str_h) free(str_h);
	if (str_v) free(str_v);
	if (str_range) free(str_range);
	str_h=str_v=0;
	str_range=0;
}

/* mapping MUST be done from top to buttom */

void vid13_stretch_map_hline(BMP dst,BMP src,coord x1,coord x2,coord y) {
	coord t;
	struct StretchRow *row;
	struct StretchRow *column;
	uchar *vm;
	long pre=0;

	if (x2>=dst->x2) x2=dst->x2-1;
	if (x1<dst->x1) pre=dst->x1-x1;
	if (y<dst->y1) pre=0x7FFFFFFF;
	if (y>=dst->y2) return;

	x1-=str_x1;
	x2-=str_x1;
	y-=str_y1;

	if (x2<x1) { t=x2; x2=x1; x1=t; }
	if (x1>=str_xs || x2<0 || y<0 || y>=str_ys) return;

	if (x1<0) x1=0;
	if (x2>=str_xs) x2=str_xs-1;

	vid_tap_rect2(dst,str_x1+x1,str_y1+y,str_x1+x2,str_y1+y);

	vm=(uchar *)(dst->vm)+(x1+str_x1)+(dst->bpr)*(y+str_y1);
	row=str_h+y;
	column=str_v+x1;
	x2-=x1;

	do {
		if (pre) --pre;
		else *vm=*(src->vm+row->pixoffset+column->pixoffset);
		++vm;
		row->pixcntr-=brush_sprite.width;

		while(row->pixcntr<0) {
			row->pixoffset++;
			row->pixcntr+=row->pixtotal;
		};
		column->pixcntr-=brush_sprite.height;

		while(column->pixcntr<0) {
			column->pixoffset+=brush_sprite.imgbm.bpr;
			column->pixcntr+=column->pixtotal;
		};
		column++;
	} while(x2--);
}

void vid13_stretch_blend_map_hline(BMP dst,BMP src,coord x1,coord x2,coord y) {
	coord t;
	struct StretchRow *row;
	struct StretchRow *column;
	uchar *vm;
	uchar *srcp;
	long pre=0;
	long ptotal,rpt,cpt;
	long rpc1,rpc2,cpc1,cpc2;
	uchar *pp1,*pp2,*pp3,*pp4;

	if (x2>=dst->x2) x2=dst->x2-1;
	if (x1<dst->x1) pre=dst->x1-x1;
	if (y<dst->y1) pre=0x7FFFFFFF;
	if (y>=dst->y2) return;

	x1-=str_x1;
	x2-=str_x1;
	y-=str_y1;

	if (x2<x1) { t=x2; x2=x1; x1=t; }
	if (x1>=str_xs || x2<0 || y<0 || y>=str_ys) return;

	if (x1<0) x1=0;
	if (x2>=str_xs) x2=str_xs-1;

	vid_tap_rect2(dst,str_x1+x1,str_y1+y,str_x1+x2,str_y1+y);

	vm=(uchar *)(dst->vm)+(x1+str_x1)+(dst->bpr)*(y+str_y1);
	row=str_h+y;
	column=str_v+x1;
	x2-=x1;

	rpt=row->pixtotal+1;

	do {
		if (pre) --pre;
		else {

			// row counter is from (pixtotal+1) to 0
			// column counter is from (pixtotal+1) to 0

			srcp=src->vm+row->pixoffset+column->pixoffset;

			pp1=piccol + 3*( srcp[0] );
			pp2=piccol + 3*( srcp[1] );
			pp3=piccol + 3*( srcp[brush_sprite.imgbm.bpr] );
			pp4=piccol + 3*( srcp[brush_sprite.imgbm.bpr+1] );

			cpt=column->pixtotal+1;
			ptotal=rpt*cpt;

			rpc1=row->pixcntr;	rpc2=rpt-rpc1;
			cpc1=column->pixcntr;	cpc2=cpt-cpc1;

			*vm=FPick_Choose(fpcol,256,
	    	 		(pp1[0]*rpc1*cpc1 + pp2[0]*rpc2*cpc1 + pp3[0]*rpc1*cpc2 + pp4[0]*rpc2*cpc2)/ptotal,
	    	 		(pp1[1]*rpc1*cpc1 + pp2[1]*rpc2*cpc1 + pp3[1]*rpc1*cpc2 + pp4[1]*rpc2*cpc2)/ptotal,
	    	 		(pp1[2]*rpc1*cpc1 + pp2[2]*rpc2*cpc1 + pp3[2]*rpc1*cpc2 + pp4[2]*rpc2*cpc2)/ptotal);
		};
		++vm;
		row->pixcntr-=brush_sprite.width;

		while(row->pixcntr<0) {
			row->pixoffset++;
			row->pixcntr+=row->pixtotal;
		};
		column->pixcntr-=brush_sprite.height;

		while(column->pixcntr<0) {
			column->pixoffset+=brush_sprite.imgbm.bpr;
			column->pixcntr+=column->pixtotal;
		};
		column++;
	} while(x2--);
}

void	vid13_hgradient_map_hline(BMP dst,BMP src,coord x1,coord x2,coord y) {
	coord t;
	struct StretchRow *row;
	uchar *vm;
	long pre=0;

	if (x2>=dst->x2) x2=dst->x2-1;
	if (x1<dst->x1) pre=dst->x1-x1;
	if (y<dst->y1) pre=0x7FFFFFFF;
	if (y>=dst->y2) return;

	x1-=str_x1;
	x2-=str_x1;
	y-=str_y1;

	if (x2<x1) { t=x2; x2=x1; x1=t; }
	if (x1>=str_xs || x2<0 || y<0 || y>=str_ys) return;

	if (x1<0) x1=0;
	if (x2>=str_xs) x2=str_xs-1;

	vid_tap_rect2(dst,str_x1+x1,str_y1+y,str_x1+x2,str_y1+y);

	vm=(uchar *)(dst->vm)+(x1+str_x1)+(dst->bpr)*(y+str_y1);
	row=str_h+y;
	x2-=x1;
	do {
		if (pre) --pre;
		else *vm=str_range[row->pixoffset+2];
		++vm;
		row->pixcntr-=*(short*)str_range;

		while(row->pixcntr<0) {
			row->pixoffset++;
			row->pixcntr+=row->pixtotal;
		};
	} while(x2--);
}

void	vid13_vgradient_map_hline(BMP dst,BMP src,coord x1,coord x2,coord y) {
	coord t;
	struct StretchRow *column;
	uchar *vm;
	long pre=0;

	if (x2>=dst->x2) x2=dst->x2-1;
	if (x1<dst->x1) pre=dst->x1-x1;
	if (y<dst->y1) pre=0x7FFFFFFF;
	if (y>=dst->y2) return;

	x1-=str_x1;
	x2-=str_x1;
	y-=str_y1;

	if (x2<x1) { t=x2; x2=x1; x1=t; }
	if (x1>=str_xs || x2<0 || y<0 || y>=str_ys) return;

	if (x1<0) x1=0;
	if (x2>=str_xs) x2=str_xs-1;

	vid_tap_rect2(dst,str_x1+x1,str_y1+y,str_x1+x2,str_y1+y);

	vm=(uchar *)(dst->vm)+(x1+str_x1)+(dst->bpr)*(y+str_y1);
	column=str_v+x1;
	x2-=x1;
	do {
		if (pre) --pre;
		else *vm=str_range[column->pixoffset+2];
		++vm;
		column->pixcntr-=*(short*)str_range;

		while(column->pixcntr<0) {
			column->pixoffset++;
			column->pixcntr+=column->pixtotal;
		};
		column++;
	} while(x2--);
}

/////////////////////  ANTIALIASING //////////////////////

#define shortswap(x,y) { t=x; x=y; y=t; }

void vid_antialias_line(struct BitMap *dst,coord x1,coord y1,coord x2,coord y2,color f,color b,boolean blend) {
	boolean steep;
	uchar r1,g1,b1,r2,g2,b2;
	long ypos,yinc,tac1,tac2;
	coord dx,dy;
	coord t;

	uchar *cp;

	f=vid_ctran(f)*3;
	b=vid_ctran(b)*3;

	r1=piccol[f];	g1=piccol[f+1];	b1=piccol[f+2];
	r2=piccol[b];	g2=piccol[b+1];	b2=piccol[b+2];

	if (vid_line_clip(dst,&x1,&y1,&x2,&y2)) return;

	dx=x2-x1; dy=y2-y1;

	if (steep=(abs(dx)<abs(dy))) {
		t=dx;
		dx=dy;
		dy=t;
	};

	if (!dx) {
		vd->plot(winbm,x1,y1,f);
		return;
	};

	if (steep) {
		if (y1>y2) {
			swap(coord,x1,x2);
			shortswap(y1,y2);
		};

		yinc=((long)(x2-x1)<<16)/abs(dx);
		ypos=(long)x1<<16;

		do {
			tac2=(ushort)ypos;
			tac1=65536-tac2;
			if (blend) {
				cp=piccol+vd->point(dst,ypos>>16,y1)*3;
				r2=*cp++;	g2=*cp++;	b2=*cp;
			};
			vd->plot(dst,ypos>>16,y1,FPick_Choose(fpcol,256,(r1*tac1+r2*tac2)>>16,(g1*tac1+g2*tac2)>>16,(b1*tac1+b2*tac2)>>16));
			if (blend) {
				cp=piccol+vd->point(dst,(ypos>>16)+1,y1)*3;
				r2=*cp++;	g2=*cp++;	b2=*cp;
			};
			vd->plot(dst,(ypos>>16)+1,y1,FPick_Choose(fpcol,256,(r2*tac1+r1*tac2)>>16,(g2*tac1+g1*tac2)>>16,(b2*tac1+b1*tac2)>>16));
			ypos+=yinc;
		} while(++y1<=y2);
	} else {
		if (x1>x2) {
			shortswap(x1,x2);
			shortswap(y1,y2);
		};

		yinc=((long)(y2-y1)<<16)/abs(dx);
		ypos=((long)y1<<16);

		do {
			tac2=(ushort)ypos;
			tac1=65536-tac2;
			if (blend) {
				cp=piccol+vd->point(dst,x1,ypos>>16)*3;
				r2=*cp++;	g2=*cp++;	b2=*cp;
			};
			vd->plot(dst,x1,ypos>>16,FPick_Choose(fpcol,256,(r1*tac1+r2*tac2)>>16,(g1*tac1+g2*tac2)>>16,(b1*tac1+b2*tac2)>>16));
			if (blend) {
				cp=piccol+vd->point(dst,x1,(ypos>>16)+1)*3;
				r2=*cp++;	g2=*cp++;	b2=*cp;
			};
			vd->plot(dst,x1,(ypos>>16)+1,FPick_Choose(fpcol,256,(r2*tac1+r1*tac2)>>16,(g2*tac1+g1*tac2)>>16,(b2*tac1+b1*tac2)>>16));
//			vd->plot(dst,x1,ypos>>16,f);
//			vd->plot(dst,x1,(ypos+65536)>>16,f);
			ypos+=yinc;
		} while(++x1<=x2);
	};
}


void vid_antialias_plot(struct BitMap *dst,coord x,coord y,ulong fgcol,ulong bgcol,ulong cv,boolean blend) {
	if (blend) {
		uchar	c;

		c = vd->point(dst,x,y);

		bgcol = *(ushort *)(dst->pal + 3*c) + ((ulong)dst->pal[3*c+2]<<16);
	}
	vd->plot(dst,x,y,FPick_Choose(fpcol,256
		,(((uchar)(bgcol    ))*cv+((uchar)(fgcol    ))*(0x10000L-cv))>>16
		,(((uchar)(bgcol>> 8))*cv+((uchar)(fgcol>>8 ))*(0x10000L-cv))>>16
		,(((uchar)(bgcol>>16))*cv+((uchar)(fgcol>>16))*(0x10000L-cv))>>16
		)
	);
}


void vid_antialias_circle(struct BitMap *dst,coord x,coord y,coord r,color f,color b,boolean blend) {
	coord ix,iy,ie;
	ulong	r2,rx,ry,ra;
	ulong	col_lo,col_hi;

	// get color values

	f = vid_ctran(f);
	b = vid_ctran(b);

	// avoid page fault on last color

	col_lo = *(ushort *)(piccol + b*3) + ((ulong)piccol[b*3+2] << 16);
	col_hi = *(ushort *)(piccol + f*3) + ((ulong)piccol[f*3+2] << 16);

	// prepare for modified bresenham-style circle algorithm

	ix = 0;
	iy = r;
	r2 = r*r;	// squared radius of circle
	ra = r2;	// squared distance from center to current point
	rx = 1;		// delta for ra when x is incremented
	ry = 2*r-1;	// delta for ra when y is decremented

	// plot the four quadrant points at full intensity, since they are prefectly on
	// the circle

	vd->plot(dst,x+r,y  ,f);
	vd->plot(dst,x-r,y  ,f);
	vd->plot(dst,x  ,y+r,f);
	vd->plot(dst,x  ,y-r,f);

	// draw the remainder of the circle

	while(ix<=iy) {
		if (ix) {	// skip this the first time around
			double	sqinc = iy*2;
			double	sqp = ra - sqinc;
			double	d;
			int	ay;
			ulong	coverage;

			for(ay=-1;ay<=1;ay++) {
				d = fabs(sqrt(sqp + abs(ay))-r);
				sqp += sqinc;
				if (d<1.0) {
					coverage = d*65536.0;
					vid_antialias_plot(dst,x+ix,y+iy+ay,col_hi,col_lo,coverage,blend);
					vid_antialias_plot(dst,x-ix,y+iy+ay,col_hi,col_lo,coverage,blend);
					vid_antialias_plot(dst,x+ix,y-iy-ay,col_hi,col_lo,coverage,blend);
					vid_antialias_plot(dst,x-ix,y-iy-ay,col_hi,col_lo,coverage,blend);
					if (iy+ay != ix) {
						vid_antialias_plot(dst,x+iy+ay,y+ix,col_hi,col_lo,coverage,blend);
						vid_antialias_plot(dst,x+iy+ay,y-ix,col_hi,col_lo,coverage,blend);
						vid_antialias_plot(dst,x-iy-ay,y+ix,col_hi,col_lo,coverage,blend);
						vid_antialias_plot(dst,x-iy-ay,y-ix,col_hi,col_lo,coverage,blend);
					}
				}
			}
		}
		if (ra>r2) {
			ra -= ry;
			ry -= 2;
			--iy;
		}
		ra += rx;
		rx += 2;
		++ix;
	}
}

void	vid_antialias_ellipse(struct BitMap *dst,coord x,coord y,coord a,coord b,color fc,color bc,boolean blend,boolean fill) {
	ulong	col_lo,col_hi;
	ushort	*stripmem;
	double	y2=0.0,y2inc=1.0;
	double	a1,a2,bsq1,bsq2;
	long	x1,x2;
	int	y0,y_strip,i;

	fc = vid_ctran(fc);
	bc = vid_ctran(bc);

	// check for really stupid cases, such as points and lines

	if (!a) {
		if (!b)
			vd->plot(dst,x,y,fc);
		else	vd->vline(dst,x,y-b,y+b,fc);
		return;
	} else if (!b) {
		vd->hline(dst,x-a,x+a,y,fc);
		return;
	}

	col_lo = *(ushort *)(piccol + bc*3) + ((ulong)piccol[bc*3+2] << 16);
	col_hi = *(ushort *)(piccol + fc*3) + ((ulong)piccol[fc*3+2] << 16);

	// allocate strip memory

	if (!(stripmem = malloc(sizeof(ushort)*(a+1)))) return;

	// trace through ellipse, 16 rows at a time

	bsq1	= (b*16+15)*(b*16+15);
	bsq2	= b*b*256;
	a1	= a*16+15;
	a2	= a*16;

	if (fill) {
		for(y0=0;y0<=b;y0++) {
			memset(stripmem,0,sizeof(ushort)*(a+1));
			for(y_strip=0;y_strip<16;y_strip++) {
				x1 = a1*sqrt(1.0-y2/bsq1);

				// add strip from 0 to x1

				for(i=0;i<((x1+1)/16);i++) stripmem[i]+=16;
				if (15!=(x1 & 15)) stripmem[x1/16] += (x1&15)+1;

				y2+=y2inc; y2inc+=2.0;
			}

			// render strip to screen

			i=0;
			while(stripmem[i]==0x100 && i<=a)
				++i;

			vd->hline(dst,x+1-i,x+i-1,y-y0,fc);
			vd->hline(dst,x+1-i,x+i-1,y+y0,fc);

			for(;i<=a;i++) {
				if (stripmem[i]) {
					if (i) {
						vid_antialias_plot(dst,x-i,y-y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
						vid_antialias_plot(dst,x-i,y+y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
					}
					vid_antialias_plot(dst,x+i,y-y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
					vid_antialias_plot(dst,x+i,y+y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
				}
			}
		}
	} else {
		for(y0=0;y0<=b;y0++) {
			memset(stripmem,0,sizeof(ushort)*(a+1));
			for(y_strip=0;y_strip<16;y_strip++) {
				x1 = a1*sqrt(1.0-y2/bsq1);
				if (y2>bsq2) {
					// add strip from 0 to x1

					for(i=0;i<((x1+1)/16);i++) stripmem[i]+=16;
					if (15!=(x1 & 15)) stripmem[x1/16] += (x1&15)+1;
				} else {
					x2 = a2*sqrt(1.0-y2/bsq2);

					// add strip from x2 to x1
					// case I: x1...x2 in a single block
					// case II: x1...x2 in two or more blocks

					if ((x1/16)==(x2/16))	// case I
						stripmem[x1/16] += x1-x2;
					else {			// case II
						stripmem[x2/16] += 16-(x1&15);
						stripmem[x1/16] += (x2&15)+1;
						for(i = x2/16+1 ; i < x1/16 ; i++)
							stripmem[i]+=16;
					}
				}

				y2+=y2inc; y2inc+=2.0;
			}

			// render strip to screen

			for(i=0;i<=a;i++) {
				if (stripmem[i]) {
					vid_antialias_plot(dst,x-i,y-y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
					vid_antialias_plot(dst,x+i,y-y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
					vid_antialias_plot(dst,x-i,y+y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
					vid_antialias_plot(dst,x+i,y+y0,col_hi,col_lo,0x10000L-(stripmem[i]<<8),blend);
				}
			}
		}
	}

	free(stripmem);
}


///////////////////////////////////// FASTPICK ROUTINES ////////////////////////////////////

#pragma off (check_stack)

#define	FP_QBArrayBIAS	(765)

uchar	FP_QBArray[765*2+1]={
              0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  2,  3,  4,  4,  5,  5,  6,  6,  7,  7,  8,  8,  8,  9,  9, 10, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 15,
 15, 15, 16, 16, 16, 17, 17, 17, 17, 18, 18, 18, 18, 19, 19, 19, 20, 20, 20, 20, 21, 21, 21, 21, 22, 22, 22, 22, 23, 23, 23, 23,
 24, 24, 24, 24, 25, 25, 25, 25, 25, 26, 26, 26, 26, 27, 27, 27, 27, 28, 28, 28, 28, 28, 29, 29, 29, 29, 29, 30, 30, 30, 30, 31,
 31, 31, 31, 31, 32, 32, 32, 32, 32, 33, 33, 33, 33, 33, 34, 34, 34, 34, 34, 35, 35, 35, 35, 35, 36, 36, 36, 36, 36, 37, 37, 37,
 37, 37, 38, 38, 38, 38, 38, 38, 39, 39, 39, 39, 39, 40, 40, 40, 40, 40, 40, 41, 41, 41, 41, 41, 42, 42, 42, 42, 42, 42, 43, 43,
 43, 43, 43, 44, 44, 44, 44, 44, 44, 45, 45, 45, 45, 45, 45, 46, 46, 46, 46, 46, 46, 47, 47, 47, 47, 47, 47, 48, 48, 48, 48, 48,
 49, 49, 49, 49, 49, 49, 49, 50, 50, 50, 50, 50, 50, 51, 51, 51, 51, 51, 51, 52, 52, 52, 52, 52, 52, 53, 53, 53, 53, 53, 53, 54,
 54, 54, 54, 54, 54, 54, 55, 55, 55, 55, 55, 55, 56, 56, 56, 56, 56, 56, 57, 57, 57, 57, 57, 57, 57, 58, 58, 58, 58, 58, 58, 58,
 59, 59, 59, 59, 59, 59, 60, 60, 60, 60, 60, 60, 60, 61, 61, 61, 61, 61, 61, 61, 62, 62, 62, 62, 62, 62, 63, 63, 63, 63, 63, 63,
 63, 64, 64, 64, 64, 64, 64, 64, 65, 65, 65, 65, 65, 65, 65, 66, 66, 66, 66, 66, 66, 66, 67, 67, 67, 67, 67, 67, 67, 68, 68, 68,
 68, 68, 68, 68, 69, 69, 69, 69, 69, 69, 69, 70, 70, 70, 70, 70, 70, 70, 70, 71, 71, 71, 71, 71, 71, 71, 72, 72, 72, 72, 72, 72,
 72, 73, 73, 73, 73, 73, 73, 73, 73, 74, 74, 74, 74, 74, 74, 74, 75, 75, 75, 75, 75, 75, 75, 76, 76, 76, 76, 76, 76, 76, 76, 77,
 77, 77, 77, 77, 77, 77, 77, 78, 78, 78, 78, 78, 78, 78, 79, 79, 79, 79, 79, 79, 79, 79, 80, 80, 80, 80, 80, 80, 80, 80, 81, 81,
 81, 81, 81, 81, 81, 82, 82, 82, 82, 82, 82, 82, 82, 83, 83, 83, 83, 83, 83, 83, 83, 84, 84, 84, 84, 84, 84, 84, 84, 85, 85, 85,
 85, 85, 85, 85, 85, 86, 86, 86, 86, 86, 86, 86, 86, 87, 87, 87, 87, 87, 87, 87, 87, 88, 88, 88, 88, 88, 88, 88, 88, 89, 89, 89,
 89, 89, 89, 89, 89, 90, 90, 90, 90, 90, 90, 90, 90, 91, 91, 91, 91, 91, 91, 91, 91, 91, 92, 92, 92, 92, 92, 92, 92, 92, 93, 93,
 93, 93, 93, 93, 93, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 95, 95, 95, 95, 95, 95, 95, 95, 96, 96, 96, 96, 96, 96, 96, 96, 96,
 97, 97, 97, 97, 97, 97, 97, 97, 98, 98, 98, 98, 98, 98, 98, 98, 98, 99, 99, 99, 99, 99, 99, 99, 99,100,100,100,100,100,100,100,
100,100,101,101,101,101,101,101,101,101,101,102,102,102,102,102,102,102,102,103,103,103,103,103,103,103,103,103,104,104,104,104,
104,104,104,104,104,105,105,105,105,105,105,105,105,105,106,106,106,106,106,106,106,106,106,107,107,107,107,107,107,107,107,107,
108,108,108,108,108,108,108,108,109,109,109,109,109,109,109,109,109,109,110,110,110,110,110,110,110,110,110,111,111,111,111,111,
111,111,111,111,112,112,112,112,112,112,112,112,112,113,113,113,113,113,113,113,113,113,114,114,114,114,114,114,114,114,114,115,
115,115,115,115,115,115,115,115,115,116,116,116,116,116,116,116,116,116,117,117,117,117,117,117,117,117,117,118,118,118,118,118,
118,118,118,118,118,119,119,119,119,119,119,119,119,119,120,120,120,120,120,120,120,120,120,121,121,121,121,121,121,121,
};

//ulong	FP_SQArray[]={
//     0,     1,     4,     9,    16,    25,    36,    49,    64,    81,   100,   121,   144,   169,   196,   225,	//16
//   256,   289,   324,   361,   400,   441,   484,   529,   576,   625,   676,   729,   784,   841,   900,   961,	//32
//  1024,  1089,  1156,  1225,  1296,  1369,  1444,  1521,  1600,  1681,  1764,  1849,  1936,  2025,  2116,  2209,
//  2304,  2401,  2500,  2601,  2704,  2809,  2916,  3025,  3136,  3249,  3364,  3481,  3600,  3721,  3844,  3969,	//64
//  4096,  4225,  4356,  4489,  4624,  4761,  4900,  5041,  5184,  5329,  5476,  5625,  5776,  5929,  6084,  6241,
//  6400,  6561,  6724,  6889,  7056,  7225,  7396,  7569,  7744,  7921,  8100,  8281,  8464,  8649,  8836,  9025,
//  9216,  9409,  9604,  9801, 10000, 10201, 10404, 10609, 10816, 11025, 11236, 11449, 11664, 11881, 12100, 12321,
// 12544, 12769, 12996, 13225, 13456, 13689, 13924, 14161, 14400, 14641, 14884, 15129, 15376, 15625, 15876, 16129,	//128
// 16384, 16641, 16900, 17161, 17424, 17689, 17956, 18225, 18496, 18769, 19044, 19321, 19600, 19881, 20164, 20449,
// 20736, 21025, 21316, 21609, 21904, 22201, 22500, 22801, 23104, 23409, 23716, 24025, 24336, 24649, 24964, 25281,
// 25600, 25921, 26244, 26569, 26896, 27225, 27556, 27889, 28224, 28561, 28900, 29241, 29584, 29929, 30276, 30625,
// 30976, 31329, 31684, 32041, 32400, 32761, 33124, 33489, 33856, 34225, 34596, 34969, 35344, 35721, 36100, 36481,	//192
// 36864, 37249, 37636, 38025, 38416, 38809, 39204, 39601, 40000, 40401, 40804, 41209, 41616, 42025, 42436, 42849,
// 43264, 43681, 44100, 44521, 44944, 45369, 45796, 46225, 46656, 47089, 47524, 47961, 48400, 48841, 49284, 49729,
// 50176, 50625, 51076, 51529, 51984, 52441, 52900, 53361, 53824, 54289, 54756, 55225, 55696, 56169, 56644, 57121,
// 57600, 58081, 58564, 59049, 59536, 60025, 60516, 61009, 61504, 62001, 62500, 63001, 63504, 64009, 64516, 65025,	//256
//};

ulong	FP_SQArray[511];
#define	FP_SQArrayBIAS	(255L)

boolean	FP_SQArrayInitialized=FALSE;

int	FPick_QCompare(const struct FastPickEntry *a1,const struct FastPickEntry *a2) {
	return (int)a1->lum-(int)a2->lum;
}

struct FastPick *FPick_Allocate(ushort colors) {
	struct FastPick *fp;

	if (fp=(struct FastPick *)malloc(sizeof(struct FastPick))) {
		if (fp->fpent = (struct FastPickEntry *)malloc(sizeof(struct FastPickEntry)*colors)) {
			if (fp->fplum = (uchar *)malloc(768)) {
				return fp;
			}
			free(fp->fpent);
		}
		free(fp);
	}
	return NULL;
}

void	FPick_Free(struct FastPick *fp) {
	free(fp->fpent);
	free(fp->fplum);
	free(fp);
}

void	FPick_Initialize(struct FastPick *fba,ushort start,ushort length,uchar *dstpal) {
	uchar *pc=dstpal+start*3;
	struct FastPickEntry *fbp=fba->fpent;
	ushort length2=length;

	if (!FP_SQArrayInitialized) {
		int i;

		for(i=0;i<256;i++)	FP_SQArray[FP_SQArrayBIAS+i]=FP_SQArray[FP_SQArrayBIAS-i]=i*i;
		for(i=0;i<766;i++)	FP_QBArray[FP_QBArrayBIAS-i]=FP_QBArray[FP_QBArrayBIAS+i];

		FP_SQArrayInitialized=TRUE;
	}

	do {
		fbp->r=*pc++;
		fbp->g=*pc++;
		fbp->b=*pc++;
		fbp->lum=(ushort)pc[-3]+(ushort)pc[-2]+(ushort)pc[-1];
		fbp->index=start++;
		++fbp;
	} while(--length2);

	qsort(fba->fpent,length,sizeof(struct FastPickEntry),FPick_QCompare);

	{
		uchar *lumtbl=fba->fplum;
		int i,j,k,l;

		fbp=fba->fpent;

		for(i=0;i<768;i++) {

			// binary search for luminance <i>

			j=0;
			k=length-1;
			while( (l=(j+k)>>1) , (j < k) ) {
				if (fbp[l].lum==i) break;
				else if (fbp[l].lum<i) j=l+1;
				else if (fbp[l].lum>i) k=l-1;
			}

			// obtain span of the luminance we picked

			j=k=l;
			while(j>0)
				if (fbp[j-1].lum==fbp[j].lum) --j; else break;

			while(k<length-1)
				if (fbp[k+1].lum==fbp[k].lum) ++k; else break;

			// write average

			*lumtbl++ = (j+k)>>1;
		}
	}
}

uchar	FPick_Choose(struct FastPick *fbp,ushort count,uchar r,uchar g,uchar b) {
	struct FastPickEntry *fp,*fp2;
	int lum;
	int mc,x;
	ulong dist=0xFFFFFFFF;
	int cindex;
	ulong sqdist;
	uchar *QBArrayPtr;

	lum=(int)r+(int)g+(int)b;

	QBArrayPtr = FP_QBArray + FP_QBArrayBIAS + lum;

//	choose an index by interpolating brightness into array

//	x=mc=((lum-fbp->lum)*(count-1))/(fbp[count-1].lum-fbp->lum);
	x=mc=fbp->fplum[lum];

	fp=fbp->fpent + x;
	fp2=fbp->fpent + mc;

	for(;;) {
//		if (FP_QBArray[FP_QBArrayBIAS+lum-fp->lum]>dist) goto up_only;
		if (QBArrayPtr[-fp->lum]>dist) goto up_only;

		sqdist =FP_SQArray[FP_SQArrayBIAS + (int)fp->r - (int)r];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp->g - (int)g];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp->b - (int)b];

		if (sqdist<dist) {
			cindex=x;
			if (!(dist=sqdist)) return fbp->fpent[cindex].index;
		};
		--fp;
		++fp2;

		if (++mc >= count) goto down_only;

//		if (FP_QBArray[FP_QBArrayBIAS+lum-fp2->lum]>dist) goto down_only;
		if (QBArrayPtr[-fp2->lum]>dist) goto down_only;

		sqdist =FP_SQArray[FP_SQArrayBIAS + (int)fp2->r - (int)r];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp2->g - (int)g];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp2->b - (int)b];

		if (sqdist<dist) {
			cindex=mc;
			if (!(dist=sqdist)) return fbp->fpent[cindex].index;
		};

		if (--x < 0) goto up_only;
	};

down_only:
	do {
//		if (FP_QBArray[FP_QBArrayBIAS+lum-fp->lum]>dist) break;
		if (QBArrayPtr[-fp->lum]>dist) break;

		sqdist =FP_SQArray[FP_SQArrayBIAS + (int)fp->r - (int)r];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp->g - (int)g];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp->b - (int)b];

		if (sqdist<dist) {
			cindex=x;
			if (!(dist=sqdist)) break; //return fbp->fpent[cindex].index;
		};
		--fp;
	} while(x-->0);
	return fbp->fpent[cindex].index;

up_only:
	while(++mc<count) {
		++fp2;
//		if (FP_QBArray[FP_QBArrayBIAS+lum-fp2->lum]>dist) break;
		if (QBArrayPtr[-fp2->lum]>dist) break;

		sqdist =FP_SQArray[FP_SQArrayBIAS + (int)fp2->r - (int)r];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp2->g - (int)g];
		sqdist+=FP_SQArray[FP_SQArrayBIAS + (int)fp2->b - (int)b];

		if (sqdist<dist) {
			cindex=mc;
			if (!(dist=sqdist)) break; //return fbp->fpent[cindex].index;
		};
	};
	return fbp->fpent[cindex].index;
}


	static uchar greypat[4]={ 0,0,0,0 };
	static struct BitMap greybm={NULL};

void	vid13_grey(BMP dst,coord x1,coord y1,coord xs,coord ys,color pri,color sec) {
	if (!greybm.vm)	vid_initBitMap(&greybm,greypat,0,0,2,2,2,dst->pal);
	greypat[0]=greypat[3]=vid_ctran(pri);
	greypat[1]=greypat[2]=vid_ctran(sec);
	vid13_tile(&greybm,dst,x1,y1,xs,ys,x1,y1);
}

