#include <a/types.h>
#include <a/mem.h>
#include "loadlib.h"
#include "toolicon.h"
#include "video.h"
#include "extdata.h"

#pragma off (check_stack)

boolean	ovid_init(enum VD_InitCode vdic,long x,long y);
void	ovid_rbitblt(coord x1,coord y1,coord xs,coord ys,uchar *mem,long dmod,boolean no_offset);
void	ovid_rhline(coord x,coord y,uchar *mem,coord xs);
void	ovid_rhlineread(coord x,coord y,uchar *mem,coord xs);
void	ovid_drawptr(struct MousePointerInfo *mpi,coord x,coord y);
void	ovid_eraseptr(struct MousePointerInfo *mpi);

struct VDevSpecs ovds={
	640,400,640,400,640,400,640,400
};

struct VDev ovd={
	NULL,				// set
	NULL,				// plot
	NULL,				// point
	NULL,				// bar
	NULL,				// bitblt
	NULL,				// bitbltcolor
	NULL,				// bitblttrans
	NULL,				// bitbltswap
	NULL,				// bitbltswaptrans
	NULL,				// bitbltswapmask,
	NULL,				// bltpattern
	NULL,				// bltpattern3d
	NULL,				// text6
	NULL,				// text6trans
	NULL,				// text6lim,
	NULL,				// hline
	NULL,				// vline
	NULL,				// line
	NULL,				// box
	NULL,				// circle
	NULL,				// ellipse
	NULL,				// blocksave
	NULL,				// blockrestore
	NULL,				// hblit
	NULL,				// vblit
	NULL,				// tile
	NULL,				// rassize
	NULL,				// bpr
	&ovid_init,
	NULL,				// refresh
	&ovid_rhline,
	NULL,				// quickfill
	&ovid_rbitblt,
	&ovid_drawptr,
	&ovid_eraseptr,
	&ovid_rhlineread
};

struct VideoDriver ourVideoDriver={
	{NULL},"640x400 ModeX paged",&ovd,NULL,VDF_FONT8x8,256,VDEV_SIZE,VDRV_SIZE,2,2,0x11c71,0x0e666
};

static long scroll_x=0,scroll_y=0;	// current y position

long main(struct Extension *ext,enum ExtCommand comm,long xval,long yval) {
	switch(comm) {
	case	EC_STARTUP:
		return (long)&main;

	case	EC_STARTUP1:
		ext->name = "640x400 Mode X paged video driver";

		// add our video driver to the list

		return !AddVideoDriver(&ourVideoDriver);

	case	EC_CLEANUP:

		// remove our video driver from the list

		RemVideoDriver(&ourVideoDriver);

		return 0;

	default:
		return -1;	// whatever it was, we didn't do it
	};
}

boolean	ovid_init(enum VD_InitCode vdic,long x,long y) {
	void *realvm;

	switch(vdic) {
	case VD_STARTUP:
		_vidbm->vm=NULL;
		if (!(realvm=VPmalloc(640*480))) return FALSE;
		vid_initBitMap(_vidbm,(uchar *)realvm,0,0,640,400,640,_vidbm->pal);
	case VD_REINIT:
	case VD_FROM13:
		vid_setmode(0x13);
		disable();
		outportb(0x3d4,0x14);
		outportb(0x3d5,0);
		outportb(0x3d4,0x17);
		outportb(0x3d5,0xe3);
		outportb(0x3d4,0x13);	// offset
		outportb(0x3d5,80);
		outportb(0x3c4,4);
		outportb(0x3c5,6);
		enable();
		vid_initBitMap(_realvidbm,(uchar *)0xA0000,0,0,640,400,160,_vidbm->pal);	/* the bpr is funny but necessary */
		_vidbm->syncrect=TRUE;
		_vidbm->syncx1=_vidbm->syncy1=0x7FFF;
		_vidbm->syncx2=_vidbm->syncy2=-0x8000;
		clipReset();
		mouse_bounds_reset();
		break;
	case VD_CLOSEDOWN:
	case VD_TO13:
		disable();
		outportb(0x3d4,0x14);
		outportb(0x3d5,0x40);
		outportb(0x3d4,0x17);
		outportb(0x3d5,0xa3);
		outportb(0x3d4,0x13);	// offset
		outportb(0x3d5,40);
		outportb(0x3c4,4);
		outportb(0x3c5,14);
		enable();
		if (vdic==VD_CLOSEDOWN) {
//			if (_vidbm->vm) VPfree(_vidbm->vm);
			_vidbm->vm=NULL;
		};
		break;
	case VD_RESIZE:
		break;
	};
	return TRUE;
}

void	ovid_rhline(coord x,coord y,uchar *dm1,coord xs) {
	uchar *vm1,*vm2;
	uchar *dm2;
	uchar p;
	coord xst;

	vm1=(uchar *)0xA0000+(x>>2)+y*_realvidbm->bpr;

	for(p=0;p<4;p++) {
		outportb(0x3c4,2);
		outportb(0x3c5,1<<p);
		vm2=vm1;
		dm2=dm1++;
		xst=xs;
		if (p<(x&3)) {
			--xst;
			++vm2;
		};
		if (xst)
			do {
				*vm2++=*dm2++;
				dm2+=3;
			} while(--xst);
	};
}

void	ovid_rhlineread(coord x,coord y,uchar *dm1,coord xs) {
	uchar *vm1,*vm2;
	uchar *dm2;
	uchar p;
	coord xst;

	vm1=(uchar *)0xA0000+(x>>2)+y*_realvidbm->bpr;

	for(p=0;p<4;p++) {
		outportb(0x3ce,4);
		outportb(0x3cf,p);
		vm2=vm1;
		dm2=dm1++;
		xst=xs;
		if (p<(x&3)) {
			--xst;
			++vm2;
		};
		if (xst)
			do {
				*dm2++=*vm2++;
				dm2+=3;
			} while(--xst);
	};
}

void	ovid_rbitblt(coord x,coord y,coord xs,coord ys,uchar *dm1,long dmod,boolean no_offset) {
	uchar *vm1,*vm2;
	uchar *dm2;
	uchar p;
	long vmod;
	coord xst,yst;

	if (xs<=0) return;
	if (ys<=0) return;

	xs=(xs+3+(x&3))>>2;
	x&=0xFFFFFFFC;

	vm1=(uchar *)0xA0000+(x>>2)+y*_realvidbm->bpr;
	if (!no_offset) dm1+=x+y*_vidbm->bpr;

	vmod=_realvidbm->bpr-xs;
	dmod-=(xs<<2);
	if(vmod<0 || dmod<0) return;

	for(p=1;p<16;p+=p) {
		outportb(0x3c4,2);
		outportb(0x3c5,p);
		vm2=vm1;
		dm2=dm1++;
		yst=ys;
		do {
			xst=xs;
			do {
				*vm2++=*dm2++;
				dm2+=3;
			} while(--xst);
			vm2+=vmod;
			dm2+=dmod;
		} while(--yst);
	};
}

/* be very careful with the pointer functions -- they are called
** when SS != DS, and ES != DS.  That means use qmovebes() and no
** pointers to automatic variables. */

#pragma off (check_stack)

void	scroll_to(ushort x,ushort y) {
	if ((x-scroll_x) & 3) {
		// wait for hblank
		while(!(inportb(0x3da)&1));
		// tweak attribute controller
#if 0
		outportb(0x3c0,0x13);
		outportb(0x3c0,(x & 3)<<1);
		outportb(0x3c0,0x20);
#else
		outportb(0x3c0,0x33);
		outportb(0x3c0,(x & 3)<<1);
#endif
	}
	scroll_x=x;
	scroll_y=y;
	y*=160;
	y+=(x>>2);

	outportb(0x3d4,0x0c);	// high
	outportb(0x3d5,y>>8);
	outportb(0x3d4,0x0d);	// low
	outportb(0x3d5,y&255);
}

static	uchar	v13x_ms_plane;

void	ovid_drawptr(struct MousePointerInfo *mpi,coord x,coord y) {
	uchar *vm1=(uchar *)0xA0000,*vm2;
	uchar *dm1=mpi->image,*dm2;
	uchar *bm1=mpi->backup;
	ushort xh=mpi->s_width,yh=mpi->s_height;
	ushort yht;
	ulong vmod;
	uchar c,p;

 	// check to see if we need to scroll the VGA

	if (mpi->cur_y < scroll_y) {
		if (mpi->cur_x < scroll_x)
			scroll_to(mpi->cur_x,mpi->cur_y);
		else if (mpi->cur_x >= scroll_x+320)
			scroll_to(mpi->cur_x-319,mpi->cur_y);
		else
			scroll_to(scroll_x,mpi->cur_y);
	} else if (mpi->cur_y >= scroll_y+200) {
		if (mpi->cur_x < scroll_x)
			scroll_to(mpi->cur_x,mpi->cur_y-199);
		else if (mpi->cur_x >= scroll_x+320)
			scroll_to(mpi->cur_x-319,mpi->cur_y-199);
		else
			scroll_to(scroll_x,mpi->cur_y-199);
	} else	if (mpi->cur_x < scroll_x)
			scroll_to(mpi->cur_x,scroll_y);
		else if (mpi->cur_x >= scroll_x+320)
			scroll_to(mpi->cur_x-319,scroll_y);

	if (x>=0)	vm1+=(x>>2);
	if (y>=0)	vm1+=(y*_realvidbm->bpr);
	mpi->vdp=vm1;

	if (y<0) {
		yh=mpi->s_height+y;
		dm1-=y*mpi->s_width;
	} else if (y>(_vidbm->height-mpi->s_height)) yh=_vidbm->height-y;

	if (x<0) {
		xh+=x;
		dm1-=x;
		x=0;
	} else if (x>(_vidbm->width-mpi->s_width)) xh=_vidbm->width-x;

	mpi->width=xh;
	mpi->height=yh;

	if (xh<=0) return;
	if (yh<=0) return;

	vmod=_realvidbm->bpr;
	v13x_ms_plane=p=1<<(x & 3);

	outportb(0x3ce,5);
	outportb(0x3cf,inportb(0x3cf) & 0xF7);	/* use read mode 0 */

	do {
		outportb(0x3c4,2);	/* sequencer set write mask */
		outportb(0x3c5,p);
		outportb(0x3ce,4);	/* set read plane */
		outportb(0x3cf,x++ & 3);
		vm2=vm1;
		if ((p+=p)==16) {
			p=1;
			vm1++;
		};
		dm2=dm1++;
		yht=yh;
		do {
			*bm1++=*vm2;
			if (c=*dm2) *vm2=(c-1 ? mpi->high : mpi->low);
			dm2+=mpi->s_width;
			vm2+=vmod;
		} while(--yht);
	} while(--xh);
}

void	ovid_eraseptr(struct MousePointerInfo *mpi) {
	uchar *vm1=mpi->vdp,*vm2;
	uchar *dm=mpi->backup;
	uchar p;
	ulong vbpr;
	ushort xs,ys,yst;

	if ((xs=mpi->width)<=0) return;
	if ((ys=mpi->height)<=0) return;

	vbpr=_realvidbm->bpr;

	p=v13x_ms_plane;

	do {
		outportb(0x3c4,2);
		outportb(0x3c5,p);
		vm2=vm1;
		if ((p+=p)==16) {
			p=1;
			vm1++;
		};
		yst=ys;
		do {
			*vm2=*dm++;
			vm2+=vbpr;
		} while(--yst);
	} while(--xs);
}

